@@ -1,7 +1,7 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -29,56 +29,18 @@
   (unit driver)
   (disable-warning var))
 
-
-(private
- compiler
-  compiler-arguments process-command-line dump-nodes dump-undefined-globals
-  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
-  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
-  standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  compiler-cleanup-hook check-global-exports disabled-warnings check-global-imports
-  file-io-only undefine-shadowed-macros profiled-procedures
-  unit-name insert-timer-checks used-units inline-max-size
-  debugging perform-lambda-lifting! disable-stack-overflow-checking
-  foreign-declarations emit-trace-info block-compilation line-number-database-size
-  target-heap-size target-stack-size target-heap-growth target-heap-shrinkage
-  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  target-initial-heap-size postponed-initforms
-  current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
-  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  broken-constant-nodes inline-substitutions-enabled compiler-macros-enabled
-  emit-profile profile-lambda-list profile-lambda-index profile-info-vector-name
-  direct-call-ids foreign-type-table first-analysis emit-closure-info
-  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
-  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call emit-unsafe-marker
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! source-filename
-  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
-  string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
-  collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
-  put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
-  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
-  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list external-protos-first
-  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database dump-exported-globals
-  default-declarations units-used-by-default words-per-flonum default-debugging-declarations
-  default-profiling-declarations default-optimization-passes
-  inline-max-size file-requirements use-import-table lookup-exports-file
-  foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
-  membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  chop-separator chop-extension display-real-name-table display-line-number-database explicit-use-flag
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  export-list do-lambda-lifting compiler-warning export-file-name
-  foreign-argument-conversion foreign-result-conversion)
-
-
+(include "compiler-namespace")
 (include "tweaks")
 
 (define-constant default-profile-name "PROFILE")
-(define-constant default-inline-max-size 10)
 (define-constant funny-message-timeout 60000)
 
+(define user-options-pass (make-parameter #f))
+(define user-read-pass (make-parameter #f))
+(define user-preprocessor-pass (make-parameter #f))
+(define user-pass (make-parameter #f))
+(define user-post-analysis-pass (make-parameter #f))
+
 
 ;;; Compile a complete source file:
 
@@ -93,12 +55,11 @@
   (initialize-compiler)
   (set! explicit-use-flag (memq 'explicit-use options))
   (let ([initforms `((##core#declare
-		      ,@(map (lambda (x) `(quote ,x))
-			     (append 
-			      default-declarations
-			      (if explicit-use-flag
-				  '()
-				  `((uses ,@units-used-by-default)) ) ) ) ) ) ]
+		      ,@(append 
+			 default-declarations
+			 (if explicit-use-flag
+			     '()
+			     `((uses ,@units-used-by-default)) ) ) ) ) ]
         [verbose (memq 'verbose options)]
 	[outfile (cond [(memq 'output-file options) 
 			=> (lambda (node)
@@ -108,7 +69,7 @@
 				   oname) ) ) ]
 		       [(memq 'to-stdout options) #f]
 		       [else (make-pathname #f (if filename (pathname-file filename) "out") "c")] ) ]
-	[ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))]
+	[ipath (map chop-separator (string-split (or (get-environment-variable "CHICKEN_INCLUDE_PATH") "") ";"))]
 	[opasses default-optimization-passes]
 	[time0 #f]
 	[time-breakdown #f]
@@ -126,15 +87,17 @@
 	[a-only (memq 'analyze-only options)]
 	[dynamic (memq 'dynamic options)]
 	[dumpnodes #f]
-	[quiet (memq 'quiet options)]
 	[start-time #f]
 	(upap #f)
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
 
     (define (cputime) (##sys#fudge 6))
 
+    (define (dribble fstr . args)
+      (when verbose (printf "~?~%~!" fstr args)))
+
     (define (print-header mode dbgmode)
-      (when verbose (printf "pass: ~a~%~!" mode))
+      (dribble "pass: ~a" mode)
       (and (memq dbgmode debugging-chicken)
 	   (begin
 	     (printf "[~a]~%" mode)
@@ -204,6 +167,11 @@
 
     (when uunit
       (set! unit-name (string->c-identifier (stringify (option-arg uunit)))) )
+    (when (or unit-name dynamic)
+      (set! standalone-executable #f))
+    (when (memq 'ignore-repository options)
+      (set! ##sys#dload-disabled #t)
+      (repository-path #f))
     (set! debugging-chicken 
       (append-map
        (lambda (do)
@@ -211,45 +179,59 @@
 	      (string->list do) ) )
        (collect-options 'debug) ) )
     (set! dumpnodes (memq '|D| debugging-chicken))
+    (set! import-libraries
+      (map (lambda (il)
+	     (cons (string->symbol il) 
+		   (string-append il ".import.scm")))
+	   (collect-options 'emit-import-library)))
     (when (memq 'lambda-lift options) (set! do-lambda-lifting #t))
-    (when (memq 'disable-compiler-macros options) (set! compiler-macros-enabled #f))
+    (when (memq 'scrutinize options) (set! do-scrutinize #t))
     (when (memq 't debugging-chicken) (##sys#start-timer))
     (when (memq 'b debugging-chicken) (set! time-breakdown #t))
-    (and-let* ((xfile (memq 'emit-exports options)))
-      (set! export-file-name (cadr xfile)) )
+    (when (memq 'emit-exports options)
+      (warning "deprecated compiler option: emit-exports") )
     (when (memq 'raw options)
       (set! explicit-use-flag #t)
       (set! cleanup-forms '())
       (set! initforms '()) )
     (when (memq 'no-lambda-info options)
       (set! emit-closure-info #f) )
-    (set! use-import-table (memq 'check-imports options))
-    (let ((imps (collect-options 'import)))
-      (when (pair? imps)
-	(set! use-import-table #t)
-	(for-each lookup-exports-file imps) ) )
+    (when (memq 'no-compiler-syntax options)
+      (set! compiler-syntax-enabled #f))
+    (when (memq 'local options)
+      (set! local-definitions #t))
+    (when (memq 'inline-global options)
+      (set! enable-inline-files #t)
+      (set! inline-locally #t)
+      (set! inline-globally #t))
     (set! disabled-warnings (map string->symbol (collect-options 'disable-warning)))
     (when (memq 'no-warnings options) 
-      (when verbose (printf "Warnings are disabled~%~!"))
+      (dribble "Warnings are disabled")
       (set! ##sys#warnings-enabled #f) )
     (when (memq 'optimize-leaf-routines options) (set! optimize-leaf-routines #t))
     (when (memq 'unsafe options) 
-      (set! unsafe #t)
-      (##match#set-error-control #:fail) )
+      (set! unsafe #t) )
     (when (and dynamic (memq 'unsafe-libraries options))
       (set! emit-unsafe-marker #t) )
+    (when (memq 'setup-mode options)
+      (set! ##sys#setup-mode #t))
     (when (memq 'disable-interrupts options) (set! insert-timer-checks #f))
     (when (memq 'fixnum-arithmetic options) (set! number-type 'fixnum))
     (when (memq 'block options) (set! block-compilation #t))
-    (when (memq 'emit-external-prototypes-first options) (set! external-protos-first #t))
-    (when (memq 'inline options) (set! inline-max-size default-inline-max-size))
+    (when (memq 'emit-external-prototypes-first options)
+      (set! external-protos-first #t))
+    (when (memq 'inline options) (set! inline-locally #t))
+    (and-let* ((ifile (memq 'emit-inline-file options)))
+      (set! inline-locally #t)		; otherwise this option makes no sense
+      (set! local-definitions #t)
+      (set! inline-output-file (option-arg ifile)))
     (and-let* ([inlimit (memq 'inline-limit options)])
       (set! inline-max-size 
 	(let ([arg (option-arg inlimit)])
 	  (or (string->number arg)
 	      (quit "invalid argument to `-inline-limit' option: `~A'" arg) ) ) ) )
     (when (memq 'case-insensitive options) 
-      (when verbose (printf "Identifiers and symbols are case insensitive~%~!"))
+      (dribble "Identifiers and symbols are case insensitive")
       (register-feature! 'case-insensitive)
       (case-sensitive #f) )
     (when (memq 'compress-literals options)
@@ -260,6 +242,18 @@
 	      [(string=? "none" val) (keyword-style #:none)]
 	      [(string=? "suffix" val) (keyword-style #:suffix)]
 	      [else (quit "invalid argument to `-keyword-style' option")] ) ) )
+    (when (memq 'no-parenthesis-synonyms options)
+      (dribble "Disabled support for parenthesis synonyms")
+      (parenthesis-synonyms #f) )
+    (when (memq 'no-symbol-escape options) 
+      (dribble "Disabled support for escaped symbols")
+      (symbol-escape #f) )
+    (when (memq '("-r5rs-syntax") options)
+      (dribble "Disabled the Chicken extensions to R5RS syntax")
+      (case-sensitive #f)
+      (keyword-style #:none)
+      (parentheses-synonyms #f)
+      (symbol-escape #f) )
     (set! verbose-mode verbose)
     (set! ##sys#read-error-with-line-number #t)
     (set! ##sys#include-pathnames
@@ -268,13 +262,26 @@
 	      ipath) )
     (when (and outfile filename (string=? outfile filename))
       (quit "source- and output-filename are the same") )
-    (set! uses-units
-      (map string->symbol 
-	   (append-map
-	    (cut string-split <> ",")
-	    (collect-options 'uses))))
+    (set! uses-units (map string->symbol (collect-options 'uses)))
     (when (memq 'keep-shadowed-macros options)
       (set! undefine-shadowed-macros #f) )
+    (when (memq 'no-argc-checks options)
+      (set! no-argc-checks #t) )
+    (when (memq 'no-bound-checks options)
+      (set! no-bound-checks #t) )
+    (when (memq 'no-procedure-checks options)
+      (set! no-procedure-checks #t) )
+    (when (memq 'no-procedure-checks-for-usual-bindings options)
+      (for-each 
+       (lambda (v)
+         (mark-variable v '##compiler#always-bound-to-procedure)
+         (mark-variable v '##compiler#always-bound) )
+       default-standard-bindings)
+      (for-each 
+       (lambda (v)
+         (mark-variable v '##compiler#always-bound-to-procedure)
+         (mark-variable v '##compiler#always-bound) )
+       default-extended-bindings) )
 
     ;; Handle feature options:
     (for-each 
@@ -284,43 +291,32 @@
     ;; Load extensions:
     (set! ##sys#features (cons #:compiler-extension ##sys#features))
     (let ([extends (collect-options 'extend)])
-      (when verbose
-	(printf "Loading compiler extensions...~%~!")
-	(load-verbose #t) )
-      (for-each (lambda (f) (load (##sys#resolve-include-filename f #f #t))) extends) )
+      (dribble "Loading compiler extensions...")
+      (when verbose (load-verbose #t))
+      (for-each
+       (lambda (f) (load (##sys#resolve-include-filename f #f #t))) 
+       extends) )
     (set! ##sys#features (delete #:compiler-extension ##sys#features eq?))
 
     (set! ##sys#features (cons '#:compiling ##sys#features))
-    (set! ##sys#features (cons #:match ##sys#features))
-    (##sys#provide 'match) 
     (set! upap (user-post-analysis-pass))
 
     ;; Insert postponed initforms:
     (set! initforms (append initforms postponed-initforms))
 
-    ;; Handle `-extension' options:
-    (when (memq 'extension options)
-      (set! initforms 
-	(append 
-	 initforms
-	 `((define-extension 
-	     ,(string->symbol
-	       (cond (outfile (pathname-file outfile))
-		     (filename (pathname-file filename))
-		     (else (quit "no filename available for `-extension' option")) ) ) ) ) ) ) )
-
-    ;; Append required extensions to initforms:
-    (let ([ids (lset-difference 
-		eq?
-		(map string->symbol
-		     (append-map
-		      (cut string-split <> ",")
-		      (collect-options 'require-extension)))
-		uses-units)])
+    (let ((se (map string->symbol (collect-options 'static-extension))))
+      ;; Append required extensions to initforms:
       (set! initforms
-	(append initforms (map (lambda (r) `(##core#require-extension ',r)) ids)) ) )
+	(append 
+	 initforms 
+	 (map (lambda (r) `(##core#require-extension (,r) #t)) 
+	      (append se (map string->symbol (collect-options 'require-extension))))))
 
-    (when (memq 'run-time-macros options)
+      ;; add static-extensions as used units:
+      (set! ##sys#explicit-library-modules
+	(append ##sys#explicit-library-modules se)))
+
+    (when (memq 'compile-syntax options)
       (set! ##sys#enable-runtime-macros #t) )
     (set! target-heap-size
       (if hsize
@@ -343,15 +339,14 @@
     (unless (memq 'no-usual-integrations options)
       (set! standard-bindings default-standard-bindings)
       (set! extended-bindings default-extended-bindings) )
-    (when verbose
-      (printf "debugging info: ~A~%~!"
-	      (if emit-trace-info
-		  "stacktrace"
-		  "none") ) )
+    (dribble "debugging info: ~A"
+	     (if emit-trace-info
+		 "calltrace"
+		 "none") )
     (when profile
       (let ([acc (eq? 'accumulate-profile (car profile))])
 	(set! emit-profile #t)
-	(set! profiled-procedures #f)
+	(set! profiled-procedures 'all)
 	(set! initforms
 	  (append
 	   initforms
@@ -359,8 +354,10 @@
 	   (if acc
 	       '((set! ##sys#profile-append-mode #t))
 	       '() ) ) )
-	(when verbose
-	  (printf "Generating ~aprofile~%~!" (if acc "accumulated " "")) ) ) )
+	(dribble "Generating ~aprofile" (if acc "accumulated " "")) ) )
+
+    ;;*** hardcoded "modules.db" is bad (also used in chicken-install.scm)
+    (load-identifier-database "modules.db")
 
     (cond ((memq 'version options)
 	   (print-version #t)
@@ -371,14 +368,12 @@
 	   (display (chicken-version)) 
 	   (newline) )
 	  ((not filename)
-	   (unless quiet
-	     (print-version #t)
-	     (display "\nEnter \"chicken -help\" for information on how to use it.\n") ) )
+	   (print-version #t)
+	   (display "\nEnter \"chicken -help\" for information on how to use it.\n") )
 	  (else
 
 	   ;; Display header:
-	   (unless quiet
-	     (printf "compiling `~a' ...~%" filename) )
+	   (dribble "compiling `~a' ..." filename)
 	   (set! source-filename filename)
 	   (debugging 'r "options" options)
 	   (debugging 'r "debugging options" debugging-chicken)
@@ -397,7 +392,7 @@
 
 	     (let ([proc (user-read-pass)])
 	       (cond [proc
-		      (when verbose (printf "User read pass...~%~!"))
+		      (dribble "User read pass...")
 		      (set! forms (proc prelude files postlude)) ]
 		     [else
 		      (do ([files files (cdr files)])
@@ -418,7 +413,7 @@
 	   ;; Start compilation passes:
 	   (let ([proc (user-preprocessor-pass)])
 	     (when proc
-	       (when verbose (printf "User preprocessing pass...~%~!"))
+	       (dribble "User preprocessing pass...")
 	       (set! forms (map proc forms))))
 
 	   (print-expr "source" '|1| forms)
@@ -450,24 +445,23 @@
 			     '() )
 			 '((##core#undefined))) ] )
 
+	     (when (and (pair? compiler-syntax-statistics)
+			(debugging 'x "applied compiler syntax:"))
+	       (for-each 
+		(lambda (cs) (printf "  ~a\t\t~a~%" (car cs) (cdr cs)))
+		compiler-syntax-statistics))
    	     (when (debugging '|N| "real name table:")
 	       (display-real-name-table) )
 	     (when (debugging 'n "line number database:")
 	       (display-line-number-database) )
 
-	     (when (and block-compilation unit-name)
-	       (compiler-warning 
-		'usage
-		"compilation of library unit `~a' in block-mode - globals may not be accessible outside this unit"
-		unit-name) )
-
 	     (when (and unit-name dynamic)
 	       (compiler-warning 'usage "library unit `~a' compiled in dynamic mode" unit-name) )
 
 	     (when (and unsafe (feature? 'compiling-extension))
 	       (compiler-warning 
 		'style
-		"compiling extensions in unsafe mode is bad practice and should be avoided as it may be surprising to an unsuspecting user") )
+		"compiling extensions in unsafe mode is bad practice and should be avoided") )
 
 	     (set! ##sys#line-number-database line-number-database-2)
 	     (set! line-number-database-2 #f)
@@ -479,42 +473,71 @@
 
 	     (let ([proc (user-pass)])
 	       (when proc
-		 (when verbose (printf "User pass...~%~!"))
+		 (dribble "User pass...")
 		 (begin-time)
 		 (set! exps (map proc exps))
 		 (end-time "user pass") ) )
 
-	     (let* ([node0 (make-node
-			    'lambda '(())
-			    (list (build-node-graph
-				   (canonicalize-begin-body exps) ) ) ) ] 
-		    [proc (user-pass-2)] )
-	       (when (debugging 'M "; requirements:")
-		 (pretty-print (##sys#hash-table->alist file-requirements)))
-	       (when proc
-		 (when verbose (printf "Secondary user pass...~%"))
+	     (let ((node0 (make-node
+			   'lambda '(())
+			   (list (build-node-graph
+				  (canonicalize-begin-body exps) ) ) ) ) 
+		   (db #f))
+
+	       (print-node "initial node tree" '|T| node0)
+	       (initialize-analysis-database)
+
+	       (when do-scrutinize
+		 ;;;*** hardcoded database file name
+		 (unless (memq 'ignore-repository options)
+		   (load-type-database "types.db"))
+		 (for-each (cut load-type-database <> #f) (collect-options 'types))
 		 (begin-time)
 		 (set! first-analysis #f)
-		 (let ([db (analyze 'user node0)])
-		   (print-db "analysis (u)" '|0| db 0)
-		   (end-time "pre-analysis (u)")
-		   (begin-time)
-		   (proc node0)
-		   (end-time "secondary user pass")
-		   (print-node "secondary user pass" '|U| node0) )
+		 (set! db (analyze 'scrutiny node0))
+		 (print-db "analysis" '|0| db 0)
+		 (end-time "pre-analysis")
+		 (begin-time)
+		 (debugging 'p "performing scrutiny")
+		 (scrutinize node0 db)
+		 (end-time "scrutiny")
 		 (set! first-analysis #t) )
 
 	       (when do-lambda-lifting
 		 (begin-time)
-		 (set! first-analysis #f)
-		 (let ([db (analyze 'lift node0)])
+		 (unless do-scrutinize	; no need to do analysis if already done above
+		   (set! first-analysis #f)
+		   (set! db (analyze 'lift node0))
 		   (print-db "analysis" '|0| db 0)
-		   (end-time "pre-analysis")
-		   (begin-time)
-		   (perform-lambda-lifting! node0 db)
-		   (end-time "lambda lifting")
-		   (print-node "lambda lifted" '|L| node0) )
+		   (end-time "pre-analysis (lambda-lift)"))
+		 (begin-time)
+		 (perform-lambda-lifting! node0 db)
+		 (end-time "lambda lifting")
+		 (print-node "lambda lifted" '|L| node0) 
 		 (set! first-analysis #t) )
+	       
+	       (let ((req (concatenate (vector->list file-requirements))))
+		 (when (debugging 'M "; requirements:")
+		   (pp req))
+		 (when enable-inline-files
+		   (for-each
+		    (lambda (id)
+		      (and-let* ((ifile (##sys#resolve-include-filename 
+					 (make-pathname #f (symbol->string id) "inline")
+					 #f #t))
+				 ((file-exists? ifile)))
+			(dribble "Loading inline file ~a ..." ifile)
+			(load-inline-file ifile)))
+		    (concatenate (map cdr req))) )
+		 (let ((ifs (collect-options 'consult-inline-file)))
+		   (unless (null? ifs)
+		     (set! inline-globally #t)
+		     (set! inline-locally #t)
+		     (for-each 
+		      (lambda (ilf)
+			(dribble "Loading inline file ~a ..." ilf)
+			(load-inline-file ilf) )
+		      ifs))))
 
 	       (set! ##sys#line-number-database #f)
 	       (set! constant-table #f)
@@ -533,10 +556,12 @@
 		   (begin-time)
 		   (let ([db (analyze 'opt node2 i progress)])
 		     (when first-analysis
-		       (when use-import-table (check-global-imports db))
-		       (check-global-exports db)
 		       (when (memq 'u debugging-chicken)
-			 (dump-undefined-globals db)) )
+			 (dump-undefined-globals db))
+		       (when (memq 'd debugging-chicken)
+			 (dump-defined-globals db))
+		       (when (memq 'v debugging-chicken)
+			 (dump-global-refs db)) )
 		     (set! first-analysis #f)
 		     (end-time "analysis")
 		     (print-db "analysis" '|4| db i)
@@ -547,7 +572,8 @@
 			    (debugging 'p "optimization pass" i)
 
 			    (begin-time)
-			    (receive (node2 progress-flag) (perform-high-level-optimizations node2 db)
+			    (receive (node2 progress-flag)
+				(perform-high-level-optimizations node2 db)
 			      (end-time "optimization")
 			      (print-node "optimized-iteration" '|5| node2)
 
@@ -569,14 +595,17 @@
 			   [else
 			    (print-node "optimized" '|7| node2)
 
+			    (when inline-output-file
+			      (let ((f inline-output-file))
+				(dribble "Generating global inline file `~a' ..." f)
+				(emit-global-inline-file f db) ) )
+
 			    (begin-time)
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
 			      (when (and ##sys#warnings-enabled (> (- (cputime) start-time) funny-message-timeout))
-				(display "(do not worry - still compiling...)\n") )
-			      (when export-file-name
-				(dump-exported-globals db export-file-name) )
+				(display "(don't worry - still compiling...)\n") )
 			      (when a-only (exit 0))
 			      (print-node "closure-converted" '|9| node3)
 
@@ -587,12 +616,10 @@
 
                                 (begin-time)
 				(let ((out (if outfile (open-output-file outfile) (current-output-port))) )
-				  (unless quiet
-				    (printf "generating `~A' ...~%" outfile) )
+				  (dribble "generating `~A' ..." outfile)
 				  (generate-code literals lliterals lambdas out filename dynamic db)
 				  (when outfile (close-output-port out)))
                                 (end-time "code generation")
                                 (when (memq 't debugging-chicken) (##sys#display-times (##sys#stop-timer)))
                                 (compiler-cleanup-hook)
-                                (when verbose 
-                                  (printf "compilation finished.~%~!") ) ) ) ] ) ) ) ) ) ) ) ) ) )
+                                (dribble "compilation finished.") ) ) ] ) ) ) ) ) ) ) ) ) )
