@@ -5,4 +5,4 @@
       n
       (+ (fib (- n 1)) (fib (- n 2))) ) )
 
-(time (fib 30))
+(time (pp (fib 30)))
