@@ -1,7 +1,7 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,50 +28,7 @@
 (declare (unit backend))
 
 
-(private compiler
-  compiler-arguments process-command-line find-early-refs
-  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
-  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings 
-  foldable-extended-bindings
-  standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home optimization-iterations debugging cleanup
-  file-io-only
-  unit-name insert-timer-checks used-units inlining external-variables
-  foreign-declarations emit-trace-info block-compilation line-number-database-size
-  target-heap-size target-stack-size target-heap-growth target-heap-shrinkage
-  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  target-initial-heap-size disable-stack-overflow-checking
-  current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants
-  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used 
-  mutable-constants encode-literal
-  broken-constant-nodes inline-substitutions-enabled
-  direct-call-ids foreign-type-table first-analysis block-variable-literal?
-  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database 
-  scan-toplevel-assignments
-  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations 
-  perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub 
-  expand-foreign-lambda*
-  transform-direct-lambdas! target-include-file emit-unsafe-marker
-  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
-  string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
-  collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
-  put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
-  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node 
-  expression-has-side-effects? source-info->string
-  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
-  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables 
-  scan-free-variables external-protos-first emit-closure-info
-  topological-sort print-version print-usage initialize-analysis-database
-  generate-external-variables real-name real-name2 unique-id
-  default-declarations units-used-by-default words-per-flonum big-fixnum?
-  foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
-  membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  default-optimization-iterations generate-foreign-callback-header generate-foreign-callback-stub-prototypes
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion)
-
+(include "compiler-namespace")
 (include "tweaks")
 
 
@@ -219,9 +176,9 @@
 	     (expr (car subs) i) )
 
 	    ((##core#global)
-	     (let ([index (first params)]
-		   [safe (second params)] 
-		   [block (third params)] )
+	     (let ((index (first params))
+		   (safe (second params)) 
+		   (block (third params)) )
 	       (cond [block
 		      (if safe
 			  (gen "lf[" index "]")
@@ -230,23 +187,28 @@
 		     [else (gen "C_retrieve(lf[" index "])")] ) ) )
 
 	    ((##core#setglobal)
-	     (let ([index (first params)]
-		   [block (second params)] )
+	     (let ((index (first params))
+		   (block (second params)) 
+		   (var (third params)))
 	       (if block
-		   (gen "C_mutate(&lf[" index "],")
-		   (gen "C_mutate((C_word*)lf[" index "]+1,") )
+		   (gen "C_mutate(&lf[" index "]")
+		   (gen "C_mutate((C_word*)lf[" index "]+1") )
+	       (gen " /* (set! " (uncommentify (symbol->string var)) " ...) */,")
 	       (expr (car subs) i)
 	       (gen #\)) ) )
 
 	    ((##core#setglobal_i)
-	     (let ([index (first params)]
-		   [block (second params)] )
+	     (let ((index (first params))
+		   (block (second params)) 
+		   (var (third params)) )
 	       (cond [block
-		      (gen "lf[" index "]=")
+		      (gen "lf[" index "] /* "
+			   (uncommentify (symbol->string var)) " */ =")
 		      (expr (car subs) i)
 		      (gen #\;) ]
 		     [else
-		      (gen "C_set_block_item(lf[" index "],0,")
+		      (gen "C_set_block_item(lf[" index "] /* "
+			   (uncommentify (symbol->string var)) " */,0,")
 		      (expr (car subs) i)
 		      (gen #\)) ] ) ) )
 
@@ -299,6 +261,33 @@
 			     (unless empty-closure (gen #\t nc #\,))
 			     (expr-args args i)
 			     (gen ");") ) ) )
+		     ((and (eq? '##core#global (node-class fn))
+			   (not unsafe) 
+			   (not no-procedure-checks)
+			   (not (first params)))
+		      (let* ((gparams (node-parameters fn))
+			     (index (first gparams))
+			     (safe (second gparams)) 
+			     (block (third gparams)) 
+			     (carg #f))
+			(gen #t "((C_proc" nf ")")
+			(cond (block
+			       (set! carg (string-append "lf[" (number->string index) "]"))
+			       (if safe
+				   (gen "C_retrieve_proc(" carg ")")
+				   (gen "C_retrieve2_symbol_proc(" carg "," 
+					(c-ify-string (symbol->string (fourth gparams))) #\)) ) )
+			      (safe 
+			       (set! carg 
+				 (string-append "*((C_word*)lf[" (number->string index) "]+1)"))
+			       (gen "C_retrieve_proc(" carg ")"))
+			      (else
+			       (set! carg 
+				 (string-append "*((C_word*)lf[" (number->string index) "]+1)"))
+			       (gen "C_retrieve_symbol_proc(lf[" index "])") ))
+			(gen ")(" nf #\, carg #\,)
+			(expr-args args i)
+			(gen ");") ) )
 		     (else
 		      (gen #t #\t nc #\=)
 		      (expr fn i)
@@ -446,30 +435,34 @@
 	(if (< n 10)
 	    (string-append "0" (number->string n))
 	    n) )
-      (match (##sys#decode-seconds (current-seconds) #f)
-	[#(_ min hour mday mon year _ _ _ _)
-	  (gen "/* Generated from " source-file " by the CHICKEN compiler" #t
-	       "   http://www.call-with-current-continuation.org" #t
-	       "   " (+ 1900 year) #\- (pad0 (add1 mon)) #\- (pad0 mday) #\space (pad0 hour) #\: (pad0 min) #t
-	       (string-intersperse
-		(map (cut string-append "   " <> "\n") 
-		     (string-split (chicken-version #t) "\n") ) 
-		"")
-	       "   command line: ")
-	  (gen-list compiler-arguments)
+      (let* ((tm (##sys#decode-seconds (current-seconds) #f))
+	     (min (vector-ref tm 1))
+	     (hour (vector-ref tm 2))
+	     (mday (vector-ref tm 3))
+	     (mon (vector-ref tm 4))
+	     (year (vector-ref tm 5)) )
+	(gen "/* Generated from " source-file " by the CHICKEN compiler" #t
+	     "   http://www.call-with-current-continuation.org" #t
+	     "   " (+ 1900 year) #\- (pad0 (add1 mon)) #\- (pad0 mday) #\space (pad0 hour) #\: (pad0 min) #t
+	     (string-intersperse
+	      (map (cut string-append "   " <> "\n") 
+		   (string-split (chicken-version #t) "\n") ) 
+	      "")
+	     "   command line: ")
+	(gen-list compiler-arguments)
+	(gen #t)
+	(cond [unit-name (gen "   unit: " unit-name)]
+	      [else 
+	       (gen "   used units: ")
+	       (gen-list used-units) ] )
+	(gen #t "*/" #t #t "#include \"" target-include-file "\"")
+	(when external-protos-first
+	  (generate-foreign-callback-stub-prototypes foreign-callback-stubs) )
+	(when (pair? foreign-declarations)
 	  (gen #t)
-	  (cond [unit-name (gen "   unit: " unit-name)]
-		[else 
-		 (gen "   used units: ")
-		 (gen-list used-units) ] )
-	  (gen #t "*/" #t #t "#include \"" target-include-file "\"")
-	  (when external-protos-first
-	    (generate-foreign-callback-stub-prototypes foreign-callback-stubs) )
-	  (when (pair? foreign-declarations)
-	    (gen #t)
-	    (for-each (lambda (decl) (gen #t decl)) foreign-declarations) )
-	  (unless external-protos-first
-	    (generate-foreign-callback-stub-prototypes foreign-callback-stubs) ) ] ) )
+	  (for-each (lambda (decl) (gen #t decl)) foreign-declarations) )
+	(unless external-protos-first
+	  (generate-foreign-callback-stub-prototypes foreign-callback-stubs) ) ) )
   
     (define (trailer)
       (gen #t "/* end of file */" #t) )
@@ -914,7 +907,7 @@
       ((null? ll)
        (gen #t "{NULL,NULL}};") )
     (let ((id (lambda-literal-id (car ll))))
-      (gen #t "{\"" id sf "\",(void*)")
+      (gen #t "{\"" id #\: (string->c-identifier sf) "\",(void*)")
       (if (eq? 'toplevel id)
 	  (if unit-name
 	      (gen "C_" unit-name "_toplevel},")
@@ -967,9 +960,11 @@
 (define (generate-external-variables vars)
   (gen #t)
   (for-each
-   (match-lambda 
-     [#(name type exported)
-      (gen #t (if exported "" "static ") (foreign-type-declaration type name) #\;) ] )
+   (lambda (v)
+     (let ((name (vector-ref v 0))
+	   (type (vector-ref v 1))
+	   (exported (vector-ref v 2)) )
+       (gen #t (if exported "" "static ") (foreign-type-declaration type name) #\;) ) )
    vars) )
 
 
@@ -1155,52 +1150,65 @@
       [(s32vector nonnull-s32vector) (str "int *")]
       [(f32vector nonnull-f32vector) (str "float *")]
       [(f64vector nonnull-f64vector) (str "double *")]
-      [(nonnull-c-string c-string nonnull-c-string* c-string* 
-			 nonnull-unsigned-c-string nonnull-unsigned-c-string* unsigned-c-string*
-			 symbol) (str "char *")]
+      [(nonnull-c-string c-string nonnull-c-string* c-string* symbol) 
+       (str "char *")]
+      [(nonnull-unsigned-c-string nonnull-unsigned-c-string* unsigned-c-string unsigned-c-string*)
+       (str "unsigned char *")]
       [(void) (str "void")]
       [else
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
 	      => (lambda (t)
 		   (foreign-type-declaration (if (vector? t) (vector-ref t 0) t) target)) ]
 	     [(string? type) (str type)]
-	     [(pair? type)
-	      (match type
-		[((or 'pointer 'nonnull-pointer 'c-pointer 'nonnull-c-pointer) ptype)
-		 (foreign-type-declaration ptype (string-append "*" target)) ]
-		[('ref rtype)
-		 (foreign-type-declaration rtype (string-append "&" target)) ]
-		[`(template ,t0 ,ts ...)
-		 (str
-		  (string-append 
-		   (foreign-type-declaration t0 "")
-		   "<"
-		   (string-intersperse (map (cut foreign-type-declaration <> "") ts) ",")
-		   "> ") ) ]
-		[`(const ,t) (string-append "const " (foreign-type-declaration t target))]
-		[`(struct ,sname) (string-append "struct " (->string sname) " " target)]
-		[`(union ,uname) (string-append "union " (->string uname) " " target)]
-		[`(enum ,ename) (string-append "enum " (->string ename) " " target)]
-		[((or 'instance 'nonnull-instance) cname sname) (string-append (->string cname) "*" target)]
-		[('instance-ref cname sname) (string-append (->string cname) "&" target)]
-		[`(function ,rtype ,argtypes . ,callconv)
-		 (string-append
-		  (foreign-type-declaration rtype "")
-		  (or (and-let* ([(pair? callconv)]
-				 [cc (car callconv)]
-				 [(string? cc)] )
-			cc)
-		      "")
-		  " (*" target ")("
-		  (string-intersperse
-		   (map (lambda (at)
-			  (if (eq? '... at) 
-			      "..."
-			      (foreign-type-declaration at "") ) )
-			argtypes) 
-		   ",")
-		  ")" ) ]
-		[_ (err)] ) ]
+	     [(list? type)
+	      (let ((len (length type)))
+		(cond 
+		 ((and (= 2 len)
+		       (memq (car type) '(pointer nonnull-pointer c-pointer 
+						  nonnull-c-pointer) ) )
+		  (foreign-type-declaration (cadr type) (string-append "*" target)) )
+		 ((and (= 2 len)
+		       (eq? 'ref (car type)))
+		  (foreign-type-declaration (cadr type) (string-append "&" target)) )
+		 ((and (> len 2)
+		       (eq? 'template (car type)))
+		  (str
+		   (string-append 
+		    (foreign-type-declaration (cadr type) "")
+		    "<"
+		    (string-intersperse
+		     (map (cut foreign-type-declaration <> "") (cddr type))
+		     ",")
+		    "> ") ) )
+		 ((and (= len 2) (eq? 'const (car type)))
+		  (string-append "const " (foreign-type-declaration (cadr type) target)))
+		 ((and (= len 2) (eq? 'struct (car type)))
+		  (string-append "struct " (->string (cadr type)) " " target))
+		 ((and (= len 2) (eq? 'union (car type)))
+		  (string-append "union " (->string (cadr type)) " " target))
+		 ((and (= len 2) (eq? 'enum (car type)))
+		  (string-append "enum " (->string (cadr type)) " " target))
+		 ((and (= len 3) (memq (car type) '(instance nonnull-instance)))
+		  (string-append (->string (cadr type)) "*" target))
+		 ((and (= len 3) (eq? 'instance-ref (car type)))
+		  (string-append (->string (cadr type)) "&" target))
+		 ((and (>= len 3) (eq? 'function (car type)))
+		  (let ((rtype (cadr type))
+			(argtypes (caddr type))
+			(callconv (optional (cdddr type) "")))
+		    (string-append
+		     (foreign-type-declaration rtype "")
+		     callconv
+		     " (*" target ")("
+		     (string-intersperse
+		      (map (lambda (at)
+			     (if (eq? '... at) 
+				 "..."
+				 (foreign-type-declaration at "") ) )
+			   argtypes) 
+		      ",")
+		     ")" ) ) )
+		 (else (err)) ) ) ]
 	     [else (err)] ) ] ) ) )
 
 
@@ -1256,21 +1264,24 @@
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
 	      => (lambda (t)
 		   (foreign-argument-conversion (if (vector? t) (vector-ref t 0) t)) ) ]
-	     [(pair? type)
-	      (match type
-                ;; pointer and nonnull-pointer are DEPRECATED
-		[('pointer ptype) "C_c_pointer_or_null("]
-		[('nonnull-pointer ptype) "C_c_pointer_nn("]
-		[('c-pointer ptype) "C_c_pointer_or_null("]
-		[('nonnull-c-pointer ptype) "C_c_pointer_nn("]
-		[`(instance ,cname ,sname) "C_c_pointer_or_null("]
-		[`(nonnull-instance ,cname ,sname) "C_c_pointer_nn("]
-		[`(function ,rtype ,@argtypes) "C_c_pointer_or_null("]
-		[`(const ,ctype) (foreign-argument-conversion ctype)]
-		[`(enum ,etype) "C_num_to_int("]
-		[`(ref ,rtype) (string-append "*(" (foreign-type-declaration rtype "*") ")C_c_pointer_nn(")]
-		[`(instance-ref ,cname ,sname) (string-append "*(" cname "*)C_c_pointer_nn(")]
-		[else (err)] ) ]
+	     [(and (list? type) (>= (length type) 2))
+	      (case (car type)
+	       ;; pointer and nonnull-pointer are DEPRECATED
+	       ((pointer) "C_c_pointer_or_null(")
+	       ((nonnull-pointer) "C_c_pointer_nn(")
+	       ((c-pointer) "C_c_pointer_or_null(")
+	       ((nonnull-c-pointer) "C_c_pointer_nn(")
+	       ((instance) "C_c_pointer_or_null(")
+	       ((nonnull-instance) "C_c_pointer_nn(")
+	       ((function) "C_c_pointer_or_null(")
+	       ((const) (foreign-argument-conversion (cadr type)))
+	       ((enum) "C_num_to_int(")
+	       ((ref)
+		(string-append "*(" (foreign-type-declaration (car type) "*")
+			       ")C_c_pointer_nn("))
+	       ((instance-ref)
+		(string-append "*(" (cadr type) "*)C_c_pointer_nn("))
+	       (else (err)) ) ]
 	     [else (err)] ) ) ) ) )
 
 
@@ -1304,24 +1315,24 @@
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
 	      => (lambda (x)
 		   (foreign-result-conversion (if (vector? x) (vector-ref x 0) x) dest)) ]
-	     [(pair? type)
-	      (match type
-		[((or 'nonnull-pointer 'nonnull-c-pointer) ptype) 
-		 (sprintf "C_mpointer(&~A,(void*)" dest) ]
-		[('ref rtype) 
-		 (sprintf "C_mpointer(&~A,(void*)&" dest) ]
-		[('instance cname sname)
-		 (sprintf "C_mpointer_or_false(&~A,(void*)" dest) ]
-		[('nonnull-instance cname sname)
-		 (sprintf "C_mpointer(&~A,(void*)" dest) ]
-		[('instance-ref cname sname)
-		 (sprintf "C_mpointer(&~A,(void*)&" dest) ]
-		[('const ctype) (foreign-result-conversion ctype dest)]
-		[((or 'pointer 'c-pointer) ptype) 
-		 (sprintf "C_mpointer_or_false(&~a,(void*)" dest) ]
-		[`(function ,rtype ,@argtypes) (sprintf "C_mpointer(&~a,(void*)" dest)]
-		[`(enum ,etype) (sprintf "C_int_to_num(&~a," dest)]
-		[else (err)] ) ]
+	     [(and (list? type) (>= (length type) 2))
+	      (case (car type)
+		((nonnull-pointer nonnull-c-pointer)
+		 (sprintf "C_mpointer(&~A,(void*)" dest) )
+		((ref)
+		 (sprintf "C_mpointer(&~A,(void*)&" dest) )
+		((instance)
+		 (sprintf "C_mpointer_or_false(&~A,(void*)" dest) )
+		((nonnull-instance)
+		 (sprintf "C_mpointer(&~A,(void*)" dest) )
+		((instance-ref)
+		 (sprintf "C_mpointer(&~A,(void*)&" dest) )
+		((const) (foreign-result-conversion (cadr type) dest))
+		((pointer c-pointer)
+		 (sprintf "C_mpointer_or_false(&~a,(void*)" dest) )
+		((function) (sprintf "C_mpointer(&~a,(void*)" dest))
+		((enum) (sprintf "C_int_to_num(&~a," dest))
+		(else (err)) ) ]
 	     [else (err)] ) ) ) ) )
 
 
@@ -1374,7 +1385,7 @@ return((C_header_bits(lit) >> 24) & 0xff);
 	     (encode-size (string-length str))
 	     str) ) )
 	 ((##sys#immediate? lit)
-	  (bomb "invalid literal - can not encode" lit))
+	  (bomb "invalid literal - cannot encode" lit))
 	 ((##core#inline "C_byteblockp" lit)
 	  (##sys#string-append ; relies on the fact that ##sys#string-append doesn't check
 	   (string-append
