@@ -1,7 +1,7 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,43 +28,7 @@
 (declare (unit platform))
 
 
-(private compiler
-  compiler-arguments process-command-line
-  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
-  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
-  standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home debugging
-  dump-nodes
-  unit-name insert-timer-checks used-units inlining
-  foreign-declarations block-compilation line-number-database-size
-  target-heap-size target-stack-size 
-  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
-  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  broken-constant-nodes inline-substitutions-enabled
-  direct-call-ids foreign-type-table first-analysis
-  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
-  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! decompose-lambda-list rewrite
-  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
-  string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
-  collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
-  put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
-  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
-  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
-  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database
-  default-declarations default-debugging-declarations units-used-by-default words-per-flonum
-  parameter-limit eq-inline-operator optimizable-rest-argument-operators
-  membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  target-include-file default-profiling-declarations
-  default-optimization-passes internal-bindings big-fixnum?
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion)
-
-
+(include "compiler-namespace")
 (include "tweaks")
 
 
@@ -77,7 +41,7 @@
      ##sys#standard-input ##sys#standard-output ##sys#standard-error)
     (bound-to-procedure
      ##sys#for-each ##sys#map ##sys#print ##sys#setter
-     ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values ##sys#match-error
+     ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values
      ##sys#start-timer ##sys#stop-timer ##sys#gcd ##sys#lcm ##sys#make-promise ##sys#structure? ##sys#slot 
      ##sys#allocate-vector ##sys#list->vector ##sys#block-ref ##sys#block-set!
      ##sys#list ##sys#cons ##sys#append ##sys#vector ##sys#foreign-char-argument ##sys#foreign-fixnum-argument
@@ -95,17 +59,19 @@
 
 (define default-profiling-declarations
   '((##core#declare
-     '(uses profiler)
-     '(bound-to-procedure
+     (uses profiler)
+     (bound-to-procedure
        ##sys#profile-entry ##sys#profile-exit) ) ) )
 
 (define units-used-by-default '(library eval data-structures ports extras srfi-69)) 
 (define words-per-flonum 4)
 (define parameter-limit 1024)
 (define small-parameter-limit 128)
+(define unlikely-variables '(unquote unquote-splicing))
 
 (define eq-inline-operator "C_eqp")
-(define optimizable-rest-argument-operators '(car cadr caddr cadddr length pair? null? list-ref))
+(define optimizable-rest-argument-operators 
+  '(car cadr caddr cadddr length pair? null? list-ref))
 (define membership-test-operators
   '(("C_i_memq" . "C_eqp") ("C_u_i_memq" . "C_eqp") ("C_i_member" . "C_i_equalp")
     ("C_i_memv" . "C_i_eqvp") ) )
@@ -113,20 +79,25 @@
 (define target-include-file "chicken.h")
 
 (define valid-compiler-options
-  '(-help h help version verbose explicit-use quiet no-trace no-warnings unsafe block
+  '(-help h help version verbose explicit-use 
+	  quiet 			; DEPRECATED
+	  no-trace no-warnings unsafe block 
     check-syntax to-stdout no-usual-integrations case-insensitive no-lambda-info 
-    profile inline keep-shadowed-macros
-    fixnum-arithmetic disable-interrupts optimize-leaf-routines check-imports
-    lambda-lift run-time-macros tag-pointers accumulate-profile
+    profile inline keep-shadowed-macros ignore-repository
+    fixnum-arithmetic disable-interrupts optimize-leaf-routines
+    lambda-lift compile-syntax tag-pointers accumulate-profile
     disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries raw 
-    emit-external-prototypes-first release disable-compiler-macros
-    analyze-only dynamic extension) )
+    emit-external-prototypes-first release local inline-global
+    analyze-only dynamic scrutinize no-argc-checks no-procedure-checks
+    no-bound-checks no-procedure-checks-for-usual-bindings no-compiler-syntax
+    no-parentheses-synonyms no-symbol-escape r5rs-syntax emit-all-import-libraries
+    setup-mode) )
 
 (define valid-compiler-options-with-argument
   '(debug output-file include-path heap-size stack-size unit uses keyword-style require-extension 
-	  inline-limit profile-name disable-warning emit-exports import
-    prelude postlude prologue epilogue nursery extend feature 
-    compress-literals			; DEPRECATED
+	  inline-limit profile-name disable-warning parenthesis-synonyms
+    prelude postlude prologue epilogue nursery extend feature types
+    emit-import-library emit-inline-file static-extension consult-inline-file
     heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path) )
 
 
@@ -153,14 +124,11 @@
     list-ref abs char-ready? peek-char list->string string->list) )
 
 (define default-extended-bindings
-  '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod
+  '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod o
     fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity fp+ fp- fp* fp/ fpmin fpmax fpneg
     fp> fp< fp= fp>= fp<= fxand fxnot fxior fxxor fxshr fxshl bit-set?
     arithmetic-shift void flush-output thread-specific thread-specific-set!
     not-pair? atom? null-list? print print* error cpu-time proper-list? call/cc
-    u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector ; DEPRECATED
-    u32vector->byte-vector s32vector->byte-vector byte-vector-length ; DEPRECATED
-    f32vector->byte-vector f64vector->byte-vector byte-vector-ref byte-vector-set! ; DEPRECATED
     blob-size u8vector->blob/shared s8vector->blob/shared u16vector->blob/shared
     s16vector->blob/shared u32vector->blob/shared s32vector->blob/shared
     f32vector->blob/shared f64vector->blob/shared
@@ -176,7 +144,8 @@
     f32vector-ref f64vector-ref
     u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
     locative-ref locative-set! locative->object locative? global-ref
-    null-pointer? pointer->object flonum? finite?) )
+    null-pointer? pointer->object flonum? finite?
+    printf sprintf format) )
 
 (define internal-bindings
   '(##sys#slot ##sys#setslot ##sys#block-ref ##sys#block-set!
@@ -187,7 +156,8 @@
     ##sys#call-with-values ##sys#fits-in-int? ##sys#fits-in-unsigned-int? ##sys#flonum-in-fixnum-range? 
     ##sys#fudge ##sys#immediate? ##sys#direct-return ##sys#context-switch
     ##sys#make-structure ##sys#apply ##sys#apply-values ##sys#continuation-graft
-    ##sys#bytevector? ##sys#make-vector ##sys#setter
+    ##sys#bytevector? ##sys#make-vector ##sys#setter ##sys#car ##sys#cdr ##sys#pair?
+    ##sys#eq? ##sys#list? ##sys#vector? ##sys#eqv?
     ##sys#foreign-char-argument ##sys#foreign-fixnum-argument ##sys#foreign-flonum-argument
     ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
     ##sys#foreign-string-argument ##sys#foreign-pointer-argument ##sys#void
@@ -195,54 +165,32 @@
     ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values ##sys#poke-double
     ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? ##sys#peek-byte) )
 
-(define side-effecting-standard-bindings
-  '(apply call-with-current-continuation set-car! set-cdr! write-char newline write display
+(define non-foldable-bindings
+  '(vector
+    cons list string make-vector make-string string->symbol values current-input-port current-output-port
+    read-char write-char printf fprintf
+    apply call-with-current-continuation set-car! set-cdr! write-char newline write display
     peek-char char-ready?
     read read-char for-each map string-set! vector-set! string-fill! vector-fill! open-input-file
     open-output-file close-input-port close-output-port call-with-input-port call-with-output-port
-    call-with-values eval) )
-
-(define non-foldable-standard-bindings
-  '(vector cons list string make-vector make-string string->symbol values current-input-port current-output-port
-	   read-char write-char) )
-
-(define foldable-standard-bindings
-  (lset-difference 
-   eq? default-standard-bindings 
-   side-effecting-standard-bindings non-foldable-standard-bindings) )
-
-(define non-foldable-extended-bindings
-  '(##sys#slot ##sys#setslot ##sys#call-with-current-continuation ##sys#fudge flush-output print void
-    u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector ; DEPRECATED
-    f32vector->byte-vector f64vector->byte-vector s32vector->byte-vector ;DEPRECATED
+    call-with-values eval
+    ##sys#slot ##sys#setslot ##sys#call-with-current-continuation ##sys#fudge flush-output print void
     u8vector->blob/shared s8vector->blob/shared u16vector->blob/shared s16vector->blob/shared u32vector->blob/shared
     f32vector->blob/shared f64vector->blob/shared
-    s32vector->blob/shared read-string read-string!
+    s32vector->blob/shared read-string read-string! o
     ##sys#make-structure print* ##sys#make-vector ##sys#apply ##sys#setislot ##sys#block-ref
     ##sys#byte ##sys#setbyte 
-    byte-vector-ref byte-vector-set!	; DEPRECATED
     u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
     f32vector-length f64vector-length ##sys#apply-values ##sys#setter setter
     u8vector-ref s8vector-ref u16vector-ref s16vector-ref u32vector-ref s32vector-ref
     u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
     ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time ##sys#block-set!) )
 
-(define foldable-extended-bindings
-  (lset-difference
-   eq? default-extended-bindings non-foldable-extended-bindings) )
-
-(define standard-bindings-that-never-return-false
-  '(cons list length * - + / current-output-port current-input-port append symbol->string char->integer
-    integer->char vector-length string-length string-ref gcd lcm reverse string->symbol max min
-    quotient remainder modulo floor ceiling truncate round exact->inexact inexact->exact exp log sin
-    cons tan atan expt sqrt asin acos number->string char-upcase char-downcase string-append string
-    string->list list->string vector->list list->vector read-char substring make-string make-vector
-    open-input-file open-output-file vector write-char) )
-
-(define side-effect-free-standard-bindings-that-never-return-false
-  (lset-difference
-   eq? standard-bindings-that-never-return-false
-   side-effecting-standard-bindings) )
+(define foldable-bindings
+  (lset-difference 
+   eq?
+   (lset-union eq? default-standard-bindings default-extended-bindings)
+   non-foldable-bindings) )
 
 
 ;;; Rewriting-definitions for this platform:
@@ -366,25 +314,26 @@
 	     '##core#call '(#t)
 	     (cons* (make-node '##core#proc '("C_quotient" #t) '()) cont callargs) ) ) ) ) )
 
-(rewrite
- 'eqv? 8
- (lambda (db classargs cont callargs)
-   ;; (eqv? <var> <var>) -> (quote #t)
-   ;; (eqv? ...) -> (##core#inline "C_eqp" ...) [one argument is a constant and not a flonum]
-   (and (= (length callargs) 2)
-	(let ([arg1 (first callargs)]
-	      [arg2 (second callargs)] )
-	  (or (and (eq? '##core#variable (node-class arg1))
-		   (eq? '##core#variable (node-class arg2))
-		   (equal? (node-parameters arg1) (node-parameters arg2))
-		   (make-node '##core#call '(#t) (list cont (qnode #t))) )
-	      (and (or (and (eq? 'quote (node-class arg1))
-			    (not (flonum? (first (node-parameters arg1)))) )
-		       (and (eq? 'quote (node-class arg2))
-			    (not (flonum? (first (node-parameters arg2)))) ) )
-		   (make-node
-		    '##core#call '(#t) 
-		    (list cont (make-node '##core#inline '("C_eqp") callargs)) ) ) ) ) ) ) )
+(let ()
+  (define (eqv?-id db classargs cont callargs)
+    ;; (eqv? <var> <var>) -> (quote #t)
+    ;; (eqv? ...) -> (##core#inline "C_eqp" ...) [one argument is a constant and not a flonum]
+    (and (= (length callargs) 2)
+	 (let ([arg1 (first callargs)]
+	       [arg2 (second callargs)] )
+	   (or (and (eq? '##core#variable (node-class arg1))
+		    (eq? '##core#variable (node-class arg2))
+		    (equal? (node-parameters arg1) (node-parameters arg2))
+		    (make-node '##core#call '(#t) (list cont (qnode #t))) )
+	       (and (or (and (eq? 'quote (node-class arg1))
+			     (not (flonum? (first (node-parameters arg1)))) )
+			(and (eq? 'quote (node-class arg2))
+			     (not (flonum? (first (node-parameters arg2)))) ) )
+		    (make-node
+		     '##core#call '(#t) 
+		     (list cont (make-node '##core#inline '("C_eqp") callargs)) ) ) ) ) ) )
+  (rewrite 'eqv? 8 eqv?-id)
+  (rewrite '##sys#eqv? 8 eqv?-id))
 
 (rewrite
  'equal? 8
@@ -431,8 +380,7 @@
 			(= 2 (length callargs))
 			(let ([name (car (node-parameters proc))])
 			  (and (memq name '(values ##sys#values))
-			       (or (get db name 'standard-binding)
-				   (get db name 'extended-binding) )
+			       (intrinsic? name)
 			       (make-node
 				'##core#call '(#t)
 				(list (make-node '##core#proc '("C_apply_values" #t) '())
@@ -474,6 +422,8 @@
 			 [else (return #f)] ) ) ) ) ) ) ) ) ) )
 
   (rewrite-c..r 'car "C_i_car" "C_u_i_car" 0)
+  (rewrite-c..r '##sys#car "C_i_car" "C_u_i_car" 0)
+  (rewrite-c..r '##sys#cdr "C_i_cdr" "C_u_i_cdr" 0)
   (rewrite-c..r 'cadr "C_i_cadr" "C_u_i_cadr" 1)
   (rewrite-c..r 'caddr "C_i_caddr" "C_u_i_caddr" 2)
   (rewrite-c..r 'cadddr "C_i_cadddr" "C_u_i_cadddr" 3)
@@ -566,11 +516,14 @@
 (rewrite 'cdr 2 1 "C_i_cdr" #t #f)
 
 (rewrite 'eq? 1 2 "C_eqp")
+(rewrite '##sys#eq? 1 2 "C_eqp")
 (rewrite 'eqv? 1 2 "C_i_eqvp")
+(rewrite '##sys#eqv? 1 2 "C_i_eqvp")
 
 (rewrite 'list-ref 2 2 "C_u_i_list_ref" #f "C_slot")
 (rewrite 'list-ref 2 2 "C_i_list_ref" #t "C_i_vector_ref")
 (rewrite 'null? 2 1 "C_i_nullp" #t "C_vemptyp")
+(rewrite '##sys#null? 2 1 "C_i_nullp" #t "C_vemptyp")
 (rewrite 'length 2 1 "C_i_length" #t "C_block_size")
 (rewrite 'not 2 1 "C_i_not" #t #f)
 (rewrite 'char? 2 1 "C_charp" #t #f)
@@ -578,19 +531,21 @@
 (rewrite 'locative? 2 1 "C_i_locativep" #t #f)
 (rewrite 'symbol? 2 1 "C_i_symbolp" #t #f)
 (rewrite 'vector? 2 1 "C_i_vectorp" #t #f)
+(rewrite '##sys#vector? 2 1 "C_i_vectorp" #t #f)
 (rewrite 'pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
+(rewrite '##sys#pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
 (rewrite 'procedure? 2 1 "C_i_closurep" #t #f)
 (rewrite 'port? 2 1 "C_i_portp" #t #f)
 (rewrite 'boolean? 2 1 "C_booleanp" #t #f)
 (rewrite 'number? 2 1 "C_i_numberp" #t #f)
 (rewrite 'complex? 2 1 "C_i_numberp" #t #f)
-(rewrite 'rational? 2 1 "C_i_numberp" #t #f)
+(rewrite 'rational? 2 1 "C_i_rationalp" #t #f)
 (rewrite 'real? 2 1 "C_i_numberp" #t #f)
 (rewrite 'integer? 2 1 "C_i_integerp" #t #f)
 (rewrite 'flonum? 2 1 "C_i_flonump" #t #f)
 (rewrite 'fixnum? 2 1 "C_fixnump" #t #f)
 (rewrite 'finite? 2 1 "C_i_finitep" #f #f)
-(rewrite '##sys#pointer? 2 1 "C_pointerp" #t #f)
+(rewrite '##sys#pointer? 2 1 "C_anypointerp" #t #f)
 (rewrite '##sys#generic-structure? 2 1 "C_structurep" #t #f)
 (rewrite 'exact? 2 1 "C_fixnump" #f #f)
 (rewrite 'exact? 2 1 "C_i_exactp" #t #f)
@@ -613,7 +568,7 @@
 (rewrite 'char>=? 2 2 "C_fixnum_greater_or_equal_p" #t #f)
 (rewrite 'char<=? 2 2 "C_fixnum_less_or_equal_p" #t #f)
 (rewrite '##sys#slot 2 2 "C_slot" #t #f)		; consider as safe, the primitive is unsafe anyway.
-(rewrite '##sys#block-ref 2 2 "C_i_block_ref" #t #f) ; must be safe for pattern matcher
+(rewrite '##sys#block-ref 2 2 "C_i_block_ref" #t #f) ;*** must be safe for pattern matcher (anymore?)
 (rewrite '##sys#size 2 1 "C_block_size" #t #f)
 (rewrite 'fxnot 2 1 "C_fixnum_not" #t #f)
 (rewrite 'fx* 2 2 "C_fixnum_times" #t #f)
@@ -704,6 +659,7 @@
 (rewrite 'integer->char 6 "C_make_character" "C_unfix" #t)
 
 (rewrite 'vector-length 2 1 "C_i_vector_length" #t #f)
+(rewrite '##sys#vector-length 2 1 "C_i_vector_length" #t #f)
 (rewrite 'string-length 2 1 "C_i_string_length" #t #f)
 (rewrite 'inexact->exact 2 1 "C_i_inexact_to_exact" #t #f)
 
@@ -748,6 +704,11 @@
 (rewrite 'vector-set! 11 3 '##sys#setslot #f)
 (rewrite 'vector-set! 2 3 "C_i_vector_set" #t #f)
 
+(rewrite '##sys#vector->list 11 1 'vector->list #t)
+(rewrite '##sys#list->vector 11 1 'list->vector #t)
+(rewrite '##sys#>= 11 2 '>= #t)
+(rewrite '##sys#= 11 2 '= #t)
+
 (rewrite 'gcd 12 '##sys#gcd #t 2)
 (rewrite 'lcm 12 '##sys#lcm #t 2)
 (rewrite 'identity 12 #f #t 1)
@@ -905,10 +866,7 @@
 (rewrite '##sys#foreign-unsigned-integer-argument 17 1 "C_i_foreign_unsigned_integer_argumentp")
 (rewrite '##sys#direct-return 17 2 "C_direct_return")
 
-(rewrite 'byte-vector-ref 2 2 "C_subbyte" #f #f) ; DEPRECATED
-(rewrite 'byte-vector-set! 2 3 "C_setbyte" #f #f) ; DEPRECATED
-(rewrite 'byte-vector-length 2 1 "C_block_size" #f #f) ; DEPRECATED
-(rewrite 'blob-size 2 1 "C_block_size" #f #f) ; DEPRECATED
+(rewrite 'blob-size 2 1 "C_block_size" #f #f)
 
 (rewrite 'u8vector-ref 2 2 "C_u_i_u8vector_ref" #f #f)
 (rewrite 's8vector-ref 2 2 "C_u_i_s8vector_ref" #f #f)
@@ -938,15 +896,6 @@
 (rewrite 'atom? 17 1 "C_i_not_pair_p")
 (rewrite 'null-list? 17 1 "C_i_null_list_p" "C_i_nullp")
 
-(rewrite 'u8vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 's8vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 'u16vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 's16vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 'u32vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 's32vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 'f32vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-(rewrite 'f64vector->byte-vector 7 1 "C_slot" 1 #f) ; DEPRECATED
-
 (rewrite 'u8vector->blob/shared 7 1 "C_slot" 1 #f)
 (rewrite 's8vector->blob/shared 7 1 "C_slot" 1 #f)
 (rewrite 'u16vector->blob/shared 7 1 "C_slot" 1 #f)
@@ -1017,7 +966,6 @@
     (cdr . set-cdr!)
     (hash-table-ref . hash-table-set!)
     (block-ref . block-set!)
-    (byte-vector-ref . byte-vector-set!) ; DEPRECATED
     (locative-ref . locative-set!)
     (u8vector-ref . u8vector-set!)
     (s8vector-ref . s8vector-set!)
@@ -1047,8 +995,7 @@
 	(let ((arg (car callargs)))
 	  (and (eq? '##core#variable (node-class arg))
 	       (let ((sym (car (node-parameters arg))))
-		 (and (or (get db sym 'standard-binding)
-			  (get db sym 'extended-binding))
+		 (and (intrinsic? sym)
 		      (and-let* ((a (assq sym setter-map)))
 			(make-node
 			 '##core#call '(#t)
