@@ -1,18 +1,18 @@
 /* chicken.h - General headerfile for compiler generated executables
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 ; conditions are met:
 ;
 ;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
-;     disclaimer. 
+;     disclaimer.
 ;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
-;     disclaimer in the documentation and/or other materials provided with the distribution. 
+;     disclaimer in the documentation and/or other materials provided with the distribution.
 ;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
-;     products derived from this software without specific prior written permission. 
+;     products derived from this software without specific prior written permission.
 ;
 ; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 ; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
@@ -27,10 +27,16 @@
 
 /* Configuration: */
 
+/*
+ * The Watcom (__WATCOMC__), Metroworks (__MWERKS__), and Delorie (__DJGPP__)
+ * compilers are not currently supported but existing references remain,
+ * just in case.
+ */
+
 #ifndef ___CHICKEN
 #define ___CHICKEN
 
-#define C_MAJOR_VERSION       3
+#define C_MAJOR_VERSION       4
 
 /*
  * N.B. This file MUST not rely upon "chicken-config.h"
@@ -39,17 +45,92 @@
 # include "chicken-config.h"
 #endif
 
-#if !defined(__GNUC__) && !defined(__WATCOMC__)
+
+/* Kind of platform */
+
+#ifndef C_SIXTY_FOUR
+# if defined (__alpha__) || defined (__sparc_v9__) || defined (__sparcv9) || defined(__ia64__) || defined(__x86_64__) || defined(__LP64__) || defined(__powerpc64__)
+#   define C_SIXTY_FOUR
+# elif defined(__mips64) && (!defined(__GNUC__) || _MIPS_SZPTR == 64)
+#   define C_SIXTY_FOUR
+# endif
+#endif
+
+#if defined(__APPLE__) && defined(__MACH__)
+# define C_MACOSX
+#endif
+
+#if defined(C_MACOSX) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__) || defined(__OpenBSD__)
+# define C_XXXBSD
+#endif
+
+#if /*defined(__GNUC__) &&*/ (defined(__linux__) || defined(C_XXXBSD))
+# define C_GNU_ENV
+#endif
+
+#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__WATCOMC__) || defined(__MWERKS__) || defined(__DJGPP__)
+# define C_NONUNIX
+#endif
+
+
+/* Headers */
+
+#include <stdio.h>
+#include <stdlib.h>
+#include <stdarg.h>
+#include <ctype.h>
+#include <string.h>
+#include <setjmp.h>
+#include <limits.h>
+#include <time.h>
+
+#if !defined(C_NONUNIX) || defined(__MINGW32__) || defined(__WATCOMC__)
+# include <unistd.h>
+# include <inttypes.h>
+# include <sys/types.h>
+#endif
+
+/* Byteorder in machine word */
+
+#if defined(__MINGW32__)
+# include <sys/param.h>
+#elif defined(__CYGWIN__)
+# include <endian.h>
+#elif defined(__linux__)
+# include <endian.h>
+#elif defined(C_XXXBSD)
+# include <machine/endian.h>
+#elif defined(__hpux__)
+# include <arpa/nameser.h>
+#elif defined(_AIX)
+# include <sys/machine.h>
+#elif defined(__sun__)
+# include <sys/isa_defs.h>
+#elif defined(__svr4__)
+# include <sys/byteorder.h>
+#endif
+
+#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__WATCOMC__)
+# include <malloc.h>
+#endif
+
+#ifdef _MSC_VER
+# include <io.h>
+#endif
+
+/* Much better with stack allocation API */
+
+#if defined(_MSC_VER)
+# if HAVE_ALLOCA_H
+#  define alloca            _alloca
+# endif
+#elif !defined(__GNUC__) && !defined(__WATCOMC__)
 # if HAVE_ALLOCA_H
 #  include <alloca.h>
-# else
-#  ifdef _AIX
+# elif defined(_AIX)
 #   pragma alloca
-#  else
-#   ifndef alloca /* predefined by HP cc +Olibcalls */
-char *alloca ();
-#   endif
-#  endif
+# elif !defined(alloca) /* predefined by HP cc +Olibcalls */
+    char *alloca ();
 # endif
 #elif (defined(__sun__) && defined(__svr4__)) || defined(__sgi__)
 # if HAVE_ALLOCA_H
@@ -57,6 +138,47 @@ char *alloca ();
 # endif
 #endif
 
+
+/* Chicken Core C API */
+
+#if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN
+# define C_BIG_ENDIAN
+#elif defined(BYTE_ORDER) && defined(BIG_ENDIAN) && BYTE_ORDER == BIG_ENDIAN
+# define C_BIG_ENDIAN
+#elif defined(__BIG_ENDIAN__)
+# define C_BIG_ENDIAN
+#elif defined(__sparc__) || defined(__POWERPC__) || defined(__MC68K__) || defined(__mips__)
+# define C_BIG_ENDIAN
+#endif
+
+#if defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN) && __BYTE_ORDER == __LITTLE_ENDIAN
+# define C_LITTLE_ENDIAN
+#elif defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && BYTE_ORDER == LITTLE_ENDIAN
+# define C_LITTLE_ENDIAN
+#elif defined(__LITTLE_ENDIAN__)
+# define C_LITTLE_ENDIAN
+#elif defined (__alpha__) || defined(_M_IX86) || defined(__i386__) || defined(__x86_64__) || defined(__ia64__)
+# define C_LITTLE_ENDIAN
+#endif
+
+/* Make sure some common C identifiers are availble w/ Windows */
+
+#ifdef _MSC_VER
+# define strncasecmp       strnicmp
+# define isatty            _isatty
+typedef __int8             int8_t;
+typedef unsigned __int8    uint8_t;
+typedef __int16            int16_t;
+typedef unsigned  __int16  uint16_t;
+typedef __int32            int32_t;
+typedef unsigned __int32   uint32_t;
+typedef __int64            int64_t;
+typedef unsigned __int64   uint64_t;
+# pragma warning(disable: 4101)
+#endif
+
+/* Could be used by C++ source */
+
 #ifdef __cplusplus
 # define C_extern                  extern "C"
 # define C_BEGIN_C_DECLS           extern "C" {
@@ -66,7 +188,11 @@ char *alloca ();
 # define C_BEGIN_C_DECLS
 # define C_END_C_DECLS
 #endif
- 
+
+
+/* Function declaration modes */
+
+/* Visibility */
 #define C_varextern                C_extern
 #define C_fctimport
 #define C_fctexport
@@ -115,6 +241,50 @@ char *alloca ();
 # endif
 #endif
 
+/* Language specifics: */
+#if defined(__GNUC__) || defined(__INTEL_COMPILER)
+# ifndef __cplusplus
+#  define C_cblock                ({
+#  define C_cblockend             })
+#  define C_noret                 __attribute__ ((noreturn))
+#  define C_noret_decl(name)
+#  define C_aligned               __attribute__ ((aligned))
+# endif
+# ifdef __i386__
+#  define C_regparm               __attribute__ ((regparm(3)))
+# endif
+#elif defined(_MSC_VER)
+# define C_fcall                  __fastcall
+#elif defined(__WATCOMC__)
+# define C_ccall                  __cdecl
+#endif
+
+#ifndef C_cblock
+# define C_cblock                 do{
+# define C_cblockend              }while(0)
+# define C_noret
+# define C_noret_decl(name)
+#endif
+
+#ifndef C_regparm
+# define C_regparm
+#endif
+
+#ifndef C_fcall
+# define C_fcall
+#endif
+
+#ifndef C_ccall
+# define C_ccall
+#endif
+
+#ifndef C_aligned
+# define C_aligned
+#endif
+
+#define C_c_regparm
+
+/* Thread Local Stoarage */
 #ifdef C_ENABLE_TLS
 # if defined(__GNUC__)
 #  define C_TLS                    __thread
@@ -127,6 +297,9 @@ char *alloca ();
 # define C_TLS
 #endif
 
+
+/* Stack growth direction; used to compute stack addresses */
+
 #ifndef C_STACK_GROWS_DOWNWARD
 # define C_STACK_GROWS_DOWNWARD    -1
 #endif
@@ -141,122 +314,22 @@ char *alloca ();
 # endif
 #endif
 
+/* Have a GUI? */
+
 #if defined(C_WINDOWS_GUI)
 # define C_MICROSOFT_WINDOWS
+# include <windows.h>
+# ifndef WINAPI
+#  define WINAPI
+# endif
 #else
 # define C_GENERIC_CONSOLE
 #endif
 
+/* Needed for pre-emptive threading */
+
 #define C_TIMER_INTERRUPTS
 
-#ifdef C_DEFAULT_TARGET_STACK_SIZE
-# define C_resize_stack(n)           C_do_resize_stack(C_DEFAULT_TARGET_STACK_SIZE)
-#else
-# define C_resize_stack(n)           C_do_resize_stack(n)
-#endif
-
-#ifndef C_SIXTY_FOUR
-# if defined (__alpha__) || defined (__sparc_v9__) || defined (__sparcv9) || defined(__ia64__) || defined(__x86_64__) || defined(__LP64__) || defined(__powerpc64__)
-#   define C_SIXTY_FOUR
-# elif defined(__mips64) && (!defined(__GNUC__) || _MIPS_SZPTR == 64)
-#   define C_SIXTY_FOUR
-# endif
-#endif
-
-#if defined(__APPLE__) && defined(__MACH__)
-# define C_MACOSX
-#endif
-
-#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__) || defined(__OpenBSD__)
-# define C_XXXBSD
-#endif
-
-#if defined(C_MACOSX) || defined(__linux__) || defined(C_XXXBSD)
-# define C_GNU_ENV
-#endif
-
-#if defined(_MSC_VER) || defined(__MWERKS__) || defined(__DJGPP__) || defined(__MINGW32__) || defined(__WATCOMC__)
-# define C_NONUNIX
-#endif
-
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <ctype.h>
-#include <string.h>
-#include <setjmp.h>
-#include <limits.h>
-#include <time.h>
-
-#if !defined(C_NONUNIX) || defined(__MINGW32__) || defined(__WATCOMC__)
-# include <unistd.h>
-# include <inttypes.h>
-# include <sys/types.h>
-#endif
-
-#if defined(__MINGW32__)
-# include <sys/param.h>
-#elif defined(__CYGWIN__)
-# include <endian.h>
-#elif defined(__linux__)
-# include <endian.h>
-#elif defined(C_MACOSX) || defined(C_XXXBSD)
-# include <machine/endian.h>
-#elif defined(__hpux__)
-# include <arpa/nameser.h>
-#elif defined(_AIX)
-# include <sys/machine.h>
-#elif defined(__sun__)
-# include <sys/isa_defs.h>
-#elif defined(__svr4__)
-# include <sys/byteorder.h>
-#endif
-
-#if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN
-# define C_BIG_ENDIAN
-#elif defined(BYTE_ORDER) && defined(BIG_ENDIAN) && BYTE_ORDER == BIG_ENDIAN
-# define C_BIG_ENDIAN
-#elif defined(__BIG_ENDIAN__)
-# define C_BIG_ENDIAN
-#elif defined(__sparc__) || defined(__POWERPC__) || defined(__MC68K__) || defined(__mips__)
-# define C_BIG_ENDIAN
-#endif
-
-#if defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN) && __BYTE_ORDER == __LITTLE_ENDIAN
-# define C_LITTLE_ENDIAN
-#elif defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && BYTE_ORDER == LITTLE_ENDIAN
-# define C_LITTLE_ENDIAN
-#elif defined(__LITTLE_ENDIAN__)
-# define C_LITTLE_ENDIAN
-#elif defined (__alpha__) || defined(_M_IX86) || defined(__i386__) || defined(__x86_64__) || defined(__ia64__)
-# define C_LITTLE_ENDIAN
-#endif
-
-#ifdef __MINGW32__
-# include <malloc.h>
-#endif
-
-#ifdef _MSC_VER
-# include <malloc.h>
-# include <io.h>
-# define alloca            _alloca
-# define strncasecmp       strnicmp
-# define isatty            _isatty
-typedef __int8             int8_t;
-typedef unsigned __int8    uint8_t;
-typedef __int16            int16_t;
-typedef unsigned  __int16  uint16_t;
-typedef __int32            int32_t;
-typedef unsigned __int32   uint32_t;
-typedef __int64            int64_t;
-typedef unsigned __int64   uint64_t;
-# pragma warning(disable: 4101)
-#endif
-
-#ifdef __WATCOMC__
-# include <malloc.h>
-#endif
-
 /* For the easy FFI: */
 
 #define ___fixnum           int
@@ -307,7 +380,9 @@ typedef unsigned __int64   uint64_t;
 #define C_FIXNUM_BIT              0x00000001
 #define C_FIXNUM_SHIFT            1
 
+/* Character range is that of a UTF-8 codepoint, not representable range */
 #define C_CHAR_BIT_MASK           0x1fffff
+#define C_CHAR_SHIFT              8
 
 #ifdef C_SIXTY_FOUR
 # define C_MOST_POSITIVE_FIXNUM   0x3fffffffffffffffL
@@ -335,15 +410,16 @@ typedef unsigned __int64   uint64_t;
 # define C_PAIR_TYPE              (0x0300000000000000L)
 # define C_CLOSURE_TYPE           (0x0400000000000000L | C_SPECIALBLOCK_BIT)
 # define C_FLONUM_TYPE            (0x0500000000000000L | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
-# define C_UNUSED_TYPE            (0x0600000000000000L)
+/*       unused                   (0x0600000000000000L ...) */
 # define C_PORT_TYPE              (0x0700000000000000L | C_SPECIALBLOCK_BIT)
 # define C_STRUCTURE_TYPE         (0x0800000000000000L)
 # define C_POINTER_TYPE           (0x0900000000000000L | C_SPECIALBLOCK_BIT)
-# define C_BUCKET_TYPE            (0x0f00000000000000L)
 # define C_LOCATIVE_TYPE          (0x0a00000000000000L | C_SPECIALBLOCK_BIT)
 # define C_TAGGED_POINTER_TYPE    (0x0b00000000000000L | C_SPECIALBLOCK_BIT)
-# define C_SWIG_POINTER_TYPE      (0x0c00000000000000L | C_BYTEBLOCK_BIT)
+# define C_SWIG_POINTER_TYPE      (0x0c00000000000000L | C_SPECIALBLOCK_BIT)
 # define C_LAMBDA_INFO_TYPE       (0x0d00000000000000L | C_BYTEBLOCK_BIT)
+/*       unused                   (0x0e00000000000000L ...) */
+# define C_BUCKET_TYPE            (0x0f00000000000000L)
 #else
 # define C_INT_SIGN_BIT           0x80000000
 # define C_INT_TOP_BIT            0x40000000
@@ -360,21 +436,52 @@ typedef unsigned __int64   uint64_t;
 # define C_PAIR_TYPE              (0x03000000)
 # define C_CLOSURE_TYPE           (0x04000000 | C_SPECIALBLOCK_BIT)
 # ifdef C_DOUBLE_IS_32_BITS
-#  define C_FLONUM_TYPE            (0x05000000 | C_BYTEBLOCK_BIT)
+#  define C_FLONUM_TYPE           (0x05000000 | C_BYTEBLOCK_BIT)
 # else
-#  define C_FLONUM_TYPE            (0x05000000 | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
+#  define C_FLONUM_TYPE           (0x05000000 | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
 # endif
-# define C_UNUSED_TYPE            (0x06000000)
+/*       unused                   (0x06000000 ...) */
 # define C_PORT_TYPE              (0x07000000 | C_SPECIALBLOCK_BIT)
 # define C_STRUCTURE_TYPE         (0x08000000)
 # define C_POINTER_TYPE           (0x09000000 | C_SPECIALBLOCK_BIT)
-# define C_BUCKET_TYPE            (0x0f000000)
 # define C_LOCATIVE_TYPE          (0x0a000000 | C_SPECIALBLOCK_BIT)
 # define C_TAGGED_POINTER_TYPE    (0x0b000000 | C_SPECIALBLOCK_BIT)
-# define C_SWIG_POINTER_TYPE      (0x0c000000 | C_BYTEBLOCK_BIT)
+# define C_SWIG_POINTER_TYPE      (0x0c000000 | C_SPECIALBLOCK_BIT)
 # define C_LAMBDA_INFO_TYPE       (0x0d000000 | C_BYTEBLOCK_BIT)
+/*       unused                   (0x0e000000 ...) */
+# define C_BUCKET_TYPE            (0x0f000000)
 #endif
+#define C_VECTOR_TYPE             0x00000000
+#define C_BYTEVECTOR_TYPE         (C_VECTOR_TYPE | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
 
+#define C_SIZEOF_LIST(n)          ((n) * 3 + 1)
+#define C_SIZEOF_PAIR             3
+#define C_SIZEOF_STRING(n)        (C_bytestowords(n) + 2)
+#define C_SIZEOF_SYMBOL           4
+#define C_SIZEOF_INTERNED_SYMBOL(n) (C_SIZEOF_SYMBOL + C_SIZEOF_BUCKET + C_SIZEOF_STRING(n))
+#ifdef C_DOUBLE_IS_32_BITS
+# define C_SIZEOF_FLONUM          2
+#else
+# define C_SIZEOF_FLONUM          4
+#endif
+#define C_SIZEOF_POINTER          2
+#define C_SIZEOF_TAGGED_POINTER   3
+#define C_SIZEOF_SWIG_POINTER     3
+#define C_SIZEOF_VECTOR(n)        ((n) + 1)
+#define C_SIZEOF_BUCKET           3
+#define C_SIZEOF_LOCATIVE         5
+#define C_SIZEOF_PORT             16
+
+/* Fixed size types have pre-computed header tags */
+#define C_PAIR_TAG                (C_PAIR_TYPE | (C_SIZEOF_PAIR - 1))
+#define C_POINTER_TAG             (C_POINTER_TYPE | (C_SIZEOF_POINTER - 1))
+#define C_LOCATIVE_TAG            (C_LOCATIVE_TYPE | (C_SIZEOF_LOCATIVE - 1))
+#define C_TAGGED_POINTER_TAG      (C_TAGGED_POINTER_TYPE | (C_SIZEOF_TAGGED_POINTER - 1))
+#define C_SWIG_POINTER_TAG        (C_SWIG_POINTER_TYPE | (C_wordstobytes(C_SIZEOF_SWIG_POINTER - 1)))
+#define C_SYMBOL_TAG              (C_SYMBOL_TYPE | (C_SIZEOF_SYMBOL - 1))
+#define C_FLONUM_TAG              (C_FLONUM_TYPE | sizeof(double))
+
+/* Locative subtypes */
 #define C_SLOT_LOCATIVE           0
 #define C_CHAR_LOCATIVE           1
 #define C_U8_LOCATIVE             2
@@ -386,35 +493,6 @@ typedef unsigned __int64   uint64_t;
 #define C_F32_LOCATIVE            8
 #define C_F64_LOCATIVE            9
 
-#define C_VECTOR_TYPE             0x00000000
-#define C_BYTEVECTOR_TYPE         (C_VECTOR_TYPE | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
-
-#define C_SIZEOF_LIST(n)          ((n) * 3 + 1)
-#define C_SIZEOF_PAIR             3
-#define C_SIZEOF_STRING(n)        (C_bytestowords(n) + 2)
-#define C_SIZEOF_SYMBOL          4
-#define C_SIZEOF_INTERNED_SYMBOL(n) (C_SIZEOF_SYMBOL + C_SIZEOF_BUCKET + C_SIZEOF_STRING(n))
-#ifdef C_DOUBLE_IS_32_BITS
-# define C_SIZEOF_FLONUM           2
-#else
-# define C_SIZEOF_FLONUM           4
-#endif
-#define C_SIZEOF_POINTER          2
-#define C_SIZEOF_TAGGED_POINTER   3
-#define C_SIZEOF_SWIG_POINTER     3
-#define C_SIZEOF_VECTOR(n)        ((n) + 1)
-#define C_SIZEOF_BUCKET           3
-#define C_SIZEOF_LOCATIVE         5
-#define C_SIZEOF_PORT             16
-
-#define C_PAIR_TAG                (C_PAIR_TYPE | (C_SIZEOF_PAIR - 1))
-#define C_POINTER_TAG             (C_POINTER_TYPE | (C_SIZEOF_POINTER - 1))
-#define C_LOCATIVE_TAG            (C_LOCATIVE_TYPE | (C_SIZEOF_LOCATIVE - 1))
-#define C_TAGGED_POINTER_TAG      (C_TAGGED_POINTER_TYPE | (C_SIZEOF_TAGGED_POINTER - 1))
-#define C_SWIG_POINTER_TAG        (C_SWIG_POINTER_TYPE | (C_wordstobytes(C_SIZEOF_SWIG_POINTER - 1)))
-#define C_SYMBOL_TAG              (C_SYMBOL_TYPE | (C_SIZEOF_SYMBOL - 1))
-#define C_FLONUM_TAG             (C_FLONUM_TYPE | sizeof(double))
-
 #ifdef C_SIXTY_FOUR
 # define C_word                   long
 # define C_u32                    uint32_t
@@ -450,6 +528,7 @@ typedef unsigned __int64   uint64_t;
 #define C_OUT_OF_RANGE_ERROR                          8
 #define C_NOT_A_CLOSURE_ERROR                         9
 #define C_CONTINUATION_CANT_RECEIVE_VALUES_ERROR      10
+#define C_BAD_ARGUMENT_TYPE_CYCLIC_LIST_ERROR         11
 #define C_TOO_DEEP_RECURSION_ERROR                    12
 #define C_CANT_REPRESENT_INEXACT_ERROR                13
 #define C_NOT_A_PROPER_LIST_ERROR                     14
@@ -477,57 +556,103 @@ typedef unsigned __int64   uint64_t;
 #define C_BAD_ARGUMENT_TYPE_NO_CLOSURE_ERROR          36
 
 
-#define CHICKEN_gc_root_ref(root)      (((C_GC_ROOT *)(root))->value)
-#define CHICKEN_gc_root_set(root, x)   C_mutate(&((C_GC_ROOT *)(root))->value, (x))
+/* Platform information */
+#if defined(C_BIG_ENDIAN)
+# define C_MACHINE_BYTE_ORDER "big-endian"
+#elif defined(C_LITTLE_ENDIAN)
+# define C_MACHINE_BYTE_ORDER "little-endian"
+#endif
 
-#define CHICKEN_global_ref(root)       C_u_i_car(((C_GC_ROOT *)(root))->value)
-#define CHICKEN_global_set(root, x)    C_mutate(&C_u_i_car(((C_GC_ROOT *)(root))->value), (x))
+#if defined(__alpha__)
+# define C_MACHINE_TYPE "alpha"
+#elif defined(__mips__)
+# define C_MACHINE_TYPE "mips"
+#elif defined(__hppa__)
+# define C_MACHINE_TYPE "hppa"
+#elif defined(__sparc_v9__) || defined(__sparcv9)
+# define C_MACHINE_TYPE "ultrasparc"
+#elif defined(__sparc__)
+# define C_MACHINE_TYPE "sparc"
+#elif defined(__powerpc64__)
+# define C_MACHINE_TYPE "ppc64"
+#elif defined(__ppc__) || defined(__powerpc__)
+# define C_MACHINE_TYPE "ppc"
+#elif defined(_M_IX86) || defined(__i386__)
+# define C_MACHINE_TYPE "x86"
+#elif defined(__ia64__)
+# define C_MACHINE_TYPE "ia64"
+#elif defined(__x86_64__)
+# define C_MACHINE_TYPE "x86-64"
+#elif defined(__arm__)
+# define C_MACHINE_TYPE "arm"
+#else
+# define C_MACHINE_TYPE "unknown"
+#endif
 
-#define CHICKEN_default_toplevel       ((void *)C_default_stub_toplevel)
+#if defined(__CYGWIN__) || defined(__MINGW32__) || defined(_WIN32) || defined(__WINNT__)
+# define C_SOFTWARE_TYPE "windows"
+#elif defined(__unix__) || defined(C_XXXBSD)
+# define C_SOFTWARE_TYPE "unix"
+#elif defined(ECOS)
+# define C_SOFTWARE_TYPE "ecos"
+#else
+# define C_SOFTWARE_TYPE "unknown"
+#endif
 
-
-/* Language specifics: */
-#if defined(__GNUC__) || defined(__INTEL_COMPILER)
-# ifndef __cplusplus
-#  define C_cblock                ({
-#  define C_cblockend             })
-#  define C_noret                 __attribute__ ((noreturn))
-#  define C_noret_decl(name)
-#  define C_aligned               __attribute__ ((aligned))
-# endif
-# ifdef __i386__
-#  define C_regparm               __attribute__ ((regparm(3)))
-# endif
+#if defined(__CYGWIN__)
+# define C_BUILD_PLATFORM "cygwin"
 #elif defined(_MSC_VER)
-# define C_fcall                  __fastcall
+# define C_BUILD_PLATFORM "msvc"
+#elif defined(__SUNPRO_C)
+# define C_BUILD_PLATFORM "sun"
+#elif defined(__MINGW32__)
+# define C_BUILD_PLATFORM "mingw32"
+#elif defined(__GNUC__)
+# define C_BUILD_PLATFORM "gnu"
+#elif defined(__MWERKS__)
+# define C_BUILD_PLATFORM "metrowerks"
+#elif defined(__INTEL_COMPILER)
+# define C_BUILD_PLATFORM "intel"
 #elif defined(__WATCOMC__)
-# define C_ccall                  __cdecl
+# define C_BUILD_PLATFORM "watcom"
+#else
+# define C_BUILD_PLATFORM "unknown"
 #endif
 
-#ifndef C_cblock
-# define C_cblock                 do{
-# define C_cblockend              }while(0)
-# define C_noret
-# define C_noret_decl(name)
+#if defined(_MSC_VER)
+# if defined(_DLL)
+#   define C_RUNTIME_VERSION "dynamic"
+# else
+#   define C_RUNTIME_VERSION "static"
+# endif
+#else
+# define C_RUNTIME_VERSION "unknown"
 #endif
 
-#ifndef C_regparm
-# define C_regparm
+#if defined(__linux__)
+# define C_SOFTWARE_VERSION "linux"
+#elif defined(__FreeBSD__)
+# define C_SOFTWARE_VERSION "freebsd"
+#elif defined(__NetBSD__)
+# define C_SOFTWARE_VERSION "netbsd"
+#elif defined(__OpenBSD__)
+# define C_SOFTWARE_VERSION "openbsd"
+#elif defined(C_MACOSX)
+# define C_SOFTWARE_VERSION "macosx"
+#elif defined(__hpux__)
+# define C_SOFTWARE_VERSION "hpux"
+#elif defined(__DragonFly__)
+# define C_SOFTWARE_VERSION "dragonfly"
+#elif defined(__sun__)
+# if defined(__svr4__)
+#   define C_SOFTWARE_VERSION "solaris"
+# else
+#   define C_SOFTWARE_VERSION "sunos"
+# endif
+#else
+# define C_SOFTWARE_VERSION "unknown"
 #endif
 
-#ifndef C_fcall
-# define C_fcall
-#endif
-
-#ifndef C_ccall
-# define C_ccall
-#endif
-
-#ifndef C_aligned
-# define C_aligned
-#endif
-
-#define C_c_regparm
 
 /* Types: */
 
@@ -549,6 +674,7 @@ typedef struct C_gc_root_struct
 {
   C_word value;
   struct C_gc_root_struct *next, *prev;
+  int finalizable;
 } C_GC_ROOT;
 
 typedef struct C_ptable_entry_struct
@@ -560,7 +686,7 @@ typedef struct C_ptable_entry_struct
 #ifdef __x86_64__
 # define C_AMD64_ABI_WEIRDNESS      , ...
 #else
-# define C_AMD64_ABI_WEIRDNESS      
+# define C_AMD64_ABI_WEIRDNESS
 #endif
 
 /* C_WORD_p<P>_<B>: List of ((2 ** P) * B) 'C_word' parameters */
@@ -629,6 +755,14 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 
 /* Macros: */
 
+#define CHICKEN_gc_root_ref(root)      (((C_GC_ROOT *)(root))->value)
+#define CHICKEN_gc_root_set(root, x)   C_mutate(&((C_GC_ROOT *)(root))->value, (x))
+
+#define CHICKEN_global_ref(root)       C_u_i_car(((C_GC_ROOT *)(root))->value)
+#define CHICKEN_global_set(root, x)    C_mutate(&C_u_i_car(((C_GC_ROOT *)(root))->value), (x))
+
+#define CHICKEN_default_toplevel       ((void *)C_default_stub_toplevel)
+
 #define C_align4(n)                (((n) + 3) & ~3)
 #define C_align8(n)                (((n) + 7) & ~7)
 #define C_align16(n)               (((n) + 15) & ~15)
@@ -698,6 +832,11 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 # define C_fputs                    fputs
 # define C_fputc                    fputc
 # define C_putchar                  putchar
+# if (defined getc_unlocked || _POSIX_C_SOURCE >= 199506L)
+#  define C_getc                    getc_unlocked
+# else
+#  define C_getc                    getc
+# endif
 # define C_fgetc                    fgetc
 # define C_fgets                    fgets
 # define C_ungetc                   ungetc
@@ -726,7 +865,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #endif
 
 #define C_return(x)                return(x)
-
+#define C_resize_stack(n)          C_do_resize_stack(n)
 #define C_memcpy_slots(t, f, n)    C_memcpy((t), (f), (n) * sizeof(C_word))
 #define C_block_header(x)          (((C_SCHEME_BLOCK *)(x))->header)
 #define C_header_bits(x)           (C_block_header(x) & C_HEADER_BITS_MASK)
@@ -750,8 +889,8 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_demand_2(n)              (((C_word *)C_fromspace_top + (n)) < (C_word *)C_fromspace_limit)
 #define C_fix(n)                   (((C_word)(n) << C_FIXNUM_SHIFT) | C_FIXNUM_BIT)
 #define C_unfix(x)                 ((x) >> C_FIXNUM_SHIFT)
-#define C_make_character(c)        ((((c) & C_CHAR_BIT_MASK) << 8) | C_CHARACTER_BITS)
-#define C_character_code(x)        (((x) >> 8) & C_CHAR_BIT_MASK)
+#define C_make_character(c)        ((((c) & C_CHAR_BIT_MASK) << C_CHAR_SHIFT) | C_CHARACTER_BITS)
+#define C_character_code(x)        (((x) >> C_CHAR_SHIFT) & C_CHAR_BIT_MASK)
 #define C_flonum_magnitude(x)      (*((double *)(((C_SCHEME_BLOCK *)(x))->data)))
 #define C_c_string(x)              ((C_char *)(((C_SCHEME_BLOCK *)(x))->data))
 #define C_c_pointer(x)             ((void *)(x))
@@ -805,6 +944,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 
 #define C_zero_length_p(x)        C_mk_bool(C_header_size(x) == 0)
 #define C_boundp(x)               C_mk_bool(((C_SCHEME_BLOCK *)(x))->data[ 0 ] != C_SCHEME_UNBOUND)
+#define C_unboundvaluep(x)        C_mk_bool((x) == C_SCHEME_UNBOUND)
 #define C_blockp(x)               C_mk_bool(!C_immediatep(x))
 #define C_forwardedp(x)           C_mk_bool((C_block_header(x) & C_GC_FORWARDING_BIT) != 0)
 #define C_immp(x)                 C_mk_bool(C_immediatep(x))
@@ -900,7 +1040,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 						  C_unfix(end1) - C_unfix(start1) ), C_SCHEME_UNDEFINED)
 #define C_words(n)                      C_fix(C_bytestowords(C_unfix(n)))
 #define C_bytes(n)                      C_fix(C_wordstobytes(C_unfix(n)))
-#define C_random_fixnum(n)              C_fix(rand() % C_unfix(n))
+#define C_random_fixnum(n)              C_fix((int)(((double)rand())/(RAND_MAX + 1.0) * C_unfix(n)))
 #define C_randomize(n)                  (srand(C_unfix(n)), C_SCHEME_UNDEFINED)
 #define C_block_size(x)                 C_fix(C_header_size(x))
 #define C_pointer_address(x)            ((C_byte *)C_u_i_car(x))
@@ -913,6 +1053,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_null_pointerp(x)              C_mk_bool((void *)C_u_i_car(x) == NULL)
 #define C_update_pointer(p, ptr)        (C_set_block_item(ptr, 0, C_num_to_unsigned_int(p)), C_SCHEME_UNDEFINED)
 #define C_copy_pointer(from, to)        (C_set_block_item(to, 0, C_u_i_car(from)), C_SCHEME_UNDEFINED)
+#define C_pointer_to_object(ptr)        ((C_word*)C_block_item(ptr, 0))
 
 #define C_direct_return(dk, x)          (C_kontinue(dk, x), C_SCHEME_UNDEFINED)
 
@@ -938,11 +1079,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_poke_pointer_or_null(b, i, x) (C_set_block_item(b, C_unfix(i), (C_word)C_data_pointer_or_null(x)), C_SCHEME_UNDEFINED)
 #define C_qfree(ptr)                    (C_free(C_c_pointer_nn(ptr)), C_SCHEME_UNDEFINED)
 
-#if defined(__MWERKS__) && !defined(__INTEL__)
-# define C_tty_portp(p)                 C_SCHEME_FALSE
-#else
-# define C_tty_portp(p)                 C_mk_bool(isatty(fileno(C_port_file(p))))
-#endif
+#define C_tty_portp(p)                 C_mk_bool(isatty(fileno(C_port_file(p))))
 
 #define C_emit_eval_trace_info(x, y, z) C_emit_trace_info2("<eval>", x, y, z)
 #define C_emit_syntax_trace_info(x, y, z) C_emit_trace_info2("<syntax>", x, y, z)
@@ -989,14 +1126,14 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_i_nullp(x)                    C_mk_bool((x) == C_SCHEME_END_OF_LIST)
 #define C_i_structurep(x, s)            C_mk_bool(!C_immediatep(x) && C_header_bits(x) == C_STRUCTURE_TYPE && C_block_item(x, 0) == (s))
 
-#define C_u_i_char_alphabeticp(x)       C_mk_bool(C_isalpha(C_character_code(x)))
-#define C_u_i_char_numericp(x)          C_mk_bool(C_isdigit(C_character_code(x)))
-#define C_u_i_char_whitespacep(x)       C_mk_bool(C_isspace(C_character_code(x)))
-#define C_u_i_char_upper_casep(x)       C_mk_bool(C_isupper(C_character_code(x)))
-#define C_u_i_char_lower_casep(x)       C_mk_bool(C_islower(C_character_code(x)))
+#define C_u_i_char_alphabeticp(x)       C_mk_bool(C_character_code(x) < 0x100 && C_isalpha(C_character_code(x)))
+#define C_u_i_char_numericp(x)          C_mk_bool(C_character_code(x) < 0x100 && C_isdigit(C_character_code(x)))
+#define C_u_i_char_whitespacep(x)       C_mk_bool(C_character_code(x) < 0x100 && C_isspace(C_character_code(x)))
+#define C_u_i_char_upper_casep(x)       C_mk_bool(C_character_code(x) < 0x100 && C_isupper(C_character_code(x)))
+#define C_u_i_char_lower_casep(x)       C_mk_bool(C_character_code(x) < 0x100 && C_islower(C_character_code(x)))
 
-#define C_u_i_char_upcase(x)            C_make_character(C_toupper(C_character_code(x)))
-#define C_u_i_char_downcase(x)          C_make_character(C_tolower(C_character_code(x)))
+#define C_u_i_char_upcase(x)            (C_character_code(x) < 0x100 ? C_make_character(C_toupper(C_character_code(x))) : (x))
+#define C_u_i_char_downcase(x)          (C_character_code(x) < 0x100 ? C_make_character(C_tolower(C_character_code(x))) : (x))
 
 #define C_i_list_ref(lst, i)            C_i_car(C_i_list_tail(lst, i))
 #define C_u_i_list_ref(lst, i)          C_u_i_car(C_i_list_tail(lst, i))
@@ -1115,7 +1252,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 /* Variables: */
 
 C_varextern C_TLS time_t C_startup_time_seconds;
-C_varextern C_TLS C_word 
+C_varextern C_TLS C_word
   *C_temporary_stack,
   *C_temporary_stack_bottom,
   *C_stack_limit;
@@ -1144,10 +1281,10 @@ C_varextern C_TLS int
   C_max_pending_finalizers,
   C_trace_buffer_size,
   C_main_argc;
-C_varextern C_TLS C_uword 
+C_varextern C_TLS C_uword
   C_heap_growth,
   C_heap_shrinkage;
-C_varextern C_TLS char 
+C_varextern C_TLS char
   **C_main_argv,
   *C_dlerror;
 C_varextern C_TLS C_uword C_maximal_heap_size;
@@ -1165,6 +1302,8 @@ C_fctexport int CHICKEN_initialize(int heap, int stack, int symbols, void *tople
 C_fctexport C_word CHICKEN_run(void *toplevel);
 C_fctexport C_word CHICKEN_continue(C_word k);
 C_fctexport void *CHICKEN_new_gc_root();
+C_fctexport void *CHICKEN_new_finalizable_gc_root();
+C_fctexport void *CHICKEN_new_gc_root_2(int finalizable);
 C_fctexport void CHICKEN_delete_gc_root(void *root);
 C_fctexport void *CHICKEN_global_lookup(char *name);
 C_fctexport int CHICKEN_is_running();
@@ -1262,6 +1401,8 @@ C_fctexport void C_fcall C_rereclaim2(C_uword size, int double_plus) C_regparm;
 C_fctexport C_word C_fcall C_retrieve(C_word sym) C_regparm;
 C_fctexport C_word C_fcall C_retrieve2(C_word val, char *name) C_regparm;
 C_fctexport void *C_fcall C_retrieve_proc(C_word closure) C_regparm;
+C_fctexport void *C_fcall C_retrieve_symbol_proc(C_word sym) C_regparm;
+C_fctexport void *C_fcall C_retrieve2_symbol_proc(C_word val, char *name) C_regparm;
 C_fctexport C_word C_fcall C_permanentp(C_word x) C_regparm;
 C_fctexport int C_in_stackp(C_word x) C_regparm;
 C_fctexport int C_fcall C_in_heapp(C_word x) C_regparm;
@@ -1390,6 +1531,7 @@ C_fctexport C_word C_fcall C_i_closurep(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_portp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_stringp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_numberp(C_word x) C_regparm;
+C_fctexport C_word C_fcall C_i_rationalp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_integerp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_flonump(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_finitep(C_word x) C_regparm;
@@ -1444,6 +1586,7 @@ C_fctexport C_word C_fcall C_u_i_memq(C_word x, C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_memv(C_word x, C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_member(C_word x, C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_length(C_word lst) C_regparm;
+C_fctexport C_word C_fcall C_u_i_length(C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_inexact_to_exact(C_word n) C_regparm;
 C_fctexport C_word C_fcall C_i_check_closure_2(C_word x, C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_i_check_exact_2(C_word x, C_word loc) C_regparm;
@@ -1471,7 +1614,7 @@ C_fctexport C_word C_fcall C_i_null_list_p(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_string_null_p(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_string_to_pbytevector(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_null_pointerp(C_word x) C_regparm;
-C_fctexport C_word C_fcall C_i_fixnum_arithmetic_shift(C_word n, C_word c) C_regparm; 
+C_fctexport C_word C_fcall C_i_fixnum_arithmetic_shift(C_word n, C_word c) C_regparm;
 C_fctexport C_word C_fcall C_i_locative_set(C_word loc, C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_locative_to_object(C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_a_i_make_locative(C_word **a, int c, C_word type, C_word object, C_word index, C_word weak) C_regparm;
