@@ -1,7 +1,7 @@
-;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Hen -*-
+;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Scheme -*-
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -27,53 +27,7 @@
 
 (declare
   (block)
-  (uses data-structures ports srfi-1 srfi-13 utils files extras ))
-
-#>
-#ifndef C_TARGET_CC
-# define C_TARGET_CC  C_INSTALL_CC
-#endif
-
-#ifndef C_TARGET_CXX
-# define C_TARGET_CXX  C_INSTALL_CXX
-#endif
-
-#ifndef C_TARGET_CFLAGS
-# define C_TARGET_CFLAGS  C_INSTALL_CFLAGS
-#endif
-
-#ifndef C_TARGET_LDFLAGS
-# define C_TARGET_LDFLAGS  C_INSTALL_LDFLAGS
-#endif
-
-#ifndef C_TARGET_BIN_HOME
-# define C_TARGET_BIN_HOME  C_INSTALL_BIN_HOME
-#endif
-
-#ifndef C_TARGET_LIB_HOME
-# define C_TARGET_LIB_HOME  C_INSTALL_LIB_HOME
-#endif
-
-#ifndef C_TARGET_STATIC_LIB_HOME
-# define C_TARGET_STATIC_LIB_HOME  C_INSTALL_STATIC_LIB_HOME
-#endif
-
-#ifndef C_TARGET_INCLUDE_HOME
-# define C_TARGET_INCLUDE_HOME  C_INSTALL_INCLUDE_HOME
-#endif
-
-#ifndef C_TARGET_SHARE_HOME
-# define C_TARGET_SHARE_HOME  C_INSTALL_SHARE_HOME
-#endif
-
-#ifndef C_TARGET_RUN_LIB_HOME
-# define C_TARGET_RUN_LIB_HOME    C_TARGET_LIB_HOME
-#endif
-
-#ifndef C_CHICKEN_PROGRAM
-# define C_CHICKEN_PROGRAM     "chicken"
-#endif
-<#
+  (uses data-structures ports srfi-1 srfi-13 utils files extras))
 
 (define-foreign-variable INSTALL_BIN_HOME c-string "C_INSTALL_BIN_HOME")
 (define-foreign-variable INSTALL_CC c-string "C_INSTALL_CC")
@@ -99,6 +53,7 @@
 (define-foreign-variable TARGET_STATIC_LIB_HOME c-string "C_TARGET_STATIC_LIB_HOME")
 (define-foreign-variable TARGET_RUN_LIB_HOME c-string "C_TARGET_RUN_LIB_HOME")
 (define-foreign-variable CHICKEN_PROGRAM c-string "C_CHICKEN_PROGRAM")
+(define-foreign-variable WINDOWS_SHELL bool "C_WINDOWS_SHELL")
 
 
 ;;; Parameters:
@@ -113,7 +68,7 @@
   (fprintf (current-error-port) "csc: ~?~%" msg args)
   (exit 64) )
 
-(define chicken-prefix (getenv "CHICKEN_PREFIX"))
+(define chicken-prefix (get-environment-variable "CHICKEN_PREFIX"))
 (define arguments (command-line-arguments))
 (define host-mode (member "-host" arguments))
 (define cross-chicken (##sys#fudge 39))
@@ -124,9 +79,7 @@
       default) )
 
 (define (quotewrap str)
-  (if (string-any char-whitespace? str)
-      (string-append "\"" str "\"") 
-      str) )
+  (qs (normalize-pathname str)))
 
 (define home
   (quotewrap 
@@ -152,6 +105,7 @@
 (define shared-library-extension ##sys#load-dynamic-extension)
 (define default-translation-optimization-options '())
 (define pic-options (if (or mingw msvc) '("-DPIC") '("-fPIC" "-DPIC")))
+(define windows-shell WINDOWS_SHELL)
 
 (define default-library (string-append
                          (if msvc "libchicken-static." "libchicken.")
@@ -160,10 +114,7 @@
                                 (if msvc "libuchicken-static." "libuchicken.")
                                 library-extension))
 
-(define cleanup-filename
-  (if (not mingw)
-      (lambda (s) (quotewrap s)) ; allow filenames w/ whitespace
-      (lambda (s) s)))
+(define cleanup-filename quotewrap)
 
 (define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
 (define best-compilation-optimization-options default-compilation-optimization-options)
@@ -173,22 +124,26 @@
 (define-constant simple-options
   '(-explicit-use -no-trace -no-warnings -no-usual-integrations -optimize-leaf-routines -unsafe
     -block -disable-interrupts -fixnum-arithmetic -to-stdout -profile -raw -accumulate-profile
-    -check-syntax -case-insensitive -benchmark-mode -shared -run-time-macros -no-lambda-info
-    -lambda-lift -dynamic -disable-stack-overflow-checks -emit-debug-info -check-imports
-    -emit-external-prototypes-first -inline -extension -release -static-extensions
-    -analyze-only -keep-shadowed-macros -disable-compiler-macros) )
+    -check-syntax -case-insensitive -benchmark-mode -shared -compile-syntax -no-lambda-info
+    -lambda-lift -dynamic -disable-stack-overflow-checks -local
+    -emit-external-prototypes-first -inline -release -scrutinize
+    -analyze-only -keep-shadowed-macros -inline-global -ignore-repository
+    -no-symbol-escape -no-parentheses-synonyms -r5rs-syntax
+    -no-argc-checks -no-bound-checks -no-procedure-checks -no-compiler-syntax
+    -emit-all-import-libraries -setup-mode
+    -no-procedure-checks-for-usual-bindings))
 
 (define-constant complex-options
   '(-debug -output-file -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
     -optimize-level -include-path -database-size -extend -prelude -postlude -prologue -epilogue 
-    -inline-limit -profile-name -disable-warning -import -require-static-extension
-    -feature -debug-level -heap-growth -heap-shrinkage -heap-initial-size -emit-exports
-    -compress-literals) )		; DEPRECATED
+    -inline-limit -profile-name -disable-warning -emit-inline-file -types
+    -feature -debug-level -heap-growth -heap-shrinkage -heap-initial-size -consult-inline-file
+    -emit-import-library -static-extension))
 
 (define-constant shortcuts
   '((-h "-help")
     (-s "-shared")
-    (|-E| "-extension")
+    (-S "-scrutinize")
     (|-P| "-check-syntax")
     (|-V| "-version")
     (|-Ob| "-benchmark-mode")
@@ -198,13 +153,15 @@
     (|-K| "-keyword-style")
     (|-X| "-extend")
     (|-N| "-no-usual-integrations")
-    (|-G| "-check-imports")
+    (|-J| "-emit-all-import-libraries")
     (-x "-explicit-use")
     (-u "-unsafe")
+    (-j "-emit-import-library")
+    (-n "-emit-inline-file")
     (-b "-block") ) )
 
 (define short-options
-  (string->list "PHhsfiENxubvwAOeWkctgG") )
+  (string->list "PHhsfiENxubvwAOeWkctgS") )
 
 
 ;;; Variables:
@@ -257,7 +214,7 @@
 (define library-files default-library-files)
 (define shared-library-files default-shared-library-files)
 
-(define translate-options '("-quiet"))
+(define translate-options '())
 
 (define include-dir
   (let ((id (prefix "" "include" 
@@ -303,7 +260,7 @@
 (define shared #f)
 (define static #f)
 (define static-libs #f)
-(define static-extensions #f)
+(define static-extensions '())
 (define required-extensions '())
 (define gui #f)
 
@@ -311,151 +268,189 @@
 ;;; Display usage information:
 
 (define (usage)
-  (display
-"Usage: csc FILENAME | OPTION ...
+  (display #<<EOF
+Usage: csc FILENAME | OPTION ...
 
-  `csc' is a driver program for the CHICKEN compiler. Any Scheme, C or object
-  files and all libraries given on the command line are translated, compiled or
-  linked as needed.
+  `csc' is a driver program for the CHICKEN compiler. Files given on the
+  command line are translated, compiled or linked as needed.
+
+  FILENAME is a Scheme source file name with optional extension or a
+  C/C++/Objective-C source, object or library file name with extension. OPTION
+  may be one of the following:
 
   General options:
 
-    -h  -help                   display this text and exit
-    -v                          show intermediate compilation stages
-    -v2  -verbose               display information about translation progress
-    -v3                         display information about all compilation stages
-    -V  -version                display Scheme compiler version and exit
-    -release                    display release number and exit
+    -h  -help                      display this text and exit
+    -v                             show intermediate compilation stages
+    -vv  -verbose                  display information about translation
+                                    progress
+    -vvv                           display information about all compilation
+                                    stages
+    -V  -version                   display Scheme compiler version and exit
+    -release                       display release number and exit
 
   File and pathname options:
 
-    -o -output-file FILENAME    specifies target executable name
-    -I -include-path PATHNAME   specifies alternative path for included files
-    -to-stdout                  write compiler to stdout (implies -t)
-    -s -shared -dynamic         generate dynamically loadable shared object file
+    -o -output-file FILENAME       specifies target executable name
+    -I -include-path PATHNAME      specifies alternative path for included
+                                    files
+    -to-stdout                     write compiler to stdout (implies -t)
+    -s -shared -dynamic            generate dynamically loadable shared object
+                                    file
 
   Language options:
 
-    -D  -DSYMBOL  -feature SYMBOL 
-                                register feature identifier
-    -c++                        Compile via a C++ source file (.cpp) 
-    -objc                       Compile via Objective-C source file (.m)
+    -D  -DSYMBOL  -feature SYMBOL  register feature identifier
+    -c++                           compile via a C++ source file (.cpp) 
+    -objc                          compile via Objective-C source file (.m)
 
   Syntax related options:
 
-    -i -case-insensitive        don't preserve case of read symbols    
-    -K -keyword-style STYLE     allow alternative keyword syntax (prefix, suffix or none)
-    -run-time-macros            macros are made available at run-time
+    -i -case-insensitive           don't preserve case of read symbols    
+    -k  -keyword-style STYLE       enable alternative keyword-syntax
+                                    (prefix, suffix or none)
+        -no-parentheses-synonyms   disables list delimiter synonyms
+        -no-symbol-escape          disables support for escaped symbols
+        -r5rs-syntax               disables the Chicken extensions to
+                                    R5RS syntax
+    -compile-syntax                macros are made available at run-time
+    -j -emit-import-library MODULE write compile-time module information into
+                                    separate file
+    -J -emit-all-import-libraries  emit import-libraries for all defined modules
+    -no-compiler-syntax            disable expansion of compiler-macros
 
   Translation options:
 
-    -x  -explicit-use           do not use units `library' and `eval' by default
-    -P  -check-syntax           stop compilation after macro-expansion
-    -A  -analyze-only           stop compilation after first analysis pass
+    -x  -explicit-use              do not use units `library' and `eval' by
+                                    default
+    -P  -check-syntax              stop compilation after macro-expansion
+    -A  -analyze-only              stop compilation after first analysis pass
 
   Debugging options:
 
-    -w  -no-warnings            disable warnings
-    -disable-warning CLASS      disable specific class of warnings
+    -w  -no-warnings               disable warnings
+    -disable-warning CLASS         disable specific class of warnings
     -d0 -d1 -d2 -debug-level NUMBER
-                                set level of available debugging information
-    -no-trace                   disable rudimentary debugging information
-    -profile                    executable emits profiling information 
-    -accumulate-profile         executable emits profiling information in append mode
-    -profile-name FILENAME      name of the generated profile information file
-    -emit-debug-info            emit additional debug-information
-    -emit-exports FILENAME      write exported toplevel variables to FILENAME
-    -G  -check-imports          look for undefined toplevel variables
-    -import FILENAME            read externally exported symbols from FILENAME
+                                   set level of available debugging information
+    -no-trace                      disable rudimentary debugging information
+    -profile                       executable emits profiling information 
+    -accumulate-profile            executable emits profiling information in
+                                    append mode
+    -profile-name FILENAME         name of the generated profile information
+                                    file
+    -S  -scrutinize                perform local flow analysis
+    -types FILENAME                load additional type database
 
   Optimization options:
 
-    -O -O1 -O2 -O3 -optimize-level NUMBER
-			        enable certain sets of optimization options
-    -optimize-leaf-routines     enable leaf routine optimization
-    -N  -no-usual-integrations  standard procedures may be redefined
-    -u  -unsafe                 disable safety checks
-    -b  -block                  enable block-compilation
-    -disable-interrupts         disable interrupts in compiled code
-    -f  -fixnum-arithmetic      assume all numbers are fixnums
-    -Ob  -benchmark-mode        equivalent to '-block -optimize-level 3 
-                                 -debug-level 0 -fixnum-arithmetic -lambda-lift 
-                                 -disable-interrupts'
-    -lambda-lift                perform lambda-lifting
-    -unsafe-libraries           link with unsafe runtime system
-    -disable-stack-overflow-checks  disables detection of stack-overflows
-    -inline                     enable inlining
-    -inline-limit               set inlining threshold
-    -disable-compiler-macros    disable expansion of compiler macros
+    -O -O1 -O2 -O3 -O4 -optimize-level NUMBER
+                                   enable certain sets of optimization options
+    -optimize-leaf-routines        enable leaf routine optimization
+    -N  -no-usual-integrations     standard procedures may be redefined
+    -u  -unsafe                    disable safety checks
+    -local                         assume globals are only modified in current
+                                    file
+    -b  -block                     enable block-compilation
+    -disable-interrupts            disable interrupts in compiled code
+    -f  -fixnum-arithmetic         assume all numbers are fixnums
+    -Ob  -benchmark-mode           equivalent to '-block -optimize-level 4
+                                    -debug-level 0 -fixnum-arithmetic
+                                    -lambda-lift -inline -disable-interrupts'
+    -lambda-lift                   perform lambda-lifting
+    -unsafe-libraries              link with unsafe runtime system
+    -disable-stack-overflow-checks disables detection of stack-overflows
+    -inline                        enable inlining
+    -inline-limit                  set inlining threshold
+    -inline-global                 enable cross-module inlining
+    -n -emit-inline-file FILENAME  generate file with globally inlinable
+                                    procedures (implies -inline -local)
+    -consult-inline-file FILENAME  explicitly load inline file
+    -no-argc-checks                disable argument count checks
+    -no-bound-checks               disable bound variable checks
+    -no-procedure-checks           disable procedure call checks
+    -no-procedure-checks-for-usual-bindings
+                                   disable procedure call checks only for usual
+                                    bindings
 
   Configuration options:
 
-    -unit NAME                  compile file as a library unit
-    -uses NAME                  declare library unit as used.
-    -heap-size NUMBER           specifies heap-size of compiled executable
-    -heap-initial-size NUMBER   specifies heap-size at startup time
-    -heap-growth PERCENTAGE     specifies growth-rate of expanding heap
-    -heap-shrinkage PERCENTAGE  specifies shrink-rate of contracting heap
+    -unit NAME                     compile file as a library unit
+    -uses NAME                     declare library unit as used.
+    -heap-size NUMBER              specifies heap-size of compiled executable
+    -heap-initial-size NUMBER      specifies heap-size at startup time
+    -heap-growth PERCENTAGE        specifies growth-rate of expanding heap
+    -heap-shrinkage PERCENTAGE     specifies shrink-rate of contracting heap
     -nursery NUMBER  -stack-size NUMBER
-		                specifies nursery size of compiled executable
-    -X -extend FILENAME         load file before compilation commences
-    -prelude EXPRESSION         add expression to beginning of source file
-    -postlude EXPRESSION        add expression to end of source file
-    -prologue FILENAME          include file before main source file
-    -epilogue FILENAME          include file after main source file
+                                   specifies nursery size of compiled
+                                   executable
+    -X -extend FILENAME            load file before compilation commences
+    -prelude EXPRESSION            add expression to beginning of source file
+    -postlude EXPRESSION           add expression to end of source file
+    -prologue FILENAME             include file before main source file
+    -epilogue FILENAME             include file after main source file
 
-    -e  -embedded               compile as embedded (don't generate `main()')
-    -W  -windows                compile as Windows GUI application (MSVC only)
-    -R  -require-extension NAME require extension in compiled code
-    -E  -extension              compile as extension (dynamic or static)
-    -dll -library               compile multiple units into a dynamic library
+    -e  -embedded                  compile as embedded
+                                    (don't generate `main()')
+    -W  -windows                   compile as Windows GUI application
+    -R  -require-extension NAME    require extension and import in compiled
+                                    code
+    -dll -library                  compile multiple units into a dynamic
+                                    library
 
   Options to other passes:
 
-    -C OPTION                   pass option to C compiler
-    -L OPTION                   pass option to linker
-    -I<DIR>                     pass \"-I<DIR>\" to C compiler (add include path)
-    -L<DIR>                     pass \"-L<DIR>\" to linker (add library path)
-    -k                          keep intermediate files
-    -c                          stop after compilation to object files
-    -t                          stop after translation to C
-    -cc COMPILER                select other C compiler than the default one
-    -cxx COMPILER               select other C++ compiler than the default one
-    -ld COMPILER                select other linker than the default one
-    -lLIBNAME                   link with given library (`libLIBNAME' on UNIX,
-                                 `LIBNAME.lib' on Windows)                                
-    -static-libs                link with static CHICKEN libraries
-    -static                     generate completely statically linked executable
-    -static-extensions          link with static extensions (if available)
-    -F<DIR>                     pass \"-F<DIR>\" to C compiler (add framework 
-                                 header path on Mac OS X)
-    -framework NAME             passed to linker on Mac OS X
-    -rpath PATHNAME             add directory to runtime library search path
-    -Wl,...                     pass linker options
-    -strip                      strip resulting binary
+    -C OPTION                      pass option to C compiler
+    -L OPTION                      pass option to linker
+    -I<DIR>                        pass \"-I<DIR>\" to C compiler
+                                    (add include path)
+    -L<DIR>                        pass \"-L<DIR>\" to linker
+                                    (add library path)
+    -k                             keep intermediate files
+    -c                             stop after compilation to object files
+    -t                             stop after translation to C
+    -cc COMPILER                   select other C compiler than the default
+    -cxx COMPILER                  select other C++ compiler than the default
+    -ld COMPILER                   select other linker than the default 
+    -lLIBNAME                      link with given library
+                                    (`libLIBNAME' on UNIX,
+                                     `LIBNAME.lib' on Windows)
+    -static-libs                   link with static CHICKEN libraries
+    -static                        generate completely statically linked
+                                    executable
+    -static-extension NAME         link extension NAME statically
+                                    (if available)
+    -F<DIR>                        pass \"-F<DIR>\" to C compiler
+                                    (add framework header path on Mac OS X)
+    -framework NAME                passed to linker on Mac OS X
+    -rpath PATHNAME                add directory to runtime library search path
+    -Wl,...                        pass linker options
+    -strip                         strip resulting binary
 
   Inquiry options:
 
-    -home                       show home-directory (where support files go)
-    -cflags                     show required C-compiler flags and exit
-    -ldflags                    show required linker flags and exit
-    -libs                       show required libraries and exit
-    -cc-name                    show name of default C compiler used
-    -cxx-name                   show name of default C++ compiler used
-    -ld-name                    show name of default linker used
-    -dry-run                    just show commands executed, don't run them 
-                                 (implies `-v')
+    -home                          show home-directory (where support files go)
+    -cflags                        show required C-compiler flags and exit
+    -ldflags                       show required linker flags and exit
+    -libs                          show required libraries and exit
+    -cc-name                       show name of default C compiler used
+    -cxx-name                      show name of default C++ compiler used
+    -ld-name                       show name of default linker used
+    -dry-run                       just show commands executed, don't run them
+                                    (implies `-v')
 
   Obscure options:
 
-    -debug MODES                display debugging output for the given modes
-    -compiler PATHNAME          use other compiler than default `chicken'
-    -disable-c-syntax-checks    disable syntax checks of C code fragments
-    -raw                        do not generate implicit init- and exit code			       
-    -emit-external-prototypes-first  emit protoypes for callbacks before foreign
-                                 declarations
-    -keep-shadowed-macros       do not remove shadowed macro
-    -host                       compile for host when configured for cross-compiling
+    -debug MODES                   display debugging output for the given modes
+    -compiler PATHNAME             use other compiler than default `chicken'
+    -disable-c-syntax-checks       disable syntax checks of C code fragments
+    -raw                           do not generate implicit init- and exit code
+    -emit-external-prototypes-first
+                                   emit prototypes for callbacks before foreign
+                                    declarations
+    -ignore-repository             do not refer to repository for extensions
+    -keep-shadowed-macros          do not remove shadowed macro
+    -host                          compile for host when configured for
+                                    cross-compiling
 
   Options can be collapsed if unambiguous, so
 
@@ -465,9 +460,10 @@
 
     -v -k -fixnum-arithmetic -optimize
 
-  The contents of the environment variable CSC_OPTIONS are implicitly
-  passed to every invocation of `csc'.
-"
+  The contents of the environment variable CSC_OPTIONS are implicitly passed to
+  every invocation of `csc'.
+
+EOF
 ) )
 
 
@@ -479,7 +475,7 @@
     (set! translate-options (append translate-options os)) )
 
   (define (check o r . n)
-    (unless (>= (length r) (:optional n 1))
+    (unless (>= (length r) (optional n 1))
       (quit "not enough arguments to option `~A'" o) ) )
 
   (define (shared-build lib)
@@ -504,7 +500,8 @@
 	     (when show-libs (print* (linker-libraries #t) #\space))
 	     (newline)
 	     (exit) )
-	   #;(when (null? scheme-files)
+	   #; ;UNUSED
+	   (when (null? scheme-files)
 	     (set! scheme-files c-files)
 	     (set! c-files '()) )
 	   (cond [(null? scheme-files)
@@ -531,7 +528,10 @@
 	       (when (member target-filename scheme-files)
 		 (printf "Warning: output file will overwrite source file `~A' - renaming source to `~A.old'~%"
 			 target-filename target-filename)
-		 (unless (zero? ($system (sprintf "mv ~A ~A.old" target-filename target-filename)))
+		 (unless (zero? ($system (sprintf "~A ~A ~A" 
+						  (if *windows-shell* "move" "mv")
+						  (quotewrap target-filename)
+						  (quotewrap (string-append target-filename ".old")))))
 		   (exit last-exit-code) ) )
 	       (run-linking)) ) ]
 	  [else
@@ -559,8 +559,6 @@
 	       [(-static-libs) 
 		(set! translate-options (cons* "-feature" "chicken-compile-static" translate-options))
 		(set! static-libs #t) ]
-	       [(-static-extensions)
-		(set! static-extensions #t) ]
 	       [(-cflags)
 		(set! inquiry-only #t) 
 		(set! show-cflags #t) ]
@@ -575,19 +573,25 @@
 		(set! inquiry-only #t)
 		(set! show-libs #t) ]
 	       [(-v)
-		(set! verbose #t) ]
-	       [(-v2 -verbose)
+		(when (and (number? verbose) (not msvc))
+		  (set! compile-options (cons* "-v" "-Q" compile-options))
+		  (set! link-options (cons (if msvc "-VERBOSE" "-v") link-options)) )
+		(cond (verbose
+		       (t-options "-verbose") 
+		       (set! verbose 2)) 
+		      (else (set! verbose #t))) ]
+	       [(-v2 -verbose)		; DEPRECATED
 		(set! verbose #t)
 		(t-options "-verbose") ]
-	       [(-w -no-warnings)
-		(set! compile-options (cons "-w" compile-options))
-		(t-options "-no-warnings") ]
-	       [(-v3)
+	       [(-v3)			; DEPRECATED
 		(set! verbose #t)
 		(t-options "-verbose")
                 (if (not msvc)
                     (set! compile-options (cons* "-v" "-Q" compile-options)))
 		(set! link-options (cons (if msvc "-VERBOSE" "-v") link-options)) ]
+	       [(-w -no-warnings)
+		(set! compile-options (cons "-w" compile-options))
+		(t-options "-no-warnings") ]
 	       [(|-A| -analyze-only)
 		(set! translate-only #t)
 		(t-options "-analyze-only") ]
@@ -605,6 +609,11 @@
 		(set! required-extensions (append required-extensions (list (car rest))))
 		(t-options "-require-extension" (car rest))
 		(set! rest (cdr rest)) ]
+	       [(-static-extension)
+		(check s rest)
+		(set! static-extensions (append static-extensions (list (car rest))))
+		(t-options "-static-extension" (car rest))
+		(set! rest (cdr rest)) ]
 	       [(-windows |-W|)
 		(set! gui #t)
 		(cond
@@ -630,6 +639,7 @@
 	       [(|-O| |-O1|) (set! rest (cons* "-optimize-level" "1" rest))]
 	       [(|-O2|) (set! rest (cons* "-optimize-level" "2" rest))]
 	       [(|-O3|) (set! rest (cons* "-optimize-level" "3" rest))]
+	       [(|-O4|) (set! rest (cons* "-optimize-level" "4" rest))]
 	       [(-d0) (set! rest (cons* "-debug-level" "0" rest))]
 	       [(-d1) (set! rest (cons* "-debug-level" "1" rest))]
 	       [(-d2) (set! rest (cons* "-debug-level" "2" rest))]
@@ -754,49 +764,26 @@
 (define (run-translation)
   (for-each
    (lambda (f)
-     (let ([cscf (pathname-replace-extension f "csc")])
-       (when (and (file-exists? cscf)
-		  (let ([x (with-input-from-file cscf read-line)])
-		    (or (eof-object? x) (string=? "#%eof" x)) ) )
-	 ($delete-file cscf) )
-       (let ([fc (pathname-replace-extension
-		  (if (= 1 (length scheme-files))
-		      target-filename
-		      f)
-		  (cond (cpp-mode "cpp")
-			(objc-mode "m")
-			(else "c") ) ) ] )
-	 (unless (zero?
-		  ($system 
-		   (string-intersperse 
-		    (cons* translator (cleanup-filename f) 
-			   (append 
-			    (if to-stdout 
-				'("-to-stdout")
-				`("-output-file" ,(cleanup-filename fc)) )
-			    (if (or static static-libs static-extensions)
-				(map (lambda (e) (conc "-uses " e)) required-extensions)
-				'() )
-			    (map quote-option (append translate-options translation-optimization-options)) ) )
-		    " ") ) )
-	   (exit last-exit-code) )
-	 (set! c-files (append (list fc) c-files))
-	 (set! generated-c-files (append (list fc) generated-c-files))
-	 (when (file-exists? cscf)
-	   (with-input-from-file cscf
-	     (lambda ()
-	       (read-line)
-	       (for-each
-		(match-lambda
- 		  [('post-process commands ...)
- 		   (for-each $system commands) ]
- 		  [('c-options opts ...)
- 		   (set! compile-options (append compile-options opts)) ]
- 		  [('link-options opts ...)
- 		   (set! link-options (append link-options opts)) ]
-		  [x (error "invalid entry in csc control file" x)] )
-		(read-file) ) ) )
-	   ($delete-file cscf) ) ) ) )
+     (let ([fc (pathname-replace-extension
+		(if (= 1 (length scheme-files))
+		    target-filename
+		    f)
+		(cond (cpp-mode "cpp")
+		      (objc-mode "m")
+		      (else "c") ) ) ] )
+       (unless (zero?
+		($system 
+		 (string-intersperse 
+		  (cons* translator (cleanup-filename f) 
+			 (append 
+			  (if to-stdout 
+			      '("-to-stdout")
+			      `("-output-file" ,(cleanup-filename fc)) )
+			  (map quote-option (append translate-options translation-optimization-options)) ) )
+		  " ") ) )
+	 (exit last-exit-code) )
+       (set! c-files (append (list fc) c-files))
+       (set! generated-c-files (append (list fc) generated-c-files))))
    scheme-files)
   (unless keep-files (for-each $delete-file generated-scheme-files)) )
 
@@ -869,8 +856,8 @@
 
 (define (static-extension-info)
   (let ((rpath (repository-path)))
-    (if (or static static-libs static-extensions)
-	(let loop ((exts required-extensions) (libs '()) (opts '()))
+    (if (and rpath (pair? static-extensions))
+	(let loop ((exts static-extensions) (libs '()) (opts '()))
 	  (if (null? exts)
 	      (values (reverse libs) (reverse opts))
 	      (let ((info (extension-information (car exts))))
@@ -922,9 +909,9 @@
 	s) ) )
 
 (define (quote-option x)
-  (if (any (lambda (c)
-	     (or (char-whitespace? c) (memq c +hairy-chars+)) )
-	   (string->list x) )
+  (if (string-any (lambda (c)
+		    (or (char-whitespace? c) (memq c +hairy-chars+)) )
+		  x)
       (cleanup x)
       x) )
 
@@ -932,15 +919,16 @@
 
 (define ($system str)
   (when verbose (print str))
-  (set! last-exit-code
-    (if dry-run 
-	0
-	(if (zero? (system str))
-	    0
-	    1)))
-  (unless (zero? last-exit-code)
-    (printf "*** Shell command terminated with exit status ~S: ~A~%" last-exit-code str) )
-  last-exit-code)
+  (let ((str (if windows-shell
+		 (string-append "\"" str "\"")
+		 str)))
+    (set! last-exit-code
+      (if dry-run 
+	  0
+	  (system str)))
+    (unless (zero? last-exit-code)
+      (printf "\nError: shell command terminated with non-zero exit status ~S: ~A~%" last-exit-code str) )
+    last-exit-code))
 
 (define ($delete-file str)
   (when verbose 
@@ -950,4 +938,4 @@
 
 ;;; Run it:
 
-(run (append (string-split (or (getenv "CSC_OPTIONS") "")) arguments))
+(run (append (string-split (or (get-environment-variable "CSC_OPTIONS") "")) arguments))
