@@ -1,7 +1,7 @@
 ;;;; csi.scm - Interpreter stub for CHICKEN
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -25,13 +25,12 @@
 ; POSSIBILITY OF SUCH DAMAGE.
 
 
-(declare (uses match srfi-69 ports))
-
 (declare
+  (uses chicken-syntax srfi-69 ports extras)
   (usual-integrations)
   (disable-interrupts)
   (disable-warning var)
-  (run-time-macros)
+  (run-time-macros)			;*** later: compile-syntax
   (foreign-declare #<<EOF
 #if (defined(_MSC_VER) && defined(_WIN32)) || defined(HAVE_DIRECT_H)
 # include <direct.h>
@@ -41,7 +40,6 @@
 EOF
 ) )
 
-(include "chicken-more-macros")
 (include "banner")
 
 (private csi 
@@ -57,7 +55,7 @@ EOF
   (always-bound
     ##sys#windows-platform)
   (hide parse-option-string bytevector-data member* canonicalize-args do-trace do-untrace
-	traced-procedures describer-table dirseparator?
+	traced-procedures describer-table dirseparator? resolve-var
 	findall trace-indent command-table do-break do-unbreak broken-procedures) )
 
 
@@ -72,36 +70,73 @@ EOF
 ;;; Print all sorts of information:
 
 (define (print-usage)
-  (display
-"Usage: csi {FILENAME | OPTION}
+  (display #<<EOF
+usage: csi [FILENAME | OPTION ...]
 
-  where OPTION may be one of the following:
+  `csi' is the CHICKEN interpreter.
+  
+  FILENAME is a Scheme source file name with optional extension. OPTION may be
+  one of the following:
 
-    -h  -help  --help           display this text and exit
-    -v  -version                display version and exit
-        -release                print release number and exit
-    -i  -case-insensitive       enable case-insensitive reading
-    -e  -eval EXPRESSION        evaluate given expression
-    -p  -print EXPRESSION       evaluate and print result(s)
+    -h  -help  --help             display this text and exit
+    -v  -version                  display version and exit
+        -release                  print release number and exit
+    -i  -case-insensitive         enable case-insensitive reading
+    -e  -eval EXPRESSION          evaluate given expression
+    -p  -print EXPRESSION         evaluate and print result(s)
     -P  -pretty-print EXPRESSION  evaluate and print result(s) prettily
-    -D  -feature SYMBOL         register feature identifier
-    -q  -quiet                  do not print banner
-    -n  -no-init                do not load initialization file `")
-  (display init-file)
-  (display 
-"'
-    -b  -batch                  terminate after command-line processing
-    -w  -no-warnings            disable all warnings
-    -k  -keyword-style STYLE    enable alternative keyword-syntax (none, prefix or suffix)
-    -s  -script PATHNAME        use interpreter for shell scripts
-        -ss PATHNAME            shell script with `main' procedure
-    -R  -require-extension NAME require extension before executing code
-    -I  -include-path PATHNAME  add PATHNAME to include path
-    --                          ignore all following options
+    -D  -feature SYMBOL           register feature identifier
+    -q  -quiet                    do not print banner
 
-") )
+EOF
+)
+  (display #<#EOF
+    -n  -no-init                  do not load initialization file #{#\`} #{init-file} #{#\'}
+
+EOF
+)
+  (display  #<<EOF
+    -b  -batch                    terminate after command-line processing
+    -w  -no-warnings              disable all warnings
+    -k  -keyword-style STYLE      enable alternative keyword-syntax
+                                   (prefix, suffix or none)
+        -no-parentheses-synonyms  disables list delimiter synonyms
+        -no-symbol-escape         disables support for escaped symbols
+        -r5rs-syntax              disables the Chicken extensions to
+                                   R5RS syntax
+    -s  -script PATHNAME          use interpreter for shell scripts
+        -ss PATHNAME              shell script with `main' procedure
+        -sx PATHNAME              same as `-s', but print each expression
+                                   as it is evaluated
+        -setup-mode               prefer the current directory when locating extensions
+    -R  -require-extension NAME   require extension and import before
+                                   executing code
+    -I  -include-path PATHNAME    add PATHNAME to include path
+    --                            ignore all following options
+
+EOF
+) )
 
 (define (print-banner)
+  (newline)
+  #; ;UNUSED
+  (when (and (tty-input?) (##sys#fudge 11))
+    (let* ((t (string-copy +product+))
+	   (len (string-length t))
+	   (c (make-string len #\x08)))
+      (do ((i (sub1 (* 2 len)) (sub1 i)))
+	  ((zero? i))
+	(let* ((p (abs (- i len)))
+	       (o (string-ref t p)))
+	  (string-set! t p #\@)
+	  (print* t)
+	  (string-set! t p o)
+	  (let ((t0 (+ (current-milliseconds) 20)))
+	    (let loop ()		; crude, but doesn't need srfi-18
+	      (when (< (current-milliseconds) t0)
+		(loop))))
+	  (print* c) ) ) ) )
+  (print +product+)
   (print +banner+ (chicken-version #t) "\n") )
 
 
@@ -155,7 +190,7 @@ EOF
 		((proc (##core#inline "C_subchar" str1 i)) i)
 		(else (loop (fx+ i 1))) ) ) ) )
     (lambda (name)
-      (let ([path (getenv "PATH")])
+      (let ([path (get-environment-variable "PATH")])
 	(and (> (##sys#size name) 0)
 	     (cond [(dirseparator? (string-ref name 0)) (addext name)]
 		   [(string-index dirseparator? name)
@@ -208,29 +243,25 @@ EOF
     (lambda ()
       (when (tty-input?) (old)) ) ) )
 
-(define command-table (make-hash-table eq?))
+(define command-table (make-vector 37 '()))
 
-(define toplevel-command
-  (let ((hash-table-set! hash-table-set!))
-    (lambda (name proc #!optional help)
-      (##sys#check-symbol name 'toplevel-command)
-      (when help (##sys#check-string help 'toplevel-command))
-      (hash-table-set! command-table name (cons proc help)) ) ) )
+(define (toplevel-command name proc #!optional help)
+  (##sys#check-symbol name 'toplevel-command)
+  (when help (##sys#check-string help 'toplevel-command))
+  (##sys#hash-table-set! command-table name (cons proc help)) )
 
 (set! ##sys#repl-eval-hook
-  ;; `macroexpand' is intentionally not shadowed (psyntax redefines it).
   (let ((eval eval)
 	(load-noisily load-noisily)
 	(read read)
 	(singlestep singlestep)
-	(hash-table-ref hash-table-ref)
-	(hash-table-walk hash-table-walk)
 	(read-line read-line)
 	(length length)
 	(display display)
 	(write write)
 	(string-split string-split)
 	(printf printf)
+	(expand expand)
 	(pretty-print pretty-print)
 	(integer? integer?)
 	(values values) )
@@ -240,7 +271,7 @@ EOF
 	    ((and (pair? form)
 		  (eq? 'unquote (##sys#slot form 0)) )
 	     (let ((cmd (cadr form)))
-	       (cond ((and (symbol? cmd) (hash-table-ref/default command-table cmd #f)) =>
+	       (cond ((and (symbol? cmd) (##sys#hash-table-ref command-table cmd)) =>
 		      (lambda (p)
 			((car p))
 			(##sys#void) ) )
@@ -248,7 +279,7 @@ EOF
 		      (case cmd
 			((x)
 			 (let ([x (read)])
-			   (pretty-print (macroexpand x))
+			   (pretty-print (##sys#strip-syntax (expand x)))
 			   (##sys#void) ) )
 			((p)
 			 (let* ([x (read)]
@@ -284,10 +315,10 @@ EOF
 			   (receive rs (time (eval x))
 			     (history-add rs)
 			     (apply values rs) ) ) )
-			((tr) (do-trace (map string->symbol (string-split (read-line)))))
-			((utr) (do-untrace (map string->symbol (string-split (read-line)))))
-			((br) (do-break (map string->symbol (string-split (read-line)))))
-			((ubr) (do-unbreak (map string->symbol (string-split (read-line)))))
+			((tr) (do-trace (map resolve-var (string-split (read-line)))))
+			((utr) (do-untrace (map resolve-var (string-split (read-line)))))
+			((br) (do-break (map resolve-var (string-split (read-line)))))
+			((ubr) (do-unbreak (map resolve-var (string-split (read-line)))))
 			((uba) (do-unbreak-all))
 			((breakall) 
 			 (set! ##sys#break-in-thread #f) ) 
@@ -343,14 +374,14 @@ EOF
  ,step EXPR        Execute EXPR in single-stepping mode
  ,exn              Describe last exception
  ,t EXP            Evaluate form and print elapsed time
- ,x EXP            Pretty print macroexpanded expression EXP\n")
-			 (hash-table-walk
-			  command-table
+ ,x EXP            Pretty print expanded expression EXP\n")
+			 (##sys#hash-table-for-each
 			  (lambda (k v) 
 			    (let ((help (cdr v)))
 			      (if help
 				  (print #\space help)
-				  (print " ," k) ) ) ) )
+				  (print " ," k) ) ) )
+			  command-table)
 			 (##sys#void) )
 			(else
 			 (printf "Undefined toplevel command ~s - enter `,?' for help~%" form) 
@@ -360,6 +391,9 @@ EOF
 	       (history-add rs)
 	       (apply values rs) ) ) ) ) ) )
 
+(define (resolve-var str)
+  (##sys#strip-syntax (string->symbol str) (##sys#current-environment) #t))
+
 
 ;;; Tracing:
 
@@ -411,14 +445,14 @@ EOF
 	(for-each (lambda (a) (print (car a))) traced-procedures) 
 	(for-each
 	 (lambda (s)
-	   (let ((s (macroexpand s)))
+	   (let ((s (expand s)))
 	     (cond ((assq s traced-procedures)
 		    (##sys#warn "procedure already traced" s) )
 		   ((assq s broken-procedures)
 		    (##sys#warn "procedure already has breakpoint") )
 		   (else
 		    (let ((old (##sys#slot s 0)))
-		      (cond ((not (procedure? old)) (##sys#error "can not trace non-procedure" s))
+		      (cond ((not (procedure? old)) (##sys#error "cannot trace non-procedure" s))
 			    (else
 			     (set! traced-procedures (cons (cons s old) traced-procedures))
 			     (##sys#setslot
@@ -435,7 +469,7 @@ EOF
   (lambda (names)
     (for-each
      (lambda (s)
-       (let* ((s (macroexpand s))
+       (let* ((s (expand s))
 	      (p (assq s traced-procedures)) )
 	 (cond ((not p) (##sys#warn "procedure not traced" s))
 	       (else
@@ -449,14 +483,14 @@ EOF
 	(for-each (lambda (b) (print (car a))) broken-procedures) 
 	(for-each
 	 (lambda (s)
-	   (let* ((s (macroexpand s))
+	   (let* ((s (expand s))
 		  (a (assq s traced-procedures)))
 	     (when a
 	       (##sys#warn "un-tracing procedure" s)
 	       (##sys#setslot s 0 (cdr a))
 	       (set! traced-procedures (del a traced-procedures eq?)) )
 	     (let ((old (##sys#slot s 0)))
-	       (cond ((not (procedure? old)) (##sys#error "can not set breakpoint on non-procedure" s))
+	       (cond ((not (procedure? old)) (##sys#error "cannot set breakpoint on non-procedure" s))
 		     (else
 		      (set! broken-procedures (cons (cons s old) broken-procedures))
 		      (##sys#setslot
@@ -470,7 +504,7 @@ EOF
   (lambda (names)
     (for-each
      (lambda (s)
-       (let* ((s (macroexpand s))
+       (let* ((s (expand s))
 	      (p (assq s broken-procedures)) )
 	 (cond ((not p) (##sys#warn "procedure has no breakpoint" s))
 	       (else
@@ -534,7 +568,7 @@ EOF
                    Include path:    \t~A~%~
                    Symbol-table load:\t~S~%  ~
                      Avg bucket length:\t~S~%  ~
-                     Total symbols:\t~S~%~
+                     Total symbol count:\t~S~%~
                    Memory:\theap size is ~S bytes~A with ~S bytes currently in use~%~  
                      nursery size is ~S bytes, stack grows ~A~%"
 		    (machine-type)
@@ -571,7 +605,7 @@ EOF
 
 (define-constant max-describe-lines 40)
 
-(define describer-table (make-hash-table eq?))
+(define describer-table (make-vector 37 '()))
 
 (define describe
   (let ([sprintf sprintf]
@@ -579,137 +613,144 @@ EOF
 	[fprintf fprintf]
 	[length length]
 	[list-ref list-ref]
-	[string-ref string-ref]
-	(hash-table-ref/default hash-table-ref/default)
-	[hash-table-walk hash-table-walk] )
-    (lambda (x . port)
-      (let ([out (:optional port ##sys#standard-output)])
+	[string-ref string-ref])
+    (lambda (x #!optional (out ##sys#standard-output))
+      (define (descseq name plen pref start)
+	(let ((len (fx- (plen x) start)))
+	  (when name (fprintf out "~A of length ~S~%" name len))
+	  (let loop1 ((i 0))
+	    (cond ((fx>= i len))
+		  ((fx>= i max-describe-lines)
+		   (fprintf out "~% (~A elements not displayed)~%" (fx- len i)) )
+		  (else
+		   (let ((v (pref x (fx+ start i))))
+		     (let loop2 ((n 1) (j (fx+ i (fx+ start 1))))
+		       (cond ((fx>= j len)
+			      (fprintf out " ~S: ~S" i v)
+			      (if (fx> n 1)
+				  (fprintf out "\t(followed by ~A identical instance~a)~% ...~%" 
+					   (fx- n 1)
+					   (if (eq? n 2) "" "s"))
+				  (newline out) )
+			      (loop1 (fx+ i n)) )
+			     ((eq? v (pref x j)) (loop2 (fx+ n 1) (fx+ j 1)))
+			     (else (loop2 n len)) ) ) ) ) ) ) ) )
+      (when (##sys#permanent? x)
+	(fprintf out "statically allocated (0x~X) " (##sys#block-address x)) )
+      (cond [(char? x)
+	     (let ([code (char->integer x)])
+	       (fprintf out "character ~S, code: ~S, #x~X, #o~O~%" x code code code) ) ]
+	    [(eq? x #t) (fprintf out "boolean true~%")]
+	    [(eq? x #f) (fprintf out "boolean false~%")]
+	    [(null? x) (fprintf out "empty list~%")]
+	    [(eof-object? x) (fprintf out "end-of-file object~%")]
+	    [(eq? (##sys#void) x) (fprintf out "unspecified object~%")]
+	    [(fixnum? x)
+	     (fprintf out "exact integer ~S, #x~X, #o~O, #b~B" x x x x)
+	     (let ([code (integer->char x)])
+	       (when (fx< x #x10000) (fprintf out ", character ~S" code)) )
+	     (##sys#write-char-0 #\newline ##sys#standard-output) ]
+	    [(eq? x (##sys#slot '##sys#arbitrary-unbound-symbol 0))
+	     (fprintf out "unbound value~%") ]
+	    [(##sys#number? x) (fprintf out "number ~S~%" x)]
+	    [(string? x) (descseq "string" ##sys#size string-ref 0)]
+	    [(vector? x) (descseq "vector" ##sys#size ##sys#slot 0)]
+	    [(symbol? x)
+	     (unless (##sys#symbol-has-toplevel-binding? x) (display "unbound " out))
+	     (when (and (symbol? x) (fx= 0 (##sys#byte (##sys#slot x 1) 0)))
+	       (display "keyword " out) )
+	     (fprintf out "~asymbol with name ~S~%"
+		      (if (##sys#interned-symbol? x) "" "uninterned ")
+		      (##sys#symbol->string x))
+	     (let ((plist (##sys#slot x 2)))
+	       (unless (null? plist)
+		 (display "  \nproperties:\n\n" out)
+		 (do ((plist plist (cddr plist)))
+		     ((null? plist))
+		   (fprintf out "  ~s\t" (car plist))
+		   (##sys#with-print-length-limit
+		    1000
+		    (lambda ()
+		      (write (cadr plist) out) ) )
+		   (newline out) ) ) ) ]
+	    [(list? x) (descseq "list" length list-ref 0)]
+	    [(pair? x) (fprintf out "pair with car ~S and cdr ~S~%" (car x) (cdr x))]
+	    [(procedure? x)
+	     (let ([len (##sys#size x)])
+	       (if (and (> len 3)
+			(memq #:tinyclos ##sys#features)
+			(eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) ) ;XXX handle this in tinyclos egg (difficult)
+		   (describe-object x out)
+		   (descseq 
+		    (sprintf "procedure with code pointer ~X" (##sys#peek-unsigned-integer x 0))
+		    ##sys#size ##sys#slot 1) ) ) ]
+	    [(port? x)
+	     (fprintf out
+		      "~A port of type ~A with name ~S and file pointer ~X~%"
+		      (if (##sys#slot x 1) "input" "output")
+		      (##sys#slot x 7)
+		      (##sys#slot x 3)
+		      (##sys#peek-unsigned-integer x 0) ) ]
+	    [(and (memq #:tinyclos ##sys#features) (instance? x)) ; XXX put into tinyclos egg
+	     (describe-object x out) ]
+	    [(##sys#locative? x)
+	     (fprintf out "locative~%  pointer ~X~%  index ~A~%  type ~A~%"
+		      (##sys#peek-unsigned-integer x 0)
+		      (##sys#slot x 1)
+		      (case (##sys#slot x 2) 
+			[(0) "slot"]
+			[(1) "char"]
+			[(2) "u8vector"]
+			[(3) "s8vector"]
+			[(4) "u16vector"]
+			[(5) "s16vector"]
+			[(6) "u32vector"]
+			[(7) "s32vector"]
+			[(8) "f32vector"]
+			[(9) "f64vector"] ) ) ]
+	    [(##sys#pointer? x) (fprintf out "machine pointer ~X~%" (##sys#peek-unsigned-integer x 0))]
+	    [(##sys#bytevector? x)
+	     (let ([len (##sys#size x)])
+	       (fprintf out "blob of size ~S:~%" len)
+	       (hexdump x len ##sys#byte out) ) ]
+	    [(##core#inline "C_lambdainfop" x)
+	     (fprintf out "lambda information: ~s~%" (##sys#lambda-info->string x)) ]
+	    [(##sys#structure? x 'hash-table)
+	     (let ((n (##sys#slot x 2)))
+	       (fprintf out "hash-table with ~S element~a~%  comparison procedure: ~A~%"
+			n (if (fx= n 1) "" "s")  (##sys#slot x 3)) )
+	     (fprintf out "  hash function: ~a~%" (##sys#slot x 4))
+	     (hash-table-walk		; blindly assumes it is bound
+	      x
+	      (lambda (k v) (fprintf out " ~S\t-> ~S~%" k v)) ) ]
+	    [(##sys#structure? x 'condition)
+	     (fprintf out "condition: ~s~%" (##sys#slot x 1))
+	     (for-each
+	      (lambda (k)
+		(fprintf out " ~s~%" k)
+		(let loop ((props (##sys#slot x 2)))
+		  (unless (null? props)
+		    (when (eq? k (caar props))
+		      (fprintf out "\t~s: ~s~%" (cdar props) (cadr props)) )
+		    (loop (cddr props)) ) ) )
+	      (##sys#slot x 1) ) ]
+	    [(and (##sys#structure? x 'meroon-instance) (provided? 'meroon)) ; XXX put this into meroon egg (really!)
+	     (unveil x out) ]
+	    [(##sys#generic-structure? x)
+	     (let ([st (##sys#slot x 0)])
+	       (cond ((##sys#hash-table-ref describer-table st) => (cut <> x out))
+		     ((assq st bytevector-data) =>
+		      (lambda (data)
+			(apply descseq (append (map eval (cdr data)) (list 0)))) )
+		     (else
+		      (fprintf out "structure of type `~S':~%" (##sys#slot x 0))
+		      (descseq #f ##sys#size ##sys#slot 1) ) ) ) ]
+	    [else (fprintf out "unknown object~%")] )
+      (##sys#void) ) ) )
 
-	(define (descseq name plen pref start)
-	  (let ((len (fx- (plen x) start)))
-	    (when name (fprintf out "~A of length ~S~%" name len))
-	    (let loop1 ((i 0))
-	      (cond ((fx>= i len))
-		    ((fx>= i max-describe-lines)
-		     (fprintf out "~% (~A elements not displayed)~%" (fx- len i)) )
-		    (else
-		     (let ((v (pref x (fx+ start i))))
-		       (let loop2 ((n 1) (j (fx+ i (fx+ start 1))))
-			 (cond ((fx>= j len)
-				(fprintf out " ~S: ~S" i v)
-				(if (fx> n 1)
-				    (fprintf out "\t(followed by ~A identical instance~a)~% ...~%" 
-					     (fx- n 1)
-					     (if (eq? n 2) "" "s"))
-				    (newline out) )
-				(loop1 (fx+ i n)) )
-			       ((eq? v (pref x j)) (loop2 (fx+ n 1) (fx+ j 1)))
-			       (else (loop2 n len)) ) ) ) ) ) ) ) )
-
-	(when (##sys#permanent? x)
-	  (fprintf out "statically allocated (0x~X) " (##sys#block-address x)) )
-	(cond [(char? x)
-	       (let ([code (char->integer x)])
-		 (fprintf out "character ~S, code: ~S, #x~X, #o~O~%" x code code code) ) ]
-	      [(eq? x #t) (fprintf out "boolean true~%")]
-	      [(eq? x #f) (fprintf out "boolean false~%")]
-	      [(null? x) (fprintf out "empty list~%")]
-	      [(eof-object? x) (fprintf out "end-of-file object~%")]
-	      [(eq? (##sys#void) x) (fprintf out "unspecified object~%")]
-	      [(fixnum? x)
-	       (fprintf out "exact integer ~S, #x~X, #o~O, #b~B" x x x x)
-	       (let ([code (integer->char x)])
-		 (when (fx< code #x10000) (fprintf out ", character ~S" code)) )
-	       (##sys#write-char-0 #\newline ##sys#standard-output) ]
-	      [(eq? x (##sys#slot '##sys#arbitrary-unbound-symbol 0))
-	       (fprintf out "unbound value~%") ]
-	      [(##sys#number? x) (fprintf out "number ~S~%" x)]
-	      [(string? x) (descseq "string" ##sys#size string-ref 0)]
-	      [(vector? x) (descseq "vector" ##sys#size ##sys#slot 0)]
-	      [(symbol? x)
-	       (unless (##sys#symbol-has-toplevel-binding? x) (display "unbound " out))
-	       (when (and (symbol? x) (fx= 0 (##sys#byte (##sys#slot x 1) 0)))
-		 (display "keyword " out) )
-	       (fprintf out "symbol with name ~S~%" (##sys#symbol->string x)) ]
-	      [(list? x) (descseq "list" length list-ref 0)]
-	      [(pair? x) (fprintf out "pair with car ~S and cdr ~S~%" (car x) (cdr x))]
-	      [(procedure? x)
-	       (let ([len (##sys#size x)])
-		 (if (and (> len 3)
-			  (memq #:tinyclos ##sys#features)
-			  (eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) ) ;XXX handle this in tinyclos egg (difficult)
-		     (describe-object x out)
-		     (descseq 
-		      (sprintf "procedure with code pointer ~X" (##sys#peek-unsigned-integer x 0))
-		      ##sys#size ##sys#slot 1) ) ) ]
-	      [(port? x)
-	       (fprintf out
-			"~A port of type ~A with name ~S and file pointer ~X~%"
-			(if (##sys#slot x 1) "input" "output")
-			(##sys#slot x 7)
-			(##sys#slot x 3)
-			(##sys#peek-unsigned-integer x 0) ) ]
-	      [(and (memq #:tinyclos ##sys#features) (instance? x)) ; XXX put into tinyclos egg
-	       (describe-object x out) ]
-	      [(##sys#locative? x)
-	       (fprintf out "locative~%  pointer ~X~%  index ~A~%  type ~A~%"
-			(##sys#peek-unsigned-integer x 0)
-			(##sys#slot x 1)
-			(case (##sys#slot x 2) 
-			  [(0) "slot"]
-			  [(1) "char"]
-			  [(2) "u8vector"]
-			  [(3) "s8vector"]
-			  [(4) "u16vector"]
-			  [(5) "s16vector"]
-			  [(6) "u32vector"]
-			  [(7) "s32vector"]
-			  [(8) "f32vector"]
-			  [(9) "f64vector"] ) ) ]
-	      [(##sys#pointer? x) (fprintf out "machine pointer ~X~%" (##sys#peek-unsigned-integer x 0))]
-	      [(##sys#bytevector? x)
-	       (let ([len (##sys#size x)])
-		 (fprintf out "blob of size ~S:~%" len)
-		 (hexdump x len ##sys#byte out) ) ]
-	      [(##core#inline "C_lambdainfop" x)
-	       (fprintf out "lambda information: ~s~%" (##sys#lambda-info->string x)) ]
-	      [(##sys#structure? x 'hash-table)
-	       (let ((n (##sys#slot x 2)))
-		 (fprintf out "hash-table with ~S element~a~%  comparison procedure: ~A~%"
-			  n (if (fx= n 1) "" "s")  (##sys#slot x 3)) )
-	       (fprintf out "  hash function: ~a~%" (##sys#slot x 4))
-	       (hash-table-walk
-		x
-		(lambda (k v) (fprintf out " ~S\t-> ~S~%" k v)) ) ]
-	      [(##sys#structure? x 'condition)
-	       (fprintf out "condition: ~s~%" (##sys#slot x 1))
-	       (for-each
-		(lambda (k)
-		  (fprintf out " ~s~%" k)
-		  (let loop ((props (##sys#slot x 2)))
-		    (unless (null? props)
-		      (when (eq? k (caar props))
-			(fprintf out "\t~s: ~s~%" (cdar props) (cadr props)) )
-		      (loop (cddr props)) ) ) )
-		(##sys#slot x 1) ) ]
-	      [(and (##sys#structure? x 'meroon-instance) (provided? 'meroon)) ; XXX put this into meroon egg (really!)
-	       (unveil x out) ]
-	      [(##sys#generic-structure? x)
-	       (let ([st (##sys#slot x 0)])
-		 (cond ((hash-table-ref/default describer-table st #f) => (cut <> x out))
-		       ((assq st bytevector-data) =>
-			(lambda (data)
-			  (apply descseq (append (map eval (cdr data)) (list 0)))) )
-		       (else
-			(fprintf out "structure of type `~S':~%" (##sys#slot x 0))
-			(descseq #f ##sys#size ##sys#slot 1) ) ) ) ]
-	      [else (fprintf out "unknown object~%")] )
-	(##sys#void) ) ) ) )
-
-(define set-describer!
-  (let ((hash-table-set! hash-table-set!))
-    (lambda (tag proc)
-      (hash-table-set! describer-table tag proc) ) ) )
+(define (set-describer! tag proc)
+  (##sys#check-symbol tag 'symbol 'set-describer!)
+  (##sys#hash-table-set! describer-table tag proc) )
 
 
 ;;; Display hexdump:
@@ -720,7 +761,7 @@ EOF
 	([len #f]
 	 [out ##sys#standard-output] )
       (define (bestlen n) (if len (min len n) n))
-      (cond [(##sys#immediate? x) (##sys#error 'dump "can not dump immediate object" x)]
+      (cond [(##sys#immediate? x) (##sys#error 'dump "cannot dump immediate object" x)]
 	    [(##sys#bytevector? x) (hexdump x (bestlen (##sys#size x)) ##sys#byte out)]
 	    [(string? x) (hexdump x (bestlen (##sys#size x)) ##sys#byte out)]
 	    [(and (not (##sys#immediate? x)) (##sys#pointer? x))
@@ -728,7 +769,7 @@ EOF
 	    [(and (##sys#generic-structure? x) (assq (##sys#slot x 0) bytevector-data))
 	     (let ([bv (##sys#slot x 1)])
 	       (hexdump bv (bestlen (##sys#size bv)) ##sys#byte out) ) ]
-	    [else (##sys#error 'dump "can not dump object" x)] ) ) ) )
+	    [else (##sys#error 'dump "cannot dump object" x)] ) ) ) )
 
 (define hexdump
   (let ([display display]
@@ -773,7 +814,7 @@ EOF
 ;;; Start interpreting:
 
 (define (deldups lis . maybe-=)
-  (let ((elt= (:optional maybe-= equal?)))
+  (let ((elt= (optional maybe-= equal?)))
     (let recur ((lis lis))
       (if (null? lis) lis
 	  (let* ((x (car lis))
@@ -793,29 +834,28 @@ EOF
   '(#\k #\s #\v #\h #\D #\e #\i #\R #\b #\n #\q #\w #\- #\I #\p #\P) )
 
 (define-constant long-options
-  '("-keyword-style" "-script" "-version" "-help" "--help" "--" "-feature" 
-    "-eval" "-case-insensitive"
+  '("-ss" "-sx" "-script" "-version" "-help" "--help" "-feature" "-eval"
+    "-case-insensitive" "-keyword-style" "-no-parentheses-synonyms" "-no-symbol-escape"
+    "-r5rs-syntax" "-setup-mode"
     "-require-extension" "-batch" "-quiet" "-no-warnings" "-no-init" 
-    "-include-path" "-release" "-ss"
-    "-print" "-pretty-print") )
+    "-include-path" "-release" "-print" "-pretty-print" "--") )
 
 (define (canonicalize-args args)
   (let loop ((args args))
     (if (null? args)
 	'()
 	(let ((x (car args)))
-	  (cond 
-	   ((member x '("-s" "-ss" "-script" "--")) args)
-	   ((and (fx> (##sys#size x) 2)
-		 (char=? #\- (##core#inline "C_subchar" x 0))
-		 (not (member x long-options)) )
-	    (if (char=? #\: (##core#inline "C_subchar" x 1))
-		(loop (cdr args))
-		(let ((cs (string->list (substring x 1))))
-		  (if (findall cs short-options)
-		      (append (map (cut string #\- <>) cs) (loop (cdr args)))
-		      (##sys#error "invalid option" x) ) ) ) )
-	   (else (cons x (loop (cdr args)))))))))
+	  (cond ((member x '("-s" "-ss" "-script" "-sx" "--")) args)
+                ((and (fx> (##sys#size x) 2)
+                       (char=? #\- (##core#inline "C_subchar" x 0))
+                       (not (member x long-options)) )
+                 (if (char=? #\: (##core#inline "C_subchar" x 1))
+                     (loop (cdr args))
+                     (let ((cs (string->list (substring x 1))))
+                       (if (findall cs short-options)
+                           (append (map (cut string #\- <>) cs) (loop (cdr args)))
+                           (##sys#error "invalid option" x) ) ) ) )
+	        (else (cons x (loop (cdr args)))))))))
 
 (define (findall chars clist)
   (let loop ((chars chars))
@@ -823,11 +863,21 @@ EOF
 	(and (memq (car chars) clist)
 	     (loop (cdr chars))))))
 
+(define-constant simple-options
+  '("--" "-b" "-batch" "-q" "-quiet" "-n" "-no-init" "-w" "-no-warnings" "-i" "-case-insensitive"
+    "-no-parentheses-synonyms" "-no-symbol-escape" "-r5rs-syntax" "-setup-mode"
+    ; Not "simple" but processed early
+    "-ss" "-sx" "-s" "-script") )
+
+(define-constant complex-options
+  '("-D" "-feature" "-I" "-include-path" "-k" "-keyword-style") )
+
 (define (run)
-  (let* ([extraopts (parse-option-string (or (getenv "CSI_OPTIONS") ""))]
+  (let* ([extraopts (parse-option-string (or (get-environment-variable "CSI_OPTIONS") ""))]
 	 [args (canonicalize-args (command-line-arguments))]
+	 ; Check for these before 'args' is updated by any 'extraopts'
 	 [kwstyle (member* '("-k" "-keyword-style") args)]
-	 [script (member* '("-s" "-ss" "-script") args)])
+	 [script (member* '("-ss" "-sx" "-s" "-script") args)])
     (cond [script
 	   (when (or (not (pair? (cdr script)))
 		     (zero? (string-length (cadr script)))
@@ -848,7 +898,7 @@ EOF
 	   [batch (or script (member* '("-b" "-batch") args) eval?)]
 	   [quietflag (member* '("-q" "-quiet") args)]
 	   [quiet (or script quietflag eval?)]
-	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )      
+	   [ipath (map chop-separator (string-split (or (get-environment-variable "CHICKEN_INCLUDE_PATH") "") ";"))] )      
       (define (collect-options opt)
 	(let loop ([opts args])
 	  (cond [(member opt opts) 
@@ -861,7 +911,7 @@ EOF
 	(let ([fn (##sys#string-append "./" init-file)])
 	  (if (file-exists? fn)
 	      (load fn)
-	      (let* ([prefix (chop-separator (or (getenv "HOME") "."))]
+	      (let* ([prefix (chop-separator (or (get-environment-variable "HOME") "."))]
 		     [fn (string-append prefix "/" init-file)] )
 		(when (file-exists? fn) 
 		  (load fn) ) ) ) ) )
@@ -877,6 +927,8 @@ EOF
       (when (member* '("-v" "-version") args)
 	(print-banner)
 	(exit 0) )
+      (when (member "-setup-mode" args)
+	(set! ##sys#setup-mode #t))
       (when (member "-release" args)
 	(print (chicken-version))
 	(exit 0) )
@@ -899,8 +951,6 @@ EOF
 		 ##sys#include-pathnames
 		 ipath)
 	 string=?) )
-      (set! ##sys#features (cons #:match ##sys#features))
-      (provide 'match)
       (when kwstyle
 	(cond [(not (pair? (cdr kwstyle)))
 	       (##sys#error "missing argument to `-keyword-style' option") ]
@@ -910,46 +960,56 @@ EOF
 	       (keyword-style #:none) ]
 	      [(string=? "suffix" (cadr kwstyle))
 	       (keyword-style #:suffix) ] ) )
+      (when (member* '("-no-parentheses-synonyms") args)
+	(unless quiet (display "Disabled support for parentheses synonyms\n"))
+        (parentheses-synonyms #f) )
+      (when (member* '("-no-symbol-escape") args)
+	(unless quiet (display "Disabled support for escaped symbols\n"))
+	(symbol-escape #f) )
+      (when (member* '("-r5rs-syntax") args)
+	(unless quiet (display "Disabled the Chicken extensions to R5RS syntax\n"))
+	(case-sensitive #f)
+	(keyword-style #:none)
+	(parentheses-synonyms #f)
+	(symbol-escape #f) )
       (unless (or (member* '("-n" "-no-init") args) script) (loadinit))
       (do ([args args (cdr args)])
 	  ((null? args)
 	   (unless batch 
 	     (repl)
 	     (##sys#write-char-0 #\newline ##sys#standard-output) ) )
-	(let* ([arg (car args)]
-	       [len (string-length arg)] )
-	  (cond ((member 
-		  arg 
-		  '("--" "-batch" "-quiet" "-no-init" "-no-warnings" "-script"
-		    "-b" "-q" "-n" "-w" "-s" "-i"
-		    "-case-insensitive" "-ss") ) )
-		((member arg '("-feature" "-include-path" "-keyword-style" 
-			       "-D" "-I" "-k"))
+	(let* ((arg (car args)))
+	  (cond ((member arg simple-options) )
+		((member arg complex-options)
 		 (set! args (cdr args)) )
 		((or (string=? "-R" arg) (string=? "-require-extension" arg))
-		 (eval `(##core#require-extension ',(string->symbol (cadr args))))
+		 (eval `(##core#require-extension (,(string->symbol (cadr args))) #t))
 		 (set! args (cdr args)) )
 		((or (string=? "-e" arg) (string=? "-eval" arg))
 		 (evalstring (cadr args))
 		 (set! args (cdr args)) )
 		((or (string=? "-p" arg) (string=? "-print" arg))
-		 (evalstring 
-		  (cadr args)
-		  (cut for-each print <...>) )
+		 (evalstring (cadr args) (cut for-each print <...>))
 		 (set! args (cdr args)) )
 		((or (string=? "-P" arg) (string=? "-pretty-print" arg))
-		 (evalstring 
-		  (cadr args)
-		  (cut for-each pretty-print <...>) )
+		 (evalstring (cadr args) (cut for-each pretty-print <...>) )
 		 (set! args (cdr args)) )
 		(else
-		 (load arg) 
-		 (when (and script (string=? "-ss" (car script)))
-		   (call-with-values (cut main (command-line-arguments))
-		     (lambda results
-		       (exit
-			(if (and (pair? results) (fixnum? (car results)))
-			    (car results)
-			    0) ) ) ) ) ) ) ) ) ) ) )
+		 (let ((scr (and script (car script))))
+		   (##sys#load 
+		    arg 
+		    (and (equal? "-sx" scr)
+			 (lambda (x)
+			   (pretty-print x ##sys#standard-error)
+			   (newline ##sys#standard-error)
+			   (eval x)))
+		    #f)
+		   (when (equal? "-ss" scr)
+		     (call-with-values (cut main (command-line-arguments))
+		       (lambda results
+			 (exit
+			  (if (and (pair? results) (fixnum? (car results)))
+			      (car results)
+			      0) ) ) ) ) ) ) ) ) ) ) ) )
 
 (run)
