@@ -1,6 +1,6 @@
 ;;; data-structures.scm - Optional data structures extensions
 ;
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
@@ -43,7 +43,7 @@ EOF
     (bound-to-procedure
       ##sys#check-char ##sys#check-exact ##sys#check-port ##sys#check-string
       ##sys#substring ##sys#for-each ##sys#map ##sys#setslot
-      ##sys#allocate-vector ##sys#check-pair ##sys#not-a-proper-list-error
+      ##sys#allocate-vector ##sys#check-pair ##sys#error-not-a-proper-list
       ##sys#member ##sys#assoc ##sys#error ##sys#signal-hook ##sys#read-string!
       ##sys#check-symbol ##sys#check-vector ##sys#floor ##sys#ceiling
       ##sys#truncate ##sys#round ##sys#check-number ##sys#cons-flonum
@@ -62,31 +62,13 @@ EOF
 
 (private data-structures
   reverse-string-append
-  fprintf0 generic-write 
-   )
+  fprintf0 generic-write )
 
 (declare
   (hide
     fprintf0 generic-write ) )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-closure . _) '(##core#undefined))
-    (define-macro (##sys#check-inexact . _) '(##core#undefined))
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "data-structures.exports")) ] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'data-structures)
 
@@ -147,7 +129,7 @@ EOF
 	      h
 	      (lambda (x) (h ((loop t) x))))))))
 
-(define (list-of pred)
+(define (list-of? pred)
   (lambda (lst)
     (let loop ([lst lst])
       (cond [(null? lst) #t]
@@ -155,6 +137,8 @@ EOF
 	    [(pred (##sys#slot lst 0)) (loop (##sys#slot lst 1))]
 	    [else #f] ) ) ) )
 
+(define list-of list-of?)		; DEPRECATED
+
 (define (noop . _) (void))
 
 (define (each . procs)
@@ -257,7 +241,7 @@ EOF
     (let loop ([lsts lsts])
       (cond [(null? lsts) '()]
 	    [(cond-expand [unsafe #f] [else (not (pair? lsts))])
-	     (##sys#not-a-proper-list-error lsts) ]
+	     (##sys#error-not-a-proper-list lsts) ]
 	    [else
 	     (let ([l (##sys#slot lsts 0)]
 		   [r (##sys#slot lsts 1)] )
@@ -531,7 +515,7 @@ EOF
 	       (##sys#check-string stri 'string-intersperse)
 	       (loop1 (##sys#slot ss 1)
 		      (fx+ (##sys#size stri) (fx+ dslen n)) ) ) )
-	    (else (##sys#not-a-proper-list-error strs)) ) ) ) )
+	    (else (##sys#error-not-a-proper-list strs)) ) ) ) )
 
 
 ;;; Translate elements of a string:
@@ -796,6 +780,54 @@ EOF
 	(sort! (append seq '()) less?)))
 
 
+;;;  Simple topological sort:
+;
+; Taken from SLIB (slightly adapted): Copyright (C) 1995 Mikael Djurfeldt
+
+(define (topological-sort dag pred)
+  (if (null? dag)
+      '()
+      (let* ((adj-table '())
+	     (sorted '()))
+
+	(define (insert x y)
+	  (let loop ([at adj-table])
+	    (cond [(null? at) (set! adj-table (cons (cons x y) adj-table))]
+		  [(pred x (caar at)) (set-cdr! (car at) y)]
+		  [else (loop (cdr at))] ) ) )
+	
+	(define (lookup x)
+	  (let loop ([at adj-table])
+	    (cond [(null? at) #f]
+		  [(pred x (caar at)) (cdar at)]
+		  [else (loop (cdr at))] ) ) )
+	
+	(define (visit u adj-list)
+	  ;; Color vertex u
+	  (insert u 'colored)
+	  ;; Visit uncolored vertices which u connects to
+	  (for-each (lambda (v)
+		      (let ((val (lookup v)))
+			(if (not (eq? val 'colored))
+			    (visit v (or val '())))))
+		    adj-list)
+	  ;; Since all vertices downstream u are visited
+	  ;; by now, we can safely put u on the output list
+	  (set! sorted (cons u sorted)) )
+	
+	;; Hash adjacency lists
+	(for-each (lambda (def) (insert (car def) (cdr def)))
+		  (cdr dag))
+	;; Visit vertices
+	(visit (caar dag) (cdar dag))
+	(for-each (lambda (def)
+		    (let ((val (lookup (car def))))
+		      (if (not (eq? val 'colored))
+			  (visit (car def) (cdr def)))))
+		  (cdr dag)) 
+	sorted) ) )
+
+
 ;;; Binary search:
 
 (define binary-search
@@ -897,7 +929,7 @@ EOF
 	   ((eq? (##sys#slot lst 1) '()) lst)
 	 (if (or (not (##core#inline "C_blockp" lst))
 		 (not (##core#inline "C_pairp" lst)) )
-	     (##sys#not-a-proper-list-error lst0 'list->queue) ) ) ) ) )
+	     (##sys#error-not-a-proper-list lst0 'list->queue) ) ) ) ) )
 
 
 ; (queue-push-back! queue item)
@@ -914,9 +946,9 @@ EOF
 ; Pushes the items in item-list back onto the queue,
 ; so that (car item-list) becomes the next removable item.
 
-(define-macro (last-pair lst0)
-  `(do ((lst ,lst0 (##sys#slot lst 1)))
-       ((eq? (##sys#slot lst 1) '()) lst)))
+(define-inline (last-pair lst0)
+  (do ((lst lst0 (##sys#slot lst 1)))
+      ((eq? (##sys#slot lst 1) '()) lst)))
 
 (define (queue-push-back-list! q itemlist)
   (##sys#check-structure q 'queue 'queue-push-back-list!)
