@@ -1,7 +1,7 @@
 ;;; extras.scm - Optional non-standard extensions
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -29,11 +29,7 @@
  (unit extras)
  (uses data-structures ports)
  (usual-integrations)
- (disable-warning redef)
- (foreign-declare #<<EOF
-#define C_mem_compare(to, from, n)   C_fix(C_memcmp(C_c_string(to), C_c_string(from), C_unfix(n)))
-EOF
-) )
+ (disable-warning redef) )
 
 (cond-expand
  [paranoia]
@@ -44,7 +40,7 @@ EOF
     (bound-to-procedure
       ##sys#check-char ##sys#check-exact ##sys#check-port ##sys#check-string
       ##sys#substring ##sys#for-each ##sys#map ##sys#setslot
-      ##sys#allocate-vector ##sys#check-pair ##sys#not-a-proper-list-error
+      ##sys#allocate-vector ##sys#check-pair ##sys#error-not-a-proper-list
       ##sys#member ##sys#assoc ##sys#error ##sys#signal-hook ##sys#read-string!
       ##sys#check-symbol ##sys#check-vector ##sys#floor ##sys#ceiling
       ##sys#truncate ##sys#round ##sys#check-number ##sys#cons-flonum
@@ -52,7 +48,7 @@ EOF
       ##sys#print ##sys#check-structure ##sys#make-structure make-parameter
       ##sys#flush-output ##sys#write-char-0 ##sys#number->string
       ##sys#fragments->string ##sys#symbol->qualified-string
-      ##extras#reverse-string-append ##sys#number? ##sys#procedure->string
+      reverse-string-append ##sys#number? ##sys#procedure->string
       ##sys#pointer->string ##sys#user-print-hook ##sys#peek-char-0
       ##sys#read-char-0 ##sys#write-char ##sys#string-append ##sys#gcd ##sys#lcm
       ##sys#fudge ##sys#check-list ##sys#user-read-hook ##sys#check-closure ##sys#check-inexact
@@ -61,35 +57,15 @@ EOF
       make-string string pretty-print-width newline char-name read random
       open-input-string make-string call-with-input-file read-line reverse ) ) ] )
 
-(private extras
-  reverse-string-append
-  fprintf0 generic-write )
-
 (declare
   (hide
-    fprintf0 generic-write ) )
+    fprintf0 generic-write reverse-string-append) )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-closure . _) '(##core#undefined))
-    (define-macro (##sys#check-inexact . _) '(##core#undefined))
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "extras.exports")) ] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'extras)
 
+
 ;;; Read expressions from file:
 
 (define read-file
@@ -107,34 +83,50 @@ EOF
 	  (call-with-input-file port slurp) ) ) ) )
 
 
-
 ;;; Random numbers:
 
-(define random-seed
-    (let ((srand   (foreign-lambda void "srand" unsigned-integer)))
+(cond-expand
+  (unix
+
+    (define random-seed)
+    (define randomize)
+
+    (let ((srandom (foreign-lambda void "srandom" unsigned-integer)))
+
+      (set! random-seed
+        (lambda (#!optional (seed (current-seconds)))
+          (##sys#check-integer seed 'random-seed)
+          (srandom seed) ) )
+
+      (set! randomize
+        (lambda (#!optional (seed (##sys#fudge 2)))
+          (##sys#check-exact seed 'randomize)
+          (srandom seed) ) ) )
+
+    (define (random n)
+      (##sys#check-integer n 'random)
+      (if (eq? 0 n)
+          0
+          ((foreign-lambda* long ((integer64 n)) "return( random() % ((uint64_t) n) );") n) ) ) )
+  (else
+
+    (define random-seed
+      (let ((srand (foreign-lambda void "srand" unsigned-integer)))
         (lambda n
-            (and (> (length n) 1)
-                 (##sys#error 'random-seed "too many arguments" (length n) 1))
-            (let ((t   (if (null? n)
-                           (current-seconds)
-                           (car n))))
-                (##sys#check-integer t 'random-seed)
-                (srand t)))))
+          (let ((t (if (null? n) (current-seconds) (car n))))
+            (##sys#check-integer t 'random-seed)
+            (srand t) ) ) ) )
 
-(define (random n)
-  (##sys#check-exact n 'random)
-  (if (eq? n 0)
-      0
-      (##core#inline "C_random_fixnum" n) ) )
+    (define (randomize . n)
+      (let ((nn (if (null? n) (##sys#fudge 2) (car n))))
+        (##sys#check-exact nn 'randomize)
+        (##core#inline "C_randomize" nn) ) )
 
-(define (randomize . n)
-  (##core#inline
-   "C_randomize"
-   (if (##core#inline "C_eqp" n '())
-       (##sys#fudge 2)
-       (let ((nn (##sys#slot n 0)))
-	 (##sys#check-exact nn 'randomize)
-	 nn) ) ) )
+    (define (random n)
+      (##sys#check-exact n 'random)
+      (if (eq? n 0)
+          0
+          (##core#inline "C_random_fixnum" n) ) ) ) )
 
 
 ;;; Line I/O:
@@ -212,19 +204,26 @@ EOF
 	   (##core#inline "C_setsubchar" dest start (##sys#read-char-0 port))
 	   (set! start (fx+ start 1)) )
 	 (let ((rdstring (##sys#slot (##sys#slot port 2) 7)))
-	   (let loop ((start start) (n n) (m 0))
-	     (let ((n2 (if rdstring
-			   (rdstring port n dest start) ; *** doesn't update port-position!
-			   (let ((c (##sys#read-char-0 port)))
+           (if rdstring
+	       (let loop ((start start) (n n) (m 0))
+		 (let ((n2 (rdstring port n dest start)))
+		   (##sys#setislot port 5 ; update port-position
+				   (fx+ (##sys#slot port 5) n2))
+		   (cond ((eq? n2 0) m)
+			 ((or (not n) (fx< n2 n)) 
+			  (loop (fx+ start n2) (and n (fx- n n2)) (fx+ m n2)))
+			 (else (fx+ n2 m)))))
+	       (let loop ((start start) (n n) (m 0))
+		 (let ((n2 (let ((c (##sys#read-char-0 port)))
 			     (if (eof-object? c)
 				 0
 				 (begin
 				   (##core#inline "C_setsubchar" dest start c)
-				   1) ) ) ) ) )
-	       (cond ((eq? n2 0) m)
-		     ((or (not n) (fx< n2 n)) 
-		      (loop (fx+ start n2) (and n (fx- n n2)) (fx+ m n2)) )
-		     (else (fx+ n2 m))) ) ) ))))
+				   1) ) )  ) )
+		   (cond ((eq? n2 0) m)
+			 ((or (not n) (fx< n2 n)) 
+			  (loop (fx+ start n2) (and n (fx- n n2)) (fx+ m n2)) )
+			 (else (fx+ n2 m))) )))))))
 
 (define (read-string! n dest #!optional (port ##sys#standard-input) (start 0))
   (##sys#check-port port 'read-string!)
@@ -236,6 +235,8 @@ EOF
   (##sys#check-exact start 'read-string!)
   (##sys#read-string! n dest port start) )
 
+(define-constant read-string-buffer-size 2048)
+
 (define ##sys#read-string/port
   (let ((open-output-string open-output-string)
 	(get-output-string get-output-string) )
@@ -248,15 +249,16 @@ EOF
 		     str
 		     (##sys#substring str 0 n2))))
 	    (else
-	     (let ([str (open-output-string)])
-	       (let loop ([n n])
-		 (or (and (eq? n 0) (get-output-string str))
-		     (let ([c (##sys#read-char-0 p)])
-		       (if (eof-object? c)
-			   (get-output-string str)
-			   (begin
-			     (##sys#write-char/port c str) 
-			     (loop (and n (fx- n 1))) ) ) ) ) ) ) ) ) ) ) )
+	     (let ([out (open-output-string)]
+		   (buf (make-string read-string-buffer-size)))
+	       (let loop ()
+		 (let ((n (##sys#read-string! read-string-buffer-size
+					      buf p 0)))
+		   (cond ((eq? n 0)
+			  (get-output-string out))
+			 (else
+			  (write-string buf n out)
+			  (loop)))))))))))
 
 (define (read-string #!optional n (port ##sys#standard-input))
   (##sys#read-string/port n port) )
@@ -265,7 +267,7 @@ EOF
   (let ([open-output-string open-output-string]
 	[get-output-string get-output-string] )
     (lambda (pred . port)
-      (let ([port (:optional port ##sys#standard-input)])
+      (let ([port (optional port ##sys#standard-input)])
 	(##sys#check-port port 'read-token)
 	(let ([out (open-output-string)])
 	  (let loop ()
@@ -374,7 +376,7 @@ EOF
 
 	(cond ((pair? obj)        (wr-expr obj col))
 	      ((null? obj)        (wr-lst obj col))
-	      ((eof-object? obj)  (out "#<eof>" col))
+	      ((eof-object? obj)  (out "#!eof" col))
 	      ((vector? obj)      (wr-lst (vector->list obj) (out "#" col)))
 	      ((boolean? obj)     (out (if obj "#t" "#f") col))
 	      ((##sys#number? obj)      (out (##sys#number->string obj) col))
@@ -636,59 +638,61 @@ EOF
 
 (define fprintf0
   (let ((write write)
-        (newline newline)
-        (display display) 
-        (open-output-string open-output-string)
-        (get-output-string get-output-string))
+	(newline newline)
+	(display display) 
+	(open-output-string open-output-string)
+	(get-output-string get-output-string))
     (lambda (loc port msg args)
       (when port (##sys#check-port port loc))
-      (let ((out (if (and port (##sys#tty-port? port)) port (open-output-string))))
-        (let rec ([msg msg] [args args])
-          (##sys#check-string msg loc)
-          (let ((index 0)
-                (len (##sys#size msg)) )
-            (define (fetch)
-              (let ((c (##core#inline "C_subchar" msg index)))
-                (set! index (fx+ index 1))
-                c) )
-            (define (next)
-              (if (cond-expand [unsafe #f] [else (##core#inline "C_eqp" args '())])
-                  (##sys#error loc "too few arguments to formatted output procedure")
-                  (let ((x (##sys#slot args 0)))
-                    (set! args (##sys#slot args 1)) 
-                    x) ) )
-            (let loop ()
-              (unless (fx>= index len)
-                (let ((c (fetch)))
-                  (if (and (eq? c #\~) (fx< index len))
-                      (let ((dchar (fetch)))
-                        (case (char-upcase dchar)
-                          ((#\S) (write (next) out))
-                          ((#\A) (display (next) out))
-                          ((#\C) (##sys#write-char-0 (next) out))
-                          ((#\B) (display (##sys#number->string (next) 2) out))
-                          ((#\O) (display (##sys#number->string (next) 8) out))
-                          ((#\X) (display (##sys#number->string (next) 16) out))
-                          ((#\!) (##sys#flush-output out))
-                          ((#\?)
-                           (let* ([fstr (next)]
-                                  [lst (next)] )
-                             (##sys#check-list lst loc)
-                             (rec fstr lst) ) )
-                          ((#\~) (##sys#write-char-0 #\~ out))
-                          ((#\% #\N) (newline out))
-                          (else
-                           (if (char-whitespace? dchar)
-                               (let skip ((c (fetch)))
-                                 (if (char-whitespace? c)
-                                     (skip (fetch))
-                                     (set! index (fx- index 1)) ) )
-                               (##sys#error loc "illegal format-string character" dchar) ) ) ) )
-                      (##sys#write-char-0 c out) )
-                  (loop) ) ) )
-            (cond ((not port) (get-output-string out))
-                  ((not (eq? out port))
-                   (##sys#print (get-output-string out) #f port) ) ) ) ) ) ) ) )
+      (let ((out (if (and port (##sys#tty-port? port))
+		     port
+		     (open-output-string))))
+      (let rec ([msg msg] [args args])
+	(##sys#check-string msg loc)
+	(let ((index 0)
+	      (len (##sys#size msg)) )
+	  (define (fetch)
+	    (let ((c (##core#inline "C_subchar" msg index)))
+	      (set! index (fx+ index 1))
+	      c) )
+	  (define (next)
+	    (if (cond-expand [unsafe #f] [else (##core#inline "C_eqp" args '())])
+		(##sys#error loc "too few arguments to formatted output procedure")
+		(let ((x (##sys#slot args 0)))
+		  (set! args (##sys#slot args 1)) 
+		  x) ) )
+	  (let loop ()
+	    (unless (fx>= index len)
+	      (let ((c (fetch)))
+		(if (and (eq? c #\~) (fx< index len))
+		    (let ((dchar (fetch)))
+		      (case (char-upcase dchar)
+			((#\S) (write (next) out))
+			((#\A) (display (next) out))
+			((#\C) (##sys#write-char-0 (next) out))
+			((#\B) (display (##sys#number->string (next) 2) out))
+			((#\O) (display (##sys#number->string (next) 8) out))
+			((#\X) (display (##sys#number->string (next) 16) out))
+			((#\!) (##sys#flush-output out))
+			((#\?)
+			 (let* ([fstr (next)]
+				[lst (next)] )
+			   (##sys#check-list lst loc)
+			   (rec fstr lst) out) )
+			((#\~) (##sys#write-char-0 #\~ out))
+			((#\% #\N) (newline out))
+			(else
+			 (if (char-whitespace? dchar)
+			     (let skip ((c (fetch)))
+			       (if (char-whitespace? c)
+				   (skip (fetch))
+				   (set! index (fx- index 1)) ) )
+			     (##sys#error loc "illegal format-string character" dchar) ) ) ) )
+		    (##sys#write-char-0 c out) )
+		(loop) ) ) ) ) )
+      (cond ((not port) (get-output-string out))
+	    ((not (eq? out port))
+	     (##sys#print (get-output-string out) #f port) ) ) ) ) ) )
 
 (define (fprintf port fstr . args)
   (fprintf0 'fprintf port fstr args) )
