@@ -1,7 +1,7 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -27,49 +27,8 @@
 
 (declare (unit optimizer))
 
-(private compiler
-  compiler-arguments process-command-line perform-lambda-lifting!
-  default-standard-bindings default-extended-bindings side-effecting-standard-bindings
-  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
-  standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home decompose-lambda-list external-to-pointer
-  copy-node! export-list inline-list not-inline-list
-  unit-name insert-timer-checks used-units external-variables
-  debug-info-index debug-info-vector-name profile-info-vector-name
-  foreign-declarations emit-trace-info block-compilation line-number-database-size
-  always-bound-to-procedure block-globals make-block-variable-literal block-variable-literal? block-variable-literal-name
-  target-heap-size target-stack-size constant-declarations
-  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
-  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  broken-constant-nodes inline-substitutions-enabled loop-lambda-names expand-profile-lambda
-  profile-lambda-list profile-lambda-index emit-profile expand-profile-lambda
-  direct-call-ids foreign-type-table first-analysis expand-debug-lambda expand-debug-assignment expand-debug-call
-  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database! scan-toplevel-assignments
-  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call compiler-warning
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! expand-foreign-callback-lambda* debug-lambda-list debug-variable-list debugging
-  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
-  string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
-  collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
-  put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
-  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
-  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
-  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database
-  expand-foreign-callback-lambda default-optimization-passes default-optimization-passes-when-trying-harder
-  units-used-by-default words-per-flonum rewrite
-  parameter-limit eq-inline-operator optimizable-rest-argument-operators
-  membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  make-random-name final-foreign-type inline-max-size simplified-ops
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result)
-
-
-(eval-when (compile eval)
-  (match-error-control #:fail) )
 
+(include "compiler-namespace")
 (include "tweaks")
 
 (define-constant maximal-number-of-free-variables-for-liftable 16)
@@ -122,7 +81,7 @@
 
        (scan node '()) ) )
     (debugging 'o "safe globals" safe)
-    (set! always-bound (append safe always-bound)) ) )
+    (for-each (cut mark-variable <> '##compiler#always-bound) safe)))
 
 
 ;;; Do some optimizations:
@@ -195,9 +154,8 @@
 	       ((##core#call)
 		(if (eq? '##core#variable (node-class (car subs)))
 		    (let ((var (first (node-parameters (car subs)))))
-		      (if (and (or (test var 'standard-binding)
-				   (test var 'extended-binding) )
-			       (test var 'foldable)
+		      (if (and (intrinsic? var)
+			       (foldable? var)
 			       (every constant-node? (cddr subs)) )
 			  (let ((form (cons var (map (lambda (arg) `(quote ,(node-value arg)))
 						     (cddr subs) ) ) ) )
@@ -240,7 +198,7 @@
 
 	  ((let)
 	   (let ([var (first params)])
-	     (cond [(or (test var 'replacable)
+	     (cond [(or ;(test var 'replacable)
 			(test var 'removable)
 			(and (test var 'contractable) (not (test var 'replacing))) )
 		    (touch)
@@ -288,7 +246,9 @@
 	       [(##core#variable)
 		;; Call to named procedure:
 		(let* ([var (first (node-parameters fun))]
-		       [lval (and (not (test var 'unknown)) (test var 'value))]
+		       [lval (and (not (test var 'unknown)) 
+				  (or (test var 'value)
+				      (test var 'local-value)))]
 		       [args (cdr subs)] )
 		  (cond [(test var 'contractable)
 			 (let* ([lparams (node-parameters lval)]
@@ -302,7 +262,7 @@
 					((eq? '##core#variable (node-class k)))
 					(kvar (first (node-parameters k)))
 					(lval (and (not (test kvar 'unknown)) (test kvar 'value))) 
-					(eq? '##core#lambda (node-class lval))
+					((eq? '##core#lambda (node-class lval)))
 					(llist (third (node-parameters lval)))
 					((or (test (car llist) 'unused)
 					     (and (not (test (car llist) 'references))
@@ -313,19 +273,28 @@
 				'##core#call '(#t)
 				(list k (make-node '##core#undefined '() '())) ) ) 
 			     (walk-generic n class params subs)) ]
-			[(and lval (eq? '##core#lambda (node-class lval)))
+			[(and lval
+			      (eq? '##core#lambda (node-class lval)))
 			 (let* ([lparams (node-parameters lval)]
 				[llist (third lparams)] )
 			   (decompose-lambda-list
 			    llist
 			    (lambda (vars argc rest)
 			      (let ([fid (first lparams)])
-				(cond [(and (test fid 'simple)
+				(cond [(and inline-locally 
+					    (test fid 'simple)
 					    (test var 'inlinable)
-					    (not (memq var not-inline-list))
-					    (or (memq var inline-list)
-						(< (fourth lparams) inline-max-size) ) )
-				       (debugging 'i "procedure inlinable" var fid (fourth lparams))
+					    (case (variable-mark var '##compiler#inline) 
+					      ((yes) #t)
+					      ((no) #f)
+					      (else 
+					       (< (fourth lparams) inline-max-size) ) ))
+				       (debugging 
+					'i
+					(if (node? (variable-mark var '##compiler#inline-global))
+					    "procedure can be inlined (globally)" 
+					    "procedure can be inlined")
+					var fid (fourth lparams))
 				       (check-signature var args llist)
 				       (debugging 'o "inlining procedure" var)
 				       (touch)
@@ -342,7 +311,9 @@
 						     (map walk (cons fun (append-reverse used args))) ) ]
 						   [(test (car vars) 'unused)
 						    (touch)
-						    (debugging 'o "removed unused parameter to known procedure" (car vars) var)
+						    (debugging
+						     'o "removed unused parameter to known procedure" 
+						     (car vars) var)
 						    (if (expression-has-side-effects? (car args) db)
 							(make-node
 							 'let
@@ -356,7 +327,7 @@
 							       (cons (car args) used) ) ] ) ) ) ]
 				      [(and (test fid 'explicit-rest)
 					    (not (memq n rest-consers)) ) ; make sure we haven't inlined rest-list already
-				       (let ([n (length llist)])
+				       (let ([n (llist-length llist)])
 					 (if (< (length args) n)
 					     (walk-generic n class params subs)
 					     (begin
@@ -392,8 +363,7 @@
 		    (touch)
 		    (make-node '##core#undefined '() '()) ]
 		   [(and (or (not (test var 'global))
-			     block-compilation
-			     (and export-list (not (memq var export-list))) )
+			     (not (variable-visible? var)))
 			 (not (test var 'references)) 
 			 (not (expression-has-side-effects? (first subs) db)) )
 		    (touch)
@@ -446,7 +416,7 @@
     (debugging 'p "pre-optimization phase...")
 
     ;; Handle '(if (not ...) ...)':
-    (if (test 'not 'standard-binding)
+    (if (intrinsic? 'not)
 	(for-each 
 	 (lambda (site)
 	   (let* ((n (cdr site))
@@ -484,45 +454,6 @@
 				     (touch) ) ) ) ) ) ) ) ) ) )
 	 (or (test 'not 'call-sites) '()) ) )
     
-    ;; Handle '(if (<func> <a> ...) ...)', where <func> never returns false:
-    (for-each
-     (lambda (varname)
-       (if (test varname 'standard-binding)
-	   (for-each
-	    (lambda (site)
-	      (let* ((n (cdr site))
-		     (subs (node-subexpressions n))
-		     (kont (first (node-parameters (second subs)))) 
-		     (krefs (test kont 'references)) 
-		     (lnode (and (not (test kont 'unknown)) (test kont 'value))) )
-		;; Call-site has side-effect-free arguments and a known continuation that has only one use?
-		(if (and lnode
-			 (eq? '##core#lambda (node-class lnode))
-			 krefs (= 1 (length krefs))
-			 (not (any (lambda (sn) (expression-has-side-effects? sn db)) (cddr subs))) )
-		    (let* ((llist (third (node-parameters lnode)))
-			   (body (first (node-subexpressions lnode))) )
-		      ;; Continuation has one parameter and contains an 'if' node?
-		      (if (and (proper-list? llist)
-			       (null? (cdr llist))
-			       (eq? 'if (node-class body)) )
-			  (let* ((var (car llist))
-				 (refs (test var 'references)) 
-				 (iftest (first (node-subexpressions body))) )
-			    ;; Parameter is used only once and is the test-argument?
-			    (if (and refs (= 1 (length refs))
-				     (eq? '##core#variable (node-class iftest))
-				     (eq? var (first (node-parameters iftest))) )
-				(let ((bodysubs (node-subexpressions body)))
-				  ;; Modify call-site to call continuation directly and swap branches
-				  ;;  in the conditional:
-				  (debugging 'o "removed call in test-context" varname)
-				  (node-parameters-set! n '(#t))
-				  (node-subexpressions-set! n (list (second subs) (qnode #t)))
-				  (touch) ) ) ) ) ) ) ) )
-	    (or (test varname 'call-sites) '()) ) ) )
-     side-effect-free-standard-bindings-that-never-return-false)
-
     (when (> removed-nots 0) (debugging 'o "Removed `not' forms" removed-nots))
     dirty) )
 
@@ -854,7 +785,7 @@
     ;; (eq?/eqv?/equal? <var> <var>) -> (quote #t)
     ;; (eq?/eqv?/equal? ...) -> (##core#inline <iop> ...)
     ((1) ; classargs = (<argc> <iop>)
-     (and (test name 'standard-binding)
+     (and (intrinsic? name)
 	  (or (and (= (length callargs) (first classargs))
 		   (let ((arg1 (first callargs))
 			 (arg2 (second callargs)) )
@@ -872,7 +803,7 @@
     ((2) ; classargs = (<argc> <iop> <safe> <iopv>)
      (and inline-substitutions-enabled
 	  (= (length callargs) (first classargs))
-	  (or (test name 'extended-binding) (test name 'standard-binding))
+	  (intrinsic? name)
 	  (or (third classargs) unsafe)
 	  (let ([arg1 (first callargs)]
 		[iopv (fourth classargs)] )
@@ -890,7 +821,7 @@
     ((3) ; classargs = (<var>)
      (and inline-substitutions-enabled
 	  (null? callargs)
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (make-node '##core#call '(#t) (list cont (varnode (first classargs)))) ) )
 
     ;; (<op> a b) -> (<primitiveop> a (quote <i>) b)
@@ -898,7 +829,7 @@
      (and inline-substitutions-enabled
 	  unsafe
 	  (= 2 (length callargs))
-	  (test name 'standard-binding)
+	  (intrinsic? name)
 	  (make-node '##core#call (list #f (first classargs))
 		     (list (varnode (first classargs))
 			   cont
@@ -910,8 +841,7 @@
     ((5) ; classargs = (<iop> <x> <numtype>)
      ;; - <numtype> may be #f
      (and inline-substitutions-enabled
-	  (or (test name 'extended-binding)
-	      (test name 'standard-binding) )
+	  (intrinsic? name)
 	  (= 1 (length callargs))
 	  (let ((ntype (third classargs)))
 	    (or (not ntype) (eq? ntype number-type)) )
@@ -926,7 +856,7 @@
       (and (or (third classargs) unsafe)
 	   inline-substitutions-enabled
 	   (= 1 (length callargs))
-	   (test name 'standard-binding)
+	   (intrinsic? name)
 	   (make-node '##core#call '(#t)
 		      (list cont
 			    (make-node '##core#inline (list (first classargs))
@@ -938,7 +868,7 @@
      (and (or (fourth classargs) unsafe)
 	  inline-substitutions-enabled
 	  (= (length callargs) (first classargs))
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (make-node '##core#call '(#t)
 		     (list cont
 			   (make-node '##core#inline (list (second classargs))
@@ -948,15 +878,14 @@
     ;; (<op> ...) -> <<call procedure <proc> with <classargs>, <cont> and <callargs> >>
     ((8) ; classargs = (<proc> ...)
      (and inline-substitutions-enabled
-	  (or (test name 'standard-binding)
-	      (test name 'extended-binding) )
+	  (intrinsic? name)
 	  ((first classargs) db classargs cont callargs) ) )
 
     ;; (<op> <x1> ...) -> (##core#inline "C_and" (##core#inline <iop> <x1> <x2>) ...)
     ;; (<op> [<x>]) -> (quote #t)
     ((9) ; classargs = (<iop-fixnum> <iop-flonum> <fixnum-safe> <flonum-safe>)
      (and inline-substitutions-enabled
-	  (test name 'standard-binding)
+	  (intrinsic? name)
 	  (if (< (length callargs) 2)
 	      (make-node '##core#call '(#t) (list cont (qnode #t)))
 	      (and (or (and unsafe (not (eq? number-type 'generic)))
@@ -983,7 +912,7 @@
     ((10) ; classargs = (<primitiveop> <i> <bvar> <safe>)
      (and inline-substitutions-enabled
 	  (or (fourth classargs) unsafe)
-	  (test name 'standard-binding)
+	  (intrinsic? name)
 	  (let ((n (length callargs)))
 	    (and (< 0 n 3)
 		 (make-node '##core#call (list #f (first classargs))
@@ -1000,7 +929,7 @@
      ;; <argc> may be #f.
      (and inline-substitutions-enabled
 	  (or (third classargs) unsafe)
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (let ([argc (first classargs)])
 	    (and (or (not argc)
 		     (= (length callargs) (first classargs)) )
@@ -1013,7 +942,7 @@
     ;; (<op> ...) -> (<primitiveop> ...)
     ((12) ; classargs = (<primitiveop> <safe> <maxargc>)
      (and inline-substitutions-enabled
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (or (second classargs) unsafe)
 	  (let ((n (length callargs)))
 	    (and (<= n (third classargs))
@@ -1026,7 +955,7 @@
     ;; (<op> ...) -> ((##core#proc <primitiveop>) ...)
     ((13) ; classargs = (<primitiveop> <safe>)
      (and inline-substitutions-enabled
-	  (or (test name 'extended-binding) (test name 'standard-binding))
+	  (intrinsic? name)
 	  (or (second classargs) unsafe)
 	  (let ((pname (first classargs)))
 	    (make-node '##core#call (if (pair? params) (cons #t (cdr params)) params)
@@ -1037,8 +966,7 @@
     ((14) ; classargs = (<numtype> <argc> <iop-safe> <iop-unsafe>)
      (and inline-substitutions-enabled
 	  (= (second classargs) (length callargs))
-	  (or (test name 'extended-binding)
-	      (test name 'standard-binding) )
+	  (intrinsic? name)
 	  (eq? number-type (first classargs))
 	  (or (fourth classargs) unsafe)
 	  (make-node
@@ -1055,8 +983,7 @@
      (and inline-substitutions-enabled
 	  (= 1 (length callargs))
 	  (or unsafe (fourth classargs))
-	  (or (test name 'extended-binding)
-	      (test name 'standard-binding) )
+	  (intrinsic? name)
 	  (cond ((eq? number-type (first classargs))
 		 (make-node '##core#call (list #t (third classargs))
 			    (cons* (varnode (third classargs)) cont callargs) ) )
@@ -1076,7 +1003,7 @@
 	   [w (fourth classargs)] )
        (and inline-substitutions-enabled
 	    (or (not argc) (= rargc argc))
-	    (or (test name 'extended-binding) (test name 'standard-binding))
+	    (intrinsic? name)
 	    (or (third classargs) unsafe)
 	    (make-node
 	     '##core#call '(#t)
@@ -1093,7 +1020,7 @@
     ((17) ; classargs = (<argc> <iop-safe> [<iop-unsafe>])
      (and inline-substitutions-enabled
 	  (= (length callargs) (first classargs))
-	  (or (test name 'extended-binding) (test name 'standard-binding))
+	  (intrinsic? name)
 	  (make-node
 	   '##core#call '(#t)
 	   (list cont
@@ -1107,7 +1034,7 @@
     ((18) ; classargs = (<null>)
      (and inline-substitutions-enabled
 	  (null? callargs)
-	  (or (test name 'extended-binding) (test name 'standard-binding))
+	  (intrinsic? name)
 	  (make-node '##core#call '(#t) (list cont (qnode (first classargs))) ) ) )
 
     ;; (<op>) -> <id>
@@ -1117,7 +1044,7 @@
     ;; - Remove "<id>" from arguments.
     ((19) ; classargs = (<id> <fixop> <ufixop> <fixmode>)
      (and inline-substitutions-enabled
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (let* ([id (first classargs)]
 		 [fixop (if unsafe (third classargs) (second classargs))]
 		 [callargs 
@@ -1146,7 +1073,7 @@
        (and (or (fourth classargs) unsafe)
 	    inline-substitutions-enabled
 	    (= n (first classargs))
-	    (or (test name 'standard-binding) (test name 'extended-binding))
+	    (intrinsic? name)
 	    (make-node
 	     '##core#call '(#t)
 	     (list cont
@@ -1164,7 +1091,7 @@
     ;; - Remove "<id>" from arguments.
     ((21) ; classargs = (<id> <fixop> <ufixop> <genop> <words>)
      (and inline-substitutions-enabled
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (let* ([id (first classargs)]
 		 [words (fifth classargs)]
 		 [genop (fourth classargs)]
@@ -1198,7 +1125,7 @@
 	   [w (fourth classargs)] )
        (and inline-substitutions-enabled
 	    (= rargc argc)
-	    (or (test name 'extended-binding) (test name 'standard-binding))
+	    (intrinsic? name)
 	    (or (third classargs) unsafe)
 	    (make-node
 	     '##core#call '(#t)
@@ -1219,7 +1146,7 @@
     ;;   quoted literal
     ((23) ; classargs = (<minargc> <primitiveop> <literal1>|<varable1> ...)
      (and inline-substitutions-enabled
-	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (intrinsic? name)
 	  (let ([argc (first classargs)])
 	    (and (>= (length callargs) (first classargs))
 		 (make-node 
@@ -1370,10 +1297,12 @@
 	     [klambdas '()] 
 	     [sites (get db fnvar 'call-sites)] 
 	     [ksites '()] )
-	(match params
-	  [(id _ (kvar vars ...) _)
-	   ;; Remove continuation argument:
-	   (set-car! (cddr params) vars)
+	(if (and (list? params) (= (length params) 4) (list? (caddr params)))
+	    (let ((id (car params))
+		  (kvar (caaddr params))
+		  (vars (cdaddr params)) )
+	      ;; Remove continuation argument:
+	      (set-car! (cddr params) vars)
 	   ;; Make "##core#direct_lambda":
 	   (node-class-set! n '##core#direct_lambda)
 	   ;; Transform recursive calls and remove unused continuations:
@@ -1467,8 +1396,8 @@
 		    (let ([vn (cdr h)])
 		      (node-parameters-set! vn (list (gensym)))
 		      (set-car! (node-subexpressions vn) (make-node '##core#undefined '() '())) ) )
-		  hoistable) ) ) ) ]
-	  [_ (bomb "invalid parameter list" params)] ) ) )
+		  hoistable) ) ) ) )
+	    (bomb "invalid parameter list" params))))
 
     (debugging 'p "direct leaf routine optimization pass...")
     (walk #f node #f)
@@ -1719,7 +1648,7 @@
 	 (lambda (gn body)
 	   (let* ([name (car gn)]
 		  [lval (get db name 'value)] )
-	     (set! block-globals (cons name block-globals))
+	     (hide-variable name)
 	     (decompose-lambda-list
 	      (first (node-parameters lval))
 	      (lambda (vars argc rest)
@@ -1825,7 +1754,7 @@
 	  (let ([al (collect-accessibles g2)])
 	    (when (debugging 'l "accessibles:") (pretty-print al))
 	    (debugging 'p "eliminating liftables by access-lists and non-liftable callees...")
-	    (let ([ls (eliminate3 (eliminate4 g2))]) ;(eliminate2 g2 al)))])
+	    (let ([ls (eliminate3 (eliminate4 g2))]) ;(eliminate2 g2 al)))]) - why isn't this used?
 	      (debugging 'o "liftable local procedures" (delay (unzip1 ls)))
 	      (debugging 'p "gathering extra parameters...")
 	      (let ([extra (compute-extra-variables ls)])
@@ -1836,3 +1765,183 @@
 		(remove-local-bindings! ls)
 		(debugging 'p "moving liftables to toplevel...")
 		(reconstruct! ls extra) ) ) ) ) ) ) ) )
+
+
+;;; Compiler macros (that operate in the expansion phase)
+
+(define compiler-syntax-statistics '())
+
+(set! ##sys#compiler-syntax-hook
+  (lambda (name result)
+    (let ((a (alist-ref name compiler-syntax-statistics eq? 0)))
+      (set! compiler-syntax-statistics
+	(alist-update! name (add1 a) compiler-syntax-statistics)))))
+
+(define (r-c-s names transformer #!optional (se '()))
+  (let ((t (cons (##sys#er-transformer transformer) se)))
+    (for-each
+     (lambda (name)
+       (##sys#put! name '##compiler#compiler-syntax t) )
+     (if (symbol? names) (list names) names) ) ) )
+
+(r-c-s 
+ '(for-each ##sys#for-each #%for-each)
+ (lambda (x r c)
+   (let ((%let (r 'let))
+	 (%if (r 'if))
+	 (%loop (r 'loop))
+	 (%lst (r 'lst))
+	 (%begin (r 'begin))
+	 (%pair? (r 'pair?)))
+     (if (and (memq 'for-each standard-bindings) ; we have to check this because the db (and thus 
+	      (= 3 (length x)))			 ; intrinsic marks) isn't set up yet
+	 `(,%let ,%loop ((,%lst ,(caddr x)))
+		 (,%if (,%pair? ,%lst)
+		       (,%begin
+			(,(cadr x) (##sys#slot ,%lst 0))
+			(##core#app ,%loop (##sys#slot ,%lst 1))) ) )
+	 x)))
+ `((pair? . ,(##sys#primitive-alias 'pair?))))
+
+(r-c-s
+ '(o #%o)
+ (lambda (x r c)
+   (if (and (fx> (length x) 1)
+	    (memq 'o extended-bindings) )
+       (let ((%tmp (r 'tmp)))
+	 `(,(r 'lambda) (,%tmp) ,(fold-right list %tmp (cdr x))))
+       x)))
+
+(let ((env `((display . ,(##sys#primitive-alias 'display)) ;XXX clean this up
+	     (write . ,(##sys#primitive-alias 'write))
+	     (fprintf . ,(##sys#primitive-alias 'fprintf))
+	     (number->string . ,(##sys#primitive-alias 'number->string))
+	     (write-char . ,(##sys#primitive-alias 'write-char))
+	     (open-output-string . ,(##sys#primitive-alias 'open-output-string))
+	     (get-output-string . ,(##sys#primitive-alias 'get-output-string)) ) ) )
+  (r-c-s
+   '(sprintf #%sprintf format #%format)
+   (lambda (x r c)
+     (let* ((out (gensym 'out))
+	    (code (compile-format-string 
+		   (if (memq (car x) '(sprintf #%sprintf))
+		       'sprintf
+		       'format)
+		   out 
+		   x
+		   (cdr x)
+		   r c)))
+       (if code
+	   `(,(r 'let) ((,out (,(r 'open-output-string))))
+	     ,code
+	     (,(r 'get-output-string) ,out))
+	   x)))
+   env)
+  (r-c-s
+   '(fprintf #%fprintf)
+   (lambda (x r c)
+     (if (>= (length x) 3)
+	 (let ((code (compile-format-string 
+		      'fprintf (cadr x) 
+		      x (cddr x)
+		      r c)))
+	   (if code
+	       code
+	       x))
+	 x))
+   env)
+  (r-c-s
+   '(printf #%printf)
+   (lambda (x r c)
+     (let ((code (compile-format-string 
+		  'printf '##sys#standard-output
+		  x (cdr x)
+		  r c)))
+       (if code
+	   code
+	   x)))
+   env))
+
+(define (compile-format-string func out x args r c)
+  (call/cc
+   (lambda (return)
+     (and (>= (length args) 1)
+	  (memq func extended-bindings)	; s.a.
+	  (or (string? (car args))
+	      (and (list? (car args))
+		   (c (r 'quote) (caar args))
+		   (string? (cadar args))))
+	  (let ((fstr (if (string? (car args)) (car args) (cadar args)))
+		(args (cdr args)))
+	    (define (fail ret? msg . args)
+	      (let ((ln (get-line x)))
+		(compiler-warning 
+		 'syntax
+		 "(~a) in format string ~s~a, ~?" 
+		 func fstr 
+		 (if ln (sprintf " in line ~a" ln) "")
+		 msg args) ) 
+	      (when ret? (return #f)))
+	    (let ((code '())
+		  (index 0)
+		  (len (string-length fstr)) 
+		  (%display (r 'display))
+		  (%write (r 'write))
+		  (%write-char (r 'write-char))
+		  (%out (r 'out))
+		  (%fprintf (r 'fprintf))
+		  (%let (r 'let))
+		  (%number->string (r 'number->string)))
+	      (define (fetch)
+		(let ((c (string-ref fstr index)))
+		  (set! index (fx+ index 1))
+		  c) )
+	      (define (next)
+		(if (null? args)
+		    (fail #t "too few arguments to formatted output procedure")
+		    (let ((x (car args)))
+		      (set! args (cdr args))
+		      x) ) )
+	      (define (endchunk chunk)
+		(when (pair? chunk)
+		  (push 
+		   (if (= 1 (length chunk))
+		       `(,%write-char ,(car chunk) ,%out)
+		       `(,%display ,(reverse-list->string chunk) ,%out)))))
+	      (define (push exp)
+		(set! code (cons exp code)))
+	      (let loop ((chunk '()))
+		(cond ((>= index len)
+		       (unless (null? args)
+			 (fail #f "too many arguments to formatted output procedure"))
+		       (endchunk chunk)
+		       `(,%let ((,%out ,out))
+			       ,@(reverse code)))
+		      (else
+		       (let ((c (fetch)))
+			 (if (eq? c #\~)
+			     (let ((dchar (fetch)))
+			       (endchunk chunk)
+			       (case (char-upcase dchar)
+				 ((#\S) (push `(,%write ,(next) ,%out)))
+				 ((#\A) (push `(,%display ,(next) ,%out)))
+				 ((#\C) (push `(,%write-char ,(next) ,%out)))
+				 ((#\B) (push `(,%display (,%number->string ,(next) 2) ,%out)))
+				 ((#\O) (push `(,%display (,%number->string ,(next) 8) ,%out)))
+				 ((#\X) (push `(,%display (,%number->string ,(next) 16) ,%out)))
+				 ((#\!) (push `(##sys#flush-output ,%out)))
+				 ((#\?)
+				  (let* ([fstr (next)]
+					 [lst (next)] )
+				    (push `(##sys#apply ,%fprintf ,%out ,fstr ,lst))))
+				 ((#\~) (push `(,write-char #\~ ,%out)))
+				 ((#\% #\N) (push `(,%write-char #\newline ,%out)))
+				 (else
+				  (if (char-whitespace? dchar)
+				      (let skip ((c (fetch)))
+					(if (char-whitespace? c)
+					    (skip (fetch))
+					    (set! index (sub1 index))))
+				      (fail #t "illegal format-string character `~c'" dchar) ) ) )
+			       (loop '()) )
+			     (loop (cons c chunk)))))))))))))
