@@ -1,7 +1,7 @@
 ;;;; posixunix.scm - Miscellaneous file- and process-handling routines
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -27,11 +27,12 @@
 
 (declare
   (unit posix)
-  (uses scheduler regex extras utils files)
+  (uses scheduler regex extras utils files ports)
   (disable-interrupts)
   (usual-integrations)
   (hide ##sys#stat group-member _get-groups _ensure-groups posix-error
-	##sys#terminal-check)
+	      ##sys#terminal-check
+	      check-time-vector)
   (foreign-declare #<<EOF
 #include <signal.h>
 #include <errno.h>
@@ -194,7 +195,6 @@ static C_TLS struct stat C_statbuf;
 #define C_close(fd)         C_fix(close(C_unfix(fd)))
 #define C_sleep             sleep
 
-#define C_putenv(s)         C_fix(putenv((char *)C_data_pointer(s)))
 #define C_stat(fn)          C_fix(stat((char *)C_data_pointer(fn), &C_statbuf))
 #define C_lstat(fn)         C_fix(lstat((char *)C_data_pointer(fn), &C_statbuf))
 #define C_fstat(f)          C_fix(fstat(C_unfix(f), &C_statbuf))
@@ -212,8 +212,10 @@ static C_TLS struct stat C_statbuf;
 #endif
 
 #ifdef C_GNU_ENV
+# define C_unsetenv(s)      (unsetenv((char *)C_data_pointer(s)), C_SCHEME_TRUE)
 # define C_setenv(x, y)     C_fix(setenv((char *)C_data_pointer(x), (char *)C_data_pointer(y), 1))
 #else
+# define C_unsetenv(s)      C_fix(putenv((char *)C_data_pointer(s)))
 static C_word C_fcall C_setenv(C_word x, C_word y) {
   char *sx = C_data_pointer(x),
        *sy = C_data_pointer(y);
@@ -356,66 +358,72 @@ static time_t timegm(struct tm *t)
 }
 #endif
 
-#define C_tm_set_08(v) \
-        (memset(&C_tm, 0, sizeof(struct tm)), \
-        C_tm.tm_sec = C_unfix(C_block_item(v, 0)), \
-        C_tm.tm_min = C_unfix(C_block_item(v, 1)), \
-        C_tm.tm_hour = C_unfix(C_block_item(v, 2)), \
-        C_tm.tm_mday = C_unfix(C_block_item(v, 3)), \
-        C_tm.tm_mon = C_unfix(C_block_item(v, 4)), \
-        C_tm.tm_year = C_unfix(C_block_item(v, 5)), \
-        C_tm.tm_wday = C_unfix(C_block_item(v, 6)), \
-        C_tm.tm_yday = C_unfix(C_block_item(v, 7)), \
-        C_tm.tm_isdst = (C_block_item(v, 8) != C_SCHEME_FALSE))
+#define cpy_tmvec_to_tmstc08(ptm, v) \
+    (memset((ptm), 0, sizeof(struct tm)), \
+    (ptm)->tm_sec = C_unfix(C_block_item((v), 0)), \
+    (ptm)->tm_min = C_unfix(C_block_item((v), 1)), \
+    (ptm)->tm_hour = C_unfix(C_block_item((v), 2)), \
+    (ptm)->tm_mday = C_unfix(C_block_item((v), 3)), \
+    (ptm)->tm_mon = C_unfix(C_block_item((v), 4)), \
+    (ptm)->tm_year = C_unfix(C_block_item((v), 5)), \
+    (ptm)->tm_wday = C_unfix(C_block_item((v), 6)), \
+    (ptm)->tm_yday = C_unfix(C_block_item((v), 7)), \
+    (ptm)->tm_isdst = (C_block_item((v), 8) != C_SCHEME_FALSE))
 
-#define C_tm_set_9(v) \
-        (C_tm.tm_gmtoff = C_unfix(C_block_item(v, 9)))
+#define cpy_tmvec_to_tmstc9(ptm, v) \
+    (((struct tm *)ptm)->tm_gmtoff = C_unfix(C_block_item((v), 9)))
 
-#define C_tm_get_08(v) \
-        (C_set_block_item(v, 0, C_fix(C_tm.tm_sec)), \
-        C_set_block_item(v, 1, C_fix(C_tm.tm_min)), \
-        C_set_block_item(v, 2, C_fix(C_tm.tm_hour)), \
-        C_set_block_item(v, 3, C_fix(C_tm.tm_mday)), \
-        C_set_block_item(v, 4, C_fix(C_tm.tm_mon)), \
-        C_set_block_item(v, 5, C_fix(C_tm.tm_year)), \
-        C_set_block_item(v, 6, C_fix(C_tm.tm_wday)), \
-        C_set_block_item(v, 7, C_fix(C_tm.tm_yday)), \
-        C_set_block_item(v, 8, (C_tm.tm_isdst ? C_SCHEME_TRUE : C_SCHEME_FALSE)))
+#define cpy_tmstc08_to_tmvec(v, ptm) \
+    (C_set_block_item((v), 0, C_fix(((struct tm *)ptm)->tm_sec)), \
+    C_set_block_item((v), 1, C_fix((ptm)->tm_min)), \
+    C_set_block_item((v), 2, C_fix((ptm)->tm_hour)), \
+    C_set_block_item((v), 3, C_fix((ptm)->tm_mday)), \
+    C_set_block_item((v), 4, C_fix((ptm)->tm_mon)), \
+    C_set_block_item((v), 5, C_fix((ptm)->tm_year)), \
+    C_set_block_item((v), 6, C_fix((ptm)->tm_wday)), \
+    C_set_block_item((v), 7, C_fix((ptm)->tm_yday)), \
+    C_set_block_item((v), 8, ((ptm)->tm_isdst ? C_SCHEME_TRUE : C_SCHEME_FALSE)))
 
-#define C_tm_get_9(v) \
-        (C_set_block_item(v, 9, C_fix(C_tm.tm_gmtoff)))
+#define cpy_tmstc9_to_tmvec(v, ptm) \
+    (C_set_block_item((v), 9, C_fix((ptm)->tm_gmtoff)))
+
+#define C_tm_set_08(v)  cpy_tmvec_to_tmstc08( &C_tm, (v) )
+#define C_tm_set_9(v)   cpy_tmvec_to_tmstc9( &C_tm, (v) )
+
+#define C_tm_get_08(v)  cpy_tmstc08_to_tmvec( (v), &C_tm )
+#define C_tm_get_9(v)   cpy_tmstc9_to_tmvec( (v), &C_tm )
 
 #if !defined(C_GNU_ENV) || defined(__CYGWIN__) || defined(__uClinux__)
 
 static struct tm *
-C_tm_set (C_word v)
+C_tm_set( C_word v )
 {
-  C_tm_set_08 (v);
+  C_tm_set_08( v );
   return &C_tm;
 }
 
 static C_word
-C_tm_get (C_word v)
+C_tm_get( C_word v )
 {
-  C_tm_get_08 (v);
+  C_tm_get_08( v );
   return v;
 }
 
 #else
 
 static struct tm *
-C_tm_set (C_word v)
+C_tm_set( C_word v )
 {
-  C_tm_set_08 (v);
-  C_tm_set_9 (v);
+  C_tm_set_08( v );
+  C_tm_set_9( v );
   return &C_tm;
 }
 
 static C_word
-C_tm_get (C_word v)
+C_tm_get( C_word v )
 {
-  C_tm_get_08 (v);
-  C_tm_get_9 (v);
+  C_tm_get_08( v );
+  C_tm_get_9( v );
   return v;
 }
 
@@ -474,11 +482,11 @@ EOF
     (no-bound-checks)
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure
-     string-match glob->regexp regexp make-anchored-pattern
+     string-match glob->regexp regexp 
      ##sys#thread-yield! ##sys#make-string
      ##sys#make-port ##sys#file-info ##sys#update-errno ##sys#fudge ##sys#make-c-string ##sys#check-port
      ##sys#error ##sys#signal-hook ##sys#peek-unsigned-integer make-pathname glob directory?
-     pathname-file process-fork file-close duplicate-fileno process-execute getenv
+     pathname-file process-fork file-close duplicate-fileno process-execute get-environment-variable
      make-string make-input-port make-output-port ##sys#thread-block-for-i/o create-pipe
      process-wait pathname-strip-directory ##sys#expand-home-path directory
      decompose-pathname ##sys#cons-flonum ##sys#decode-seconds ##sys#null-pointer ##sys#pointer->address
@@ -488,22 +496,7 @@ EOF
      make-nonblocking-input-port make-nonblocking-output-port 
      canonical-path) ) ] )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "posix.exports")) ] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'posix)
 
@@ -772,7 +765,7 @@ EOF
       (posix-error #:file-error loc "cannot access file" file) ) ) )
 
 (define (file-stat f . link)
-  (##sys#stat f (:optional link #f) 'file-stat)
+  (##sys#stat f (optional link #f) 'file-stat)
   (vector _stat_st_ino _stat_st_mode _stat_st_nlink
           _stat_st_uid _stat_st_gid _stat_st_size
           _stat_st_atime _stat_st_ctime _stat_st_mtime
@@ -796,67 +789,71 @@ EOF
   (##sys#stat fname #t 'symbolic-link?)
   (foreign-value "C_islink" bool) )
 
-(define (stat-regular? fname)
+(define (stat-regular? fname)		; DEPRECATED
     (##sys#check-string fname 'stat-regular?)
     (##sys#stat fname #f 'stat-regular?)
     (foreign-value "C_isreg" bool))
 
-(define (stat-directory? fname)
+(define (stat-directory? fname)		; DEPRECATED
     (##sys#check-string fname 'stat-directory?)
     (##sys#stat fname #f 'stat-directory?)
     (foreign-value "C_isdir" bool))
 
-(define (stat-char-device? fname)
-    (##sys#check-string fname 'stat-char-device?)
-    (##sys#stat fname #f 'stat-char-device?)
+(define (character-device? fname)
+    (##sys#check-string fname 'character-device?)
+    (##sys#stat fname #f 'character-device?)
     (foreign-value "C_ischr" bool))
 
-(define (stat-block-device? fname)
-    (##sys#check-string fname 'stat-block-device?)
-    (##sys#stat fname #f 'stat-block-device?)
+(define stat-char-device? character-device?) ; DEPRECATED
+
+(define (block-device? fname)
+    (##sys#check-string fname 'block-device?)
+    (##sys#stat fname #f 'block-device?)
     (foreign-value "C_isblk" bool))
 
-(define (stat-fifo? fname)
+(define stat-block-device? block-device?) ; DEPRECATED
+
+(define (fifo? fname)
     (##sys#check-string fname 'stat-fifo?)
     (##sys#stat fname #f 'stat-fifo?)
     (foreign-value "C_isfifo" bool))
 
-(define (stat-symlink? fname)
-    (##sys#check-string fname 'stat-symlink?)
-    (##sys#stat fname #t 'stat-symlink?)
-    (foreign-value "C_islink" bool))
+(define stat-fifo? fifo?)		; DEPRECATED
+(define stat-symlink? symbolic-link?)	; DEPRECATED
 
-(define (stat-socket? fname)
-    (##sys#check-string fname 'stat-socket?)
-    (##sys#stat fname #f 'stat-socket?)
-    (foreign-value "C_issock" bool))
+(define (socket? fname)
+  (##sys#check-string fname 'socket?)
+  (##sys#stat fname #f 'socket?)
+  (foreign-value "C_issock" bool))
 
-(define set-file-position!              ; DEPRECATED
-  (lambda (port pos . whence)
-    (let ([whence (if (pair? whence) (car whence) _seek_set)])
-      (##sys#check-exact pos 'set-file-position!)
-      (##sys#check-exact whence 'set-file-position!)
-      (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
-      (unless (cond [(port? port)
-                     (and (eq? (##sys#slot port 7) 'stream)
-                          (##core#inline "C_fseek" port pos whence) ) ]
-                    [(fixnum? port) (##core#inline "C_lseek" port pos whence)]
-                    [else (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)] )
-        (posix-error #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
+(define stat-socket? socket?)		; DEPRECATED
+
+(define set-file-position!
+   (lambda (port pos . whence)
+     (let ([whence (if (pair? whence) (car whence) _seek_set)])
+       (##sys#check-exact pos 'set-file-position!)
+       (##sys#check-exact whence 'set-file-position!)
+       (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
+       (unless (cond [(port? port)
+		      (and (eq? (##sys#slot port 7) 'stream)
+			   (##core#inline "C_fseek" port pos whence) ) ]
+		     [(fixnum? port) (##core#inline "C_lseek" port pos whence)]
+		     [else (##sys#signal-hook #:type-error 'set-file-position! "invalid file" port)] )
+	 (posix-error #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
 
 (define file-position
   (getter-with-setter
    (lambda (port)
-    (let ([pos (cond [(port? port)
-                      (if (eq? (##sys#slot port 7) 'stream)
-                          (##core#inline "C_ftell" port)
-                          -1) ]
-                     [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
-                     [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
-      (when (fx< pos 0)
-        (posix-error #:file-error 'file-position "cannot retrieve file position of port" port) )
-      pos) )
-   set-file-position!) )
+     (let ([pos (cond [(port? port)
+		       (if (eq? (##sys#slot port 7) 'stream)
+			   (##core#inline "C_ftell" port)
+			   -1) ]
+		      [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
+		      [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
+       (when (fx< pos 0)
+	 (posix-error #:file-error 'file-position "cannot retrieve file position of port" port) )
+       pos) )
+   set-file-position!) )		; doesn't accept WHENCE
 
 
 ;;; Directory stuff:
@@ -892,10 +889,8 @@ EOF
   (lambda (name #!optional parents?)
     (##sys#check-string name 'create-directory)
     (if parents?
-        (create-directory-helper-parents (canonical-path name))
-        (create-directory-helper (canonical-path name)))))
-;    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string (##sys#expand-home-path name))))
-;      (posix-error #:file-error 'create-directory "cannot create directory" name) ) ) )
+        (create-directory-helper-parents name)
+        (create-directory-helper name))))
 
 (define change-directory
   (lambda (name)
@@ -954,8 +949,7 @@ EOF
                 (##sys#substring buffer 0 len)
                 (posix-error #:file-error 'current-directory "cannot retrieve current directory") ) ) ) ) ) )
 
-
-(define canonical-path
+(define canonical-path			; DEPRECATED
     (let ((null?      null?)
           (char=?     char=?)
           (string=?   string=?)
@@ -965,7 +959,7 @@ EOF
           (sappend    string-append)
           (isperse    (cut string-intersperse <> "/"))
           (sep?       (lambda (c) (or (char=? #\/ c) (char=? #\\ c))))
-          (getenv     getenv)
+          (get-environment-variable     get-environment-variable)
           (user       current-user-name)
           (cwd        (let ((cw   current-directory))
                           (lambda ()
@@ -983,7 +977,7 @@ EOF
                              ((and (char=? #\~ (sref path 0))
                                    (sep? (sref path 1)))
                                  (sappend
-                                     (or (getenv "HOME")
+                                     (or (get-environment-variable "HOME")
                                          (sappend "/home/" (user)))
                                      (##sys#substring path 1
                                          (##sys#size path))))
@@ -1259,16 +1253,13 @@ EOF
           _uname-version
           _uname-machine) ) )
 
-(define set-user-id!                  ; DEPRECATED
-  (lambda (id)
-    (when (fx< (##core#inline "C_setuid" id) 0)
-      (##sys#update-errno)
-      (##sys#error 'set-user-id! "cannot set user ID" id) ) ) )
-
 (define current-user-id
   (getter-with-setter
    (foreign-lambda int "C_getuid")
-   set-user-id!) )
+   (lambda (id)
+     (when (fx< (##core#inline "C_setuid" id) 0)
+       (##sys#update-errno)
+       (##sys#error 'set-user-id! "cannot set user ID" id) ) ) ) )
 
 (define current-effective-user-id
   (getter-with-setter
@@ -1279,16 +1270,13 @@ EOF
       (##sys#error 
 	 'effective-user-id!-setter "cannot set effective user ID" id) ) ) ) )
 
-(define set-group-id!                 ; DEPRECATED
-  (lambda (id)
-    (when (fx< (##core#inline "C_setgid" id) 0)
-      (##sys#update-errno)
-      (##sys#error 'set-user-id! "cannot set group ID" id) ) ) )
-
 (define current-group-id
   (getter-with-setter
    (foreign-lambda int "C_getgid")
-   set-group-id!) )
+   (lambda (id)
+    (when (fx< (##core#inline "C_setgid" id) 0)
+      (##sys#update-errno)
+      (##sys#error 'set-user-id! "cannot set group ID" id) ) ) ) )
 
 (define current-effective-group-id
   (getter-with-setter 
@@ -1510,13 +1498,6 @@ EOF
       (##sys#error 'create-session "cannot create session") )
     a) )
 
-(define (set-process-group-id! pid pgid) ; DEPRECATED
-  (##sys#check-exact pid 'set-process-group-id!)
-  (##sys#check-exact pgid 'set-process-group-id!)
-  (when (fx< (##core#inline "C_setpgid" pid pgid) 0)
-    (##sys#update-errno)
-    (##sys#error 'set-process-group-id! "cannot set process group ID" pid pgid) ) )
-
 (define process-group-id
   (getter-with-setter
    (lambda (pid)
@@ -1525,8 +1506,14 @@ EOF
        (when (fx< a 0)
          (##sys#update-errno)
          (##sys#error 'process-group-id "cannot retrieve process group ID" pid) )
-    a) )
-   set-process-group-id!) )
+       a))
+   (lambda (pid pgid)
+     (##sys#check-exact pid 'set-process-group-id!)
+     (##sys#check-exact pgid 'set-process-group-id!)
+     (when (fx< (##core#inline "C_setpgid" pid pgid) 0)
+       (##sys#update-errno)
+       (##sys#error 'set-process-group-id! "cannot set process group ID" pid pgid) ) ) ) )
+
 
 ;;; Hard and symbolic links:
 
@@ -1545,13 +1532,16 @@ EOF
 
 (define read-symbolic-link
   (let ([substring substring]
-      [buf (make-string (fx+ _filename_max 1))] )
-    (lambda (fname)
+        [buf (make-string (fx+ _filename_max 1))] )
+    (lambda (fname #!optional canonicalize)
       (##sys#check-string fname 'read-symbolic-link)
       (let ([len (##core#inline "C_readlink" (##sys#make-c-string (##sys#expand-home-path fname)) buf)])
       (when (fx< len 0)
         (posix-error #:file-error 'read-symbolic-link "cannot read symbolic link" fname) )
-      (substring buf 0 len) ) ) ) )
+      (let ((pathname (substring buf 0 len)))
+        (if (and canonicalize (symbolic-link? pathname))
+            (read-symbolic-link pathname 'canonicalize)
+            pathname ) ) ) ) ) )
 
 (define file-link
   (let ([link (foreign-lambda int "link" c-string c-string)])
@@ -1632,8 +1622,12 @@ EOF
         (let (
             [ready?
               (lambda ()
-                (when (fx= -1 (##sys#file-select-one fd))
-                  (posix-error #:file-error loc "cannot select" fd nam) ) )]
+                (let ((res (##sys#file-select-one fd)))
+                  (if (fx= -1 res)
+                      (if (fx= _errno _ewouldblock)
+                          #f
+                          (posix-error #:file-error loc "cannot select" fd nam))
+                      (fx= 1 res))))]
             [peek
               (lambda ()
                 (if (fx>= bufpos buflen)
@@ -1874,6 +1868,7 @@ EOF
           (fx= 3 (##sys#slot v 4))
           (posix-error #:file-error 'fifo? "file does not exist" filename) ) ) ) )
 
+
 ;;; Environment access:
 
 (define setenv
@@ -1885,10 +1880,10 @@ EOF
 
 (define (unsetenv var)
   (##sys#check-string var 'unsetenv)
-  (##core#inline "C_putenv" (##sys#make-c-string var))
+  (##core#inline "C_unsetenv" (##sys#make-c-string var))
   (##core#undefined) )
 
-(define current-environment
+(define get-environment-variables
   (let ([get (foreign-lambda c-string "C_getenventry" int)])
     (lambda ()
       (let loop ([i 0])
@@ -1902,6 +1897,9 @@ EOF
                     (scan (fx+ j 1)) ) )
               '() ) ) ) ) ) )
 
+(define current-environment get-environment-variables) ; DEPRECATED
+
+
 ;;; Memory mapped I/O:
 
 (define-foreign-variable _prot_read int "PROT_READ")
@@ -1955,6 +1953,11 @@ EOF
 
 ;;; Time related things:
 
+(define (check-time-vector loc tm)
+  (##sys#check-vector tm loc)
+  (when (fx< (##sys#size tm) 10)
+    (##sys#error loc "time vector too short" tm) ) )
+
 (define (seconds->local-time secs)
   (##sys#check-number secs 'seconds->local-time)
   (##sys#decode-seconds secs #f) )
@@ -1966,6 +1969,7 @@ EOF
 (define seconds->string
   (let ([ctime (foreign-lambda c-string "C_ctime" integer)])
     (lambda (secs)
+      (##sys#check-number secs 'seconds->string)
       (let ([str (ctime secs)])
         (if str
             (##sys#substring str 0 (fx- (##sys#size str) 1))
@@ -1975,8 +1979,7 @@ EOF
   (let ([asctime (foreign-lambda c-string "C_asctime" scheme-object)]
         [strftime (foreign-lambda c-string "C_strftime" scheme-object scheme-object)])
     (lambda (tm #!optional fmt)
-      (##sys#check-vector tm 'time->string)
-      (when (fx< (##sys#size tm) 10) (##sys#error 'time->string "time vector too short" tm))
+      (check-time-vector 'time->string tm)
       (if fmt
           (begin
             (##sys#check-string fmt 'time->string)
@@ -1995,15 +1998,13 @@ EOF
       (strptime (##sys#make-c-string tim) (##sys#make-c-string fmt) (make-vector 10 #f)) ) ) )
 
 (define (local-time->seconds tm)
-  (##sys#check-vector tm 'local-time->seconds)
-  (when (fx< (##sys#size tm) 10) (##sys#error 'local-time->seconds "time vector too short" tm))
+  (check-time-vector 'local-time->seconds tm)
   (if (##core#inline "C_mktime" tm)
       (##sys#cons-flonum)
       (##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) ) )
 
 (define (utc-time->seconds tm)
-  (##sys#check-vector tm 'utc-time->seconds)
-  (when (fx< (##sys#size tm) 10) (##sys#error 'utc-time->seconds "time vector too short" tm))
+  (check-time-vector 'utc-time->seconds tm)
   (if (##core#inline "C_timegm" tm)
       (##sys#cons-flonum)
       (##sys#error 'utc-time->seconds "cannot convert time vector to seconds" tm) ) )
@@ -2098,7 +2099,6 @@ EOF
 
 (define glob
   (let ([regexp regexp]
-        [make-anchored-pattern make-anchored-pattern]
         [string-match string-match]
         [glob->regexp glob->regexp]
         [directory directory]
@@ -2110,8 +2110,7 @@ EOF
             '()
             (let ([path (car paths)])
               (let-values ([(dir fil ext) (decompose-pathname path)])
-                (let* ([fnpatt (glob->regexp (make-pathname #f (or fil "*") ext))]
-                       [patt (make-anchored-pattern fnpatt)]
+                (let* ([patt (glob->regexp (make-pathname #f (or fil "*") ext))]
                        [rx (regexp patt)])
                   (let loop ([fns (directory (or dir ".") #t)])
                     (cond [(null? fns) (conc-loop (cdr paths))]
@@ -2205,15 +2204,14 @@ EOF
       (when (fx= r -1) (posix-error #:process-error 'process-signal "could not send signal to process" id sig) ) ) ) ) )
 
 (define (##sys#shell-command)
-  (or (getenv "SHELL") "/bin/sh") )
+  (or (get-environment-variable "SHELL") "/bin/sh") )
 
 (define (##sys#shell-command-arguments cmdlin)
   (list "-c" cmdlin) )
 
 (define process-run
   (let ([process-fork process-fork]
-        [process-execute process-execute]
-        [getenv getenv] )
+        [process-execute process-execute])
     (lambda (f . args)
       (let ([args (if (pair? args) (car args) #f)]
             [pid (process-fork)] )
