@@ -3,7 +3,7 @@
 ; By Sergey Khorev
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -65,11 +65,12 @@
 
 (declare
   (unit posix)
-  (uses scheduler regex extras utils files)
+  (uses scheduler regex extras utils files ports)
   (disable-interrupts)
   (usual-integrations)
-  (hide ##sys#stat close-handle posix-error
-	$quote-args-list $exec-setup $exec-teardown)
+  (hide ##sys#stat posix-error
+	$quote-args-list $exec-setup $exec-teardown
+	check-time-vector)
   (foreign-declare #<<EOF
 #ifndef WIN32_LEAN_AND_MEAN
 # define WIN32_LEAN_AND_MEAN
@@ -924,22 +925,7 @@ EOF
      ##sys#error ##sys#signal-hook ##sys#peek-unsigned-integer ##sys#process
      ##sys#peek-fixnum ##sys#make-structure ##sys#check-structure ##sys#enable-interrupts) ) ] )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "posix.exports"))] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'posix)
 
@@ -1125,26 +1111,17 @@ EOF
              (lambda (fname)
                  (##sys#check-string fname name)
                  #f))))
-    (set! stat-regular? regular-file?)
-    (set! stat-directory? (stat-type 'stat-directory?))
-    (set! stat-char-device? (stat-type 'stat-char-device?))
-    (set! stat-block-device? (stat-type 'stat-block-device?))
-    (set! stat-fifo? (stat-type 'stat-fifo?))
-    (set! stat-symlink? (stat-type 'stat-symlink?))
-    (set! stat-socket? (stat-type 'stat-socket?)))
-
-(define file-position
-  (lambda (port)
-    (let ([pos (cond [(port? port)
-		      (if (eq? (##sys#slot port 7) 'stream)
-			  (##core#inline "C_ftell" port)
-			  -1) ]
-		     [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
-		     [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
-      (when (fx< pos 0)
-	(##sys#update-errno)
-	(##sys#signal-hook #:file-error 'file-position "cannot retrieve file position of port" port) )
-      pos) ) )
+    (set! stat-regular? regular-file?)	; DEPRECATED
+    (set! stat-directory? (stat-type 'stat-directory?)) ; DEPRECATED
+    (set! stat-device? (stat-type 'stat-char-device?))	; DEPRECATED
+    (set! character-device? (stat-type 'character-device?))
+    (set! block-device? (stat-type 'block-device?))
+    (set! stat-block-device? (stat-type 'stat-block-device?)) ; DEPRECATED
+    (set! stat-fifo? (stat-type 'stat-fifo?))		      ; DEPRECATED
+    (set! fifo? (stat-type 'fifo?))
+    (set! stat-symlink? (stat-type 'stat-symlink?)) ; DEPRECATED
+    (set! socket? (stat-type 'socket?))
+    (set! stat-socket? (stat-type 'stat-socket?))) ; DEPRECATED
 
 (define set-file-position!
   (lambda (port pos . whence)
@@ -1160,6 +1137,21 @@ EOF
 	(##sys#update-errno)
 	(##sys#signal-hook #:file-error 'set-file-position! "cannot set file position" port pos) ) ) ) )
 
+(define file-position
+  (getter-with-setter
+   (lambda (port)
+     (let ([pos (cond [(port? port)
+		       (if (eq? (##sys#slot port 7) 'stream)
+			   (##core#inline "C_ftell" port)
+			   -1) ]
+		      [(fixnum? port) (##core#inline "C_lseek" port 0 _seek_cur)]
+		      [else (##sys#signal-hook #:type-error 'file-position "invalid file" port)] ) ] )
+       (when (fx< pos 0)
+	 (##sys#update-errno)
+	 (##sys#signal-hook #:file-error 'file-position "cannot retrieve file position of port" port) )
+       pos) )
+   set-file-position!) )		; doesn't accept WHENCE argument
+
 
 ;;; Directory stuff:
 
@@ -1666,6 +1658,7 @@ EOF
 	(##sys#signal-hook #:file-error 'duplicate-fileno "cannot duplicate file descriptor" old) )
       fd) ) )
 
+
 ;;; Environment access:
 
 (define setenv
@@ -1680,7 +1673,7 @@ EOF
   (##core#inline "C_putenv" (##sys#make-c-string var))
   (##core#undefined) )
 
-(define current-environment
+(define get-environment-variables
   (let ([get (foreign-lambda c-string "C_getenventry" int)]
 	[substring substring] )
     (lambda ()
@@ -1693,8 +1686,15 @@ EOF
 		    (scan (fx+ j 1)) ) )
 	      '() ) ) ) ) ) )
 
+(define current-environment get-environment-variables) ; DEPRECATED
+
 ;;; Time related things:
 
+(define (check-time-vector loc tm)
+  (##sys#check-vector tm loc)
+  (when (fx< (##sys#size tm) 10)
+    (##sys#error loc "time vector too short" tm) ) )
+
 (define (seconds->local-time secs)
   (##sys#check-number secs 'seconds->local-time)
   (##sys#decode-seconds secs #f) )
@@ -1715,8 +1715,7 @@ EOF
   (let ([asctime (foreign-lambda c-string "C_asctime" scheme-object)]
         [strftime (foreign-lambda c-string "C_strftime" scheme-object scheme-object)])
     (lambda (tm #!optional fmt)
-      (##sys#check-vector tm 'time->string)
-      (when (fx< (##sys#size tm) 10) (##sys#error 'time->string "time vector too short" tm))
+      (check-time-vector 'time->string tm)
       (if fmt
           (begin
             (##sys#check-string fmt 'time->string)
@@ -1728,15 +1727,14 @@ EOF
                 (##sys#error 'time->string "cannot convert time vector to string" tm) ) ) ) ) ) )
 
 (define (local-time->seconds tm)
-  (##sys#check-vector tm 'local-time->seconds)
-  (when (fx< (##sys#size tm) 10) (##sys#error 'local-time->seconds "time vector too short" tm))
+  (check-time-vector 'local-time->seconds tm)
   (if (##core#inline "C_mktime" tm)
       (##sys#cons-flonum)
       (##sys#error 'local-time->seconds "cannot convert time vector to seconds" tm) ) )
 
 (define local-timezone-abbreviation
   (foreign-lambda* c-string ()
-   "char *z = (daylight ? _tzname[1] : _tzname[0]);"
+   "char *z = (_daylight ? _tzname[1] : _tzname[0]);\n"
    "return(z);") )
 
 ;;; Other things:
@@ -1746,6 +1744,10 @@ EOF
     (lambda code
       (ex0 (if (pair? code) (car code) 0)) ) ) )
 
+(define (terminal-port? port)
+  (##sys#check-port port 'terminal-port?)
+  #f)
+
 (define-foreign-variable _iofbf int "_IOFBF")
 (define-foreign-variable _iolbf int "_IOLBF")
 (define-foreign-variable _ionbf int "_IONBF")
@@ -1771,7 +1773,6 @@ EOF
 
 (define glob
   (let ([regexp regexp]
-	[make-anchored-pattern make-anchored-pattern]
 	[string-match string-match]
 	[glob->regexp glob->regexp]
 	[directory directory]
@@ -1783,8 +1784,7 @@ EOF
 	    '()
 	    (let ([path (car paths)])
 	      (let-values ([(dir fil ext) (decompose-pathname path)])
-		(let* ([fnpatt (glob->regexp (make-pathname #f (or fil "*") ext))]
-		       [patt (make-anchored-pattern fnpatt)]
+		(let* ([patt (glob->regexp (make-pathname #f (or fil "*") ext))]
 		       [rx (regexp patt)])
 		  (let loop ([fns (directory (or dir ".") #t)])
 		    (cond [(null? fns) (conc-loop (cdr paths))]
@@ -1886,7 +1886,7 @@ EOF
 (define-foreign-variable _shlcmd c-string "C_shlcmd")
 
 (define (##sys#shell-command)
-  (or (getenv "COMSPEC")
+  (or (get-environment-variable "COMSPEC")
       (if (##core#inline "C_get_shlcmd")
 	  _shlcmd
 	  (begin
@@ -1898,7 +1898,7 @@ EOF
 
 (define process-run
   (let ([process-spawn process-spawn]
-	[getenv getenv] )
+	[get-environment-variable get-environment-variable] )
     (lambda (f . args)
       (let ([args (if (pair? args) (car args) #f)])
 	(if args
@@ -1910,9 +1910,6 @@ EOF
 (define-foreign-variable _wr0 int "C_wr0_")
 (define-foreign-variable _rd1 int "C_rd1_")
 
-(define close-handle
-  (foreign-lambda int "close_handle" bool))
-
 ; from original by Mejedi
 ;; ##sys#process
 ; loc		 caller procedure symbol
@@ -2074,9 +2071,11 @@ EOF
 
 ;;; unimplemented stuff:
 
-(define-macro (define-unimplemented name)
-  `(define (,name . _)
-     (error ',name (##core#immutable '"this function is not available on this platform")) ) )
+(define-syntax define-unimplemented
+  (syntax-rules ()
+    [(_ ?name)
+     (define (?name . _)
+       (error '?name (##core#immutable '"this function is not available on this platform")) ) ] ) )
 
 (define-unimplemented change-file-owner)
 (define-unimplemented create-fifo)
@@ -2116,7 +2115,6 @@ EOF
 (define-unimplemented signal-masked?)
 (define-unimplemented signal-unmask!)
 (define-unimplemented terminal-name)
-(define-unimplemented terminal-port?)
 (define-unimplemented terminal-size)
 (define-unimplemented unmap-file-from-memory)
 (define-unimplemented user-information)
