@@ -1,7 +1,7 @@
 ;;;; profiler.scm - Support code for profiling applications
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -41,7 +41,7 @@
     (no-bound-checks) ) ] )
 
 (foreign-declare #<<EOF
-#if !defined(_MSC_VER) && !defined(__DJGPP__) && !defined(__MWERKS__)
+#if !defined(_MSC_VER)
 # include <unistd.h>
 #endif
 EOF
