@@ -1,7 +1,7 @@
 ; scheduler.scm - Basic scheduler for multithreading
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -30,7 +30,6 @@
   (unit scheduler)
   (disable-interrupts)
   (usual-integrations)
-  (emit-exports "scheduler.exports")
   (disable-warning var)
   (hide ##sys#ready-queue-head ##sys#ready-queue-tail ##sys#timeout-list
 	##sys#update-thread-state-buffer ##sys#restore-thread-state-buffer
@@ -87,9 +86,9 @@ EOF
   (declare (unsafe)) ] )
 
 
-(define-macro (dbg . args) #f)
-#;(define-macro (dbg . args)
-  `(print "DBG: " ,@args) )
+(define-syntax dbg
+  (syntax-rules ()
+    ((_ . _) #f))) 
 
 
 (define (##sys#schedule)
@@ -212,6 +211,18 @@ EOF
 
 (define ##sys#timeout-list '())
 
+(define (##sys#remove-from-timeout-list t)
+  (let loop ((l ##sys#timeout-list) (prev #f))
+    (if (null? l)
+	l
+	(let ((h (##sys#slot l 0))
+	      (r (##sys#slot l 1)))
+	  (if (eq? (##sys#slot h 1) t)
+	      (if prev
+		  (set-cdr! prev r)
+		  (set! ##sys#timeout-list r))
+	      (loop r l))))))
+
 (define (##sys#thread-block-for-timeout! t tm)
   (dbg t " blocks for " tm)
   ;; This should really use a balanced tree:
@@ -241,6 +252,7 @@ EOF
   (##sys#setislot t 4 #f)
   (##sys#setislot t 11 #f)
   (##sys#setislot t 8 '())
+  (##sys#remove-from-timeout-list t)
   (let ([rs (##sys#slot t 12)])
     (unless (null? rs)
       (for-each
@@ -407,11 +419,27 @@ EOF
 
 
 ;;; Get list of all threads that are ready or waiting for timeout or waiting for I/O:
+;
+; (contributed by Joerg Wittenberger)
 
-(define (##sys#all-threads)
-  (append ##sys#ready-queue-head
-          (apply append (map cdr ##sys#fd-list))
-          (map cdr ##sys#timeout-list)))
+(define (##sys#all-threads #!optional
+			   (cns (lambda (queue arg val init)
+				  (cons val init)))
+			   (init '()))
+  (let loop ((l ##sys#ready-queue-head) (i init))
+    (if (pair? l)
+	(loop (cdr l) (cns 'ready #f (car l) i))
+	(let loop ((l ##sys#fd-list) (i i))
+	  (if (pair? l)
+	      (loop (cdr l)
+		    (let ((fd (caar l)))
+		      (let loop ((l (cdar l)))
+			(if (null? l) i
+			    (cns 'i/o fd (car l) (loop (cdr l)))))))
+	      (let loop ((l ##sys#timeout-list) (i i))
+		(if (pair? l)
+		    (loop (cdr l) (cns 'timeout (caar l) (cdar l) i))
+		    i)))))))
 
 
 ;;; Remove all waiting threads from the relevant queues with the exception of the current thread:
@@ -438,14 +466,7 @@ EOF
 
 (define (##sys#thread-unblock! t)
   (when (eq? 'blocked (##sys#slot t 3))
-    (set! ##sys#timeout-list
-      (let loop ((l ##sys#timeout-list))
-	(if (null? l) 
-	    l
-	    (let ((h (##sys#slot l 0)))
-	      (if (eq? (##sys#slot h 1) t)
-		  (##sys#slot l 1)
-		  (cons h (loop (##sys#slot l 1))))))))
+    (##sys#remove-from-timeout-list t)
     (set! ##sys#fd-list 
       (let loop ([fdl ##sys#fd-list])
 	(if (null? fdl)
