@@ -1,7 +1,7 @@
 ;;;; srfi-18.scm - Simple thread unit - felix
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -50,28 +50,13 @@
      ##sys#current-exception-handler ##sys#abandon-mutexes ##sys#check-structure ##sys#structure? ##sys#make-mutex
      ##sys#delq ##sys#compute-time-limit ##sys#fudge) ) ] )
 
-(cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-range . _) '(##core#undefined))
-    (define-macro (##sys#check-pair . _) '(##core#undefined))
-    (define-macro (##sys#check-list . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-char . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined))
-    (define-macro (##sys#check-port . _) '(##core#undefined))
-    (define-macro (##sys#check-number . _) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else
-  (declare (emit-exports "srfi-18.exports"))] )
+(include "unsafe-declarations.scm")
 
 (register-feature! 'srfi-18)
 
-(define-macro (dbg . args) #f)
-#;(define-macro (dbg . args)
-  `(print "DBG: " ,@args) )
+(define-syntax dbg
+  (syntax-rules ()
+    ((_ . _) #f)))
 
 
 ;;; Helper routines:
@@ -217,7 +202,7 @@ EOF
 	  (set! thread (make-thread thread))
 	  (##sys#check-structure thread 'thread 'thread-start!) )
       (unless (eq? 'created (##sys#slot thread 3))
-	(##sys#error 'thread-start! "thread can not be started a second time" thread) )
+	(##sys#error 'thread-start! "thread cannot be started a second time" thread) )
       (##sys#setslot thread 3 'ready)
       (##sys#add-to-ready-queue thread) 
       thread) ) )
@@ -239,7 +224,10 @@ EOF
 	    ct 1
 	    (lambda ()
 	      (case (##sys#slot thread 3)
-		[(dead) (apply return (##sys#slot thread 2))]
+		[(dead)
+		 (unless (##sys#slot ct 13) ; not unblocked by timeout
+		   (##sys#remove-from-timeout-list ct))
+		 (apply return (##sys#slot thread 2))]
 		[(terminated)
 		 (return 
 		  (##sys#signal
@@ -360,9 +348,11 @@ EOF
 		   ct 1 
 		   (lambda ()
 		     (##sys#setslot mutex 3 (##sys#delq ct (##sys#slot mutex 3)))
-		     (##sys#setslot ##sys#current-thread 8 (cons mutex (##sys#slot ##sys#current-thread 8)))
+		     (unless (##sys#slot ct 13)  ; not unblocked by timeout
+		       (##sys#remove-from-timeout-list ct))
+		     (##sys#setslot ct 8 (cons mutex (##sys#slot ct 8)))
 		     (##sys#setslot mutex 2 thread)
-		     #f) )
+		     (return #f) ))
 		  (##sys#thread-block-for-timeout! ct limit)
 		  (switch) ]
 		 [else
@@ -394,6 +384,8 @@ EOF
 		     ct 1
 		     (lambda () 
 		       (##sys#setslot cvar 2 (##sys#delq ct (##sys#slot cvar 2)))
+		       (unless (##sys#slot ct 13)  ; not unblocked by timeout
+			 (##sys#remove-from-timeout-list ct))
 		       (return #f) ) )
 		    (##sys#thread-block-for-timeout! ct limit) ]
 		   [else 
