@@ -1,18 +1,18 @@
 ;;; srfi-69.scm - Optional non-standard extensions
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 ; conditions are met:
 ;
 ;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
-;     disclaimer. 
+;     disclaimer.
 ;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
-;     disclaimer in the documentation and/or other materials provided with the distribution. 
+;     disclaimer in the documentation and/or other materials provided with the distribution.
 ;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
-;     products derived from this software without specific prior written permission. 
+;     products derived from this software without specific prior written permission.
 ;
 ; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 ; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
@@ -24,7 +24,6 @@
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
 
-
 (declare
  (unit srfi-69)
  (usual-integrations)
@@ -35,108 +34,86 @@
  [else
   (declare
     (no-bound-checks)
-    (no-procedure-checks-for-usual-bindings)
-    (bound-to-procedure
-      ##sys#check-string ##sys#check-symbol ##sys#check-exact ##sys#check-closure
-      ##sys#check-inexact ##sys#check-structure
-      ##sys#signal-hook
-      ##sys#peek-fixnum
-      ##sys#make-structure
-      ##sys#size
-      ##sys#slot ##sys#setslot
-      ##srfi-69#%equal?-hash ) ) ] )
-
-(private srfi-69
-  unbound-value-thunk
-  %object-uid-hash %eq?-hash %eqv?-hash %equal?-hash
-  %hash-table-copy %hash-table-ref %hash-table-update! %hash-table-merge!
-  %hash-table-for-each %hash-table-fold
-  hash-table-canonical-length hash-table-rehash )
+    (no-procedure-checks-for-usual-bindings) ) ] )
 
 (declare
+  (bound-to-procedure
+    ##sys#signal-hook
+    ##sys#peek-fixnum
+    ##sys#make-structure
+    ##sys#size
+    ##sys#slot ##sys#setslot
+    *equal?-hash )
   (hide
-    unbound-value-thunk
-    %object-uid-hash %eq?-hash %eqv?-hash %equal?-hash
-    %hash-table-copy %hash-table-ref %hash-table-update! %hash-table-merge!
-    %hash-table-for-each %hash-table-fold
-    hash-table-canonical-length hash-table-rehash) )
+    *eq?-hash *eqv?-hash *equal?-hash
+    *make-hash-table
+    *hash-table-copy *hash-table-merge! *hash-table-update!/default
+    *hash-table-for-each *hash-table-fold
+    hash-table-canonical-length hash-table-rehash! hash-table-check-resize! ) )
 
 (cond-expand
- [unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-closure . _) '(##core#undefined))
-    (define-macro (##sys#check-inexact . _) '(##core#undefined))
-    (define-macro (##sys#check-structure . _) '(##core#undefined))
-    (define-macro (##sys#check-symbol . _) '(##core#undefined))
-    (define-macro (##sys#check-string . _) '(##core#undefined))
-    (define-macro (##sys#check-exact . _) '(##core#undefined)) ) ]
+ [unsafe]
  [else
-  (declare (emit-exports "srfi-69.exports")) ] )
+   (declare
+     (bound-to-procedure
+       ##sys#check-string ##sys#check-symbol
+       ##sys#check-exact ##sy#check-inexact
+       ##sys#check-closure ##sys#check-structure ) ) ] )
+
+(include "unsafe-declarations.scm")
 
 (register-feature! 'srfi-69)
 
 
-;;; Unbound Value:
+;;; Naming Conventions:
 
-;; This only works because of '(no-bound-checks)'
-
-(define-macro ($unbound-value)
- '(##sys#slot '##sys#arbitrary-unbound-symbol 0) )
-
-(define unbound-value-thunk (lambda () ($unbound-value)))
-
-(define-macro ($unbound? ?val)
-  `(eq? ($unbound-value) ,?val) )
+;; %foo - inline primitive
+;; %%foo - local inline (no such thing but at least it looks different)
+;; $foo - local macro
+;; *foo - local unchecked variant of a checked procedure
+;; ##sys#foo - public, but undocumented, un-checked procedure
+;; foo - public checked procedure
+;;
 
 
 ;;; Core Inlines:
 
-(define-macro ($quick-flonum-truncate ?flo)
-  `(##core#inline "C_quickflonumtruncate" ,?flo) )
+(define-inline (%fix wrd)
+  (##core#inline "C_fix" wrd) )
 
-(define-macro ($fix ?wrd)
-  `(##core#inline "C_fix" ,?wrd) )
+(define-inline (%block? obj)
+  (##core#inline "C_blockp" obj) )
 
-(define-macro ($block? ?obj)
-  `(##core#inline "C_blockp" ,?obj) )
+(define-inline (%immediate? obj)
+  (not (##core#inline "C_blockp" obj)) )
 
-(define-macro ($special? ?obj)
-  `(##core#inline "C_specialp" ,?obj) )
+(define-inline (%special? obj)
+  (##core#inline "C_specialp" obj) )
 
-(define-macro ($port? ?obj)
-  `(##core#inline "C_portp" ,?obj) )
+(define-inline (%port? obj)
+  (##core#inline "C_portp" obj) )
 
-(define-macro ($byte-block? ?obj)
-  `(##core#inline "C_byteblockp" ,?obj) )
+(define-inline (%byte-block? obj)
+  (##core#inline "C_byteblockp" obj) )
 
-(define-macro ($hash-string ?str)
-  `(##core#inline "C_hash_string" ,?str) )
+(define-inline (%string-hash str)
+  (##core#inline "C_hash_string" str) )
 
-(define-macro ($hash-string-ci ?str)
-  `(##core#inline "C_hash_string_ci" ,?str) )
+(define-inline (%string-ci-hash str)
+  (##core#inline "C_hash_string_ci" str) )
 
-
-;;;
-
-(define-macro ($immediate? ?obj)
-  `(not ($block? ,?obj)) )
+(define-inline (%subbyte bytvec i)
+  (##core#inline "C_subbyte" bytvec i) )
 
 
 ;;; Generation of hash-values:
 
-;; Naming Conventions:
-;; $foo - macro
-;; $*foo - local macro (no such thing but at least it looks different)
-;; %foo - private, usually unchecked, procedure
-;; ##sys#foo - public, but undocumented, un-checked procedure
-;; foo - public checked procedure
-;;
 ;; All '%foo-hash' return a fixnum, not necessarily positive. The "overflow" of
 ;; a, supposedly, unsigned hash value into negative is not checked during
 ;; intermediate computation.
 ;;
-;; The body of '%eq?-hash' is duplicated in 'eqv?-hash' and the body of '%eqv?-hash'
-;; is duplicated in '%equal?-hash' to save on procedure calls.
+;; The body of '*eq?-hash' is duplicated in '*eqv?-hash' and the body of '*eqv?-hash'
+;; is duplicated in '*equal?-hash' to save on procedure calls.
 
 ;; Fixed hash-values:
 
@@ -153,77 +130,76 @@
 
 ;; Force Hash to Bounded Fixnum:
 
-(define-macro ($fxabs ?fxn)
-  `(let ([_fxn ,?fxn]) (if (fx< _fxn 0) (fxneg _fxn) _fxn ) ) )
+(define-inline (%fxabs fxn)
+  (if (fx< fxn 0) (fxneg fxn) fxn ) )
 
-(define-macro ($hash/limit ?hsh ?lim)
-  `(fxmod (fxand (foreign-value "C_MOST_POSITIVE_FIXNUM" int)
-		 ($fxabs ,?hsh))
-	  ,?lim) )
+(define-inline (%hash/limit hsh lim)
+  (fxmod (fxand (foreign-value "C_MOST_POSITIVE_FIXNUM" int)
+		(%fxabs hsh))
+	 lim) )
 
 ;; Number Hash:
 
 (define-constant flonum-magic 331804471)
 
-#| Not sure which is "better"; went with speed
-(define-macro ($subbyte ?bytvec ?i)
-  `(##core#inline "C_subbyte" ,?bytvec ,?i) )
-
-(define-macro ($hash-flonum ?flo)
-  `(fx* flonum-magic
-	,(let loop ([idx (fx- (##sys#size 1.0) 1)])
-	    (if (fx= 0 idx)
-		`($subbyte ,?flo 0)
-		`(fx+ ($subbyte ,?flo ,idx)
-		      (fxshl ,(loop (fx- idx 1)) 1))))) )
-|#
-
-(define-macro ($hash-flonum ?flo)
-  `(fx* flonum-magic ($quick-flonum-truncate ,?flo)) )
+(define-syntax $flonum-hash
+  (lambda (form r c)
+    (let ( (flo (cadr form))
+           (_%subbyte (r '%subbyte))
+           (_flonum-magic (r 'flonum-magic))
+           (_fx+ (r 'fx+))
+           (_fx* (r 'fx*))
+           (_fxshl (r 'fxshl)) )
+    `(,_fx* ,_flonum-magic
+            ,(let loop ( (idx (fx- (##sys#size 1.0) 1)) )
+               (if (fx= 0 idx)
+                   `(,_%subbyte ,flo 0)
+                   `(,_fx+ (,_%subbyte ,flo ,idx)
+                           (,_fxshl ,(loop (fx- idx 1)) 1)) ) ) ) ) ) )
 
 (define (##sys#number-hash-hook obj)
-  (%equal?-hash obj) )
+  (*equal?-hash obj) )
 
-(define-macro ($non-fixnum-number-hash ?obj)
-  `(cond [(flonum? obj)	($hash-flonum ,?obj)]
-	 [else		($fix (##sys#number-hash-hook ,?obj))] ) )
+(define-inline (%non-fixnum-number-hash obj)
+  (cond [(flonum? obj)	($flonum-hash obj)]
+	[else		(%fix (##sys#number-hash-hook obj))] ) )
 
-(define-macro ($number-hash ?obj)
-  `(cond [(fixnum? obj)	,?obj]
-	 [else		($non-fixnum-number-hash ?obj)] ) )
+(define-inline (%number-hash obj)
+  (cond [(fixnum? obj)	obj]
+	[else		(%non-fixnum-number-hash obj)] ) )
 
 (define (number-hash obj #!optional (bound hash-default-bound))
   (unless (number? obj)
     (##sys#signal-hook #:type 'number-hash "invalid number" obj) )
   (##sys#check-exact bound 'number-hash)
-  ($hash/limit ($number-hash obj) bound) )
+  (%hash/limit (%number-hash obj) bound) )
 
 ;; Object UID Hash:
 
 #; ;NOT YET (no weak-reference)
-(define (%object-uid-hash obj)
+(define-inline (%object-uid-hash obj)
   (%uid-hash (##sys#object->uid obj)) )
 
-(define (%object-uid-hash obj)
-  (%equal?-hash obj) )
+(define-inline (%object-uid-hash obj)
+  (*equal?-hash obj) )
 
 (define (object-uid-hash obj #!optional (bound hash-default-bound))
   (##sys#check-exact bound 'object-uid-hash)
-  ($hash/limit (%object-uid-hash obj) bound) )
+  (%hash/limit (%object-uid-hash obj) bound) )
 
 ;; Symbol Hash:
 
 #; ;NOT YET (no unique-symbol-hash)
-(define-macro ($symbol-hash ?obj)
-  `(##sys#slot ,?obj INDEX-OF-UNIQUE-HASH-VALUE-COMPUTED-DURING-SYMBOL-CREATION) )
+(define-inline (%symbol-hash obj)
+  (##sys#slot obj INDEX-OF-UNIQUE-HASH-VALUE-COMPUTED-DURING-SYMBOL-CREATION) )
 
-(define-macro ($symbol-hash ?obj)
-  `($hash-string (##sys#slot ,?obj 1)) )
+(define-inline (%symbol-hash obj)
+  (%string-hash (##sys#slot obj 1)) )
 
 (define (symbol-hash obj #!optional (bound hash-default-bound))
   (##sys#check-symbol obj 'symbol-hash)
-  (##sys#check-exact bound 'string-hash)
-  ($hash/limit ($symbol-hash obj) bound) )
+  (##sys#check-exact bound 'symbol-hash)
+  (%hash/limit (%symbol-hash obj) bound) )
 
 ;; Keyword Hash:
 
@@ -234,67 +210,67 @@
 		       "bad argument type - not a keyword" x) ) )
 
 #; ;NOT YET (no unique-keyword-hash)
-(define-macro ($keyword-hash ?obj)
-  `(##sys#slot ,?obj INDEX-OF-UNIQUE-HASH-VALUE-COMPUTED-DURING-KEYWORD-CREATION) )
+(define-inline (%keyword-hash obj)
+  (##sys#slot obj INDEX-OF-UNIQUE-HASH-VALUE-COMPUTED-DURING-KEYWORD-CREATION) )
 
-(define-macro ($keyword-hash ?obj)
-  `($hash-string (##sys#slot ,?obj 1)) )
+(define-inline (%keyword-hash obj)
+  (%string-hash (##sys#slot obj 1)) )
 
 (define (keyword-hash obj #!optional (bound hash-default-bound))
   (##sys#check-keyword obj 'keyword-hash)
   (##sys#check-exact bound 'keyword-hash)
-  ($hash/limit ($keyword-hash obj) bound) )
+  (%hash/limit (%keyword-hash obj) bound) )
 
 ;; Eq Hash:
 
-(define-macro ($eq?-hash-object? ?obj)
-  `(or ($immediate? ,?obj)
-       (symbol? ,?obj)
+(define-inline (%eq?-hash-object? obj)
+  (or (%immediate? obj)
+       (symbol? obj)
        #; ;NOT YET (no keyword vs. symbol issue)
-       (keyword? ,?obj) ) )
+       (keyword? obj) ) )
 
-(define (%eq?-hash obj)
+(define (*eq?-hash obj)
   (cond [(fixnum? obj)		obj]
 	[(char? obj)		(char->integer obj)]
 	[(eq? obj #t)		true-hash-value]
 	[(eq? obj #f)		false-hash-value]
 	[(null? obj)		null-hash-value]
 	[(eof-object? obj)	eof-hash-value]
-	[(symbol? obj)		($symbol-hash obj)]
+	[(symbol? obj)		(%symbol-hash obj)]
 	#; ;NOT YET (no keyword vs. symbol issue)
-	[(keyword? obj)		($keyword-hash obj)]
-	[($immediate? obj)	unknown-immediate-hash-value]
+	[(keyword? obj)		(%keyword-hash obj)]
+	[(%immediate? obj)	unknown-immediate-hash-value]
 	[else			(%object-uid-hash obj) ] ) )
 
 (define (eq?-hash obj #!optional (bound hash-default-bound))
   (##sys#check-exact bound 'eq?-hash)
-  ($hash/limit (%eq?-hash obj) bound) )
+  (%hash/limit (*eq?-hash obj) bound) )
 
 (define hash-by-identity eq?-hash)
 
 ;; Eqv Hash:
 
-(define-macro ($eqv?-hash-object? ?obj)
-  `(or ($eq?-hash-object? ,?obj)
-       (number? ,?obj)) )
+(define-inline (%eqv?-hash-object? obj)
+  (or (%eq?-hash-object? obj)
+      (number? obj) ) )
 
-(define (%eqv?-hash obj)
+(define (*eqv?-hash obj)
   (cond [(fixnum? obj)		obj]
 	[(char? obj)		(char->integer obj)]
 	[(eq? obj #t)		true-hash-value]
 	[(eq? obj #f)		false-hash-value]
 	[(null? obj)		null-hash-value]
 	[(eof-object? obj)	eof-hash-value]
-	[(symbol? obj)		($symbol-hash obj)]
+	[(symbol? obj)		(%symbol-hash obj)]
 	#; ;NOT YET (no keyword vs. symbol issue)
-	[(keyword? obj)		($keyword-hash obj)]
-	[(number? obj)		($non-fixnum-number-hash obj)]
-	[($immediate? obj)	unknown-immediate-hash-value]
+	[(keyword? obj)		(%keyword-hash obj)]
+	[(number? obj)		(%non-fixnum-number-hash obj)]
+	[(%immediate? obj)	unknown-immediate-hash-value]
 	[else			(%object-uid-hash obj) ] ) )
 
 (define (eqv?-hash obj #!optional (bound hash-default-bound))
   (##sys#check-exact bound 'eqv?-hash)
-  ($hash/limit (%eqv?-hash obj) bound) )
+  (%hash/limit (*eqv?-hash obj) bound) )
 
 ;; Equal Hash:
 
@@ -302,29 +278,31 @@
 (define-constant recursive-hash-max-depth 4)
 (define-constant recursive-hash-max-length 4)
 
-(define-macro ($*list-hash ?obj)
-  `(fx+ (length ,?obj)
-	(recursive-atomic-hash (##sys#slot ,?obj 0) depth)) )
+;; NOTE - These refer to identifiers available only within the body of '*equal?-hash'.
 
-(define-macro ($*pair-hash ?obj)
-  `(fx+ (fxshl (recursive-atomic-hash (##sys#slot ,?obj 0) depth) 16)
-	(recursive-atomic-hash (##sys#slot ,?obj 1) depth)) )
+(define-inline (%%list-hash obj)
+  (fx+ (length obj)
+       (recursive-atomic-hash (##sys#slot obj 0) depth)) )
 
-(define-macro ($*port-hash ?obj)
-  `(fx+ (fxshl (##sys#peek-fixnum ,?obj 0) 4) ; Little extra "identity"
-	(if (input-port? ,?obj)
+(define-inline (%%pair-hash obj)
+  (fx+ (fxshl (recursive-atomic-hash (##sys#slot obj 0) depth) 16)
+	(recursive-atomic-hash (##sys#slot obj 1) depth)) )
+
+(define-inline (%%port-hash obj)
+  (fx+ (fxshl (##sys#peek-fixnum obj 0) 4) ; Little extra "identity"
+	(if (input-port? obj)
 	    input-port-hash-value
 	    output-port-hash-value)) )
 
-(define-macro ($*special-vector-hash ?obj)
-  `(vector-hash ,?obj (##sys#peek-fixnum ,?obj 0) depth 1) )
+(define-inline (%%special-vector-hash obj)
+  (vector-hash obj (##sys#peek-fixnum obj 0) depth 1) )
 
-(define-macro ($*regular-vector-hash ?obj)
-  `(vector-hash ,?obj 0 depth 0) )
+(define-inline (%%regular-vector-hash obj)
+  (vector-hash obj 0 depth 0) )
 
-(define (%equal?-hash obj)
+(define (*equal?-hash obj)
 
-  ; Recurse into some portion of the vector's slots 
+  ; Recurse into some portion of the vector's slots
   (define (vector-hash obj seed depth start)
     (let ([len (##sys#size obj)])
       (let loop ([hsh (fx+ len seed)]
@@ -340,8 +318,8 @@
 
   ; Don't recurse into structured objects
   (define (recursive-atomic-hash obj depth)
-    (if (or ($eqv?-hash-object? obj)
-	    ($byte-block? obj))
+    (if (or (%eqv?-hash-object? obj)
+	    (%byte-block? obj))
 	(recursive-hash obj (fx+ depth 1))
 	other-hash-value ) )
 
@@ -355,38 +333,54 @@
 	  [(eq? obj #f)		  false-hash-value]
 	  [(null? obj)		  null-hash-value]
 	  [(eof-object? obj)	  eof-hash-value]
-	  [(symbol? obj)	  ($symbol-hash obj)]
+	  [(symbol? obj)	  (%symbol-hash obj)]
 	  #; ;NOT YET (no keyword vs. symbol issue)
-	  [(keyword? obj)	  ($keyword-hash obj)]
-	  [(number? obj)	  ($non-fixnum-number-hash obj)]
-	  [($immediate? obj)	  unknown-immediate-hash-value]
-	  [($byte-block? obj)	  ($hash-string obj)]
-	  [(list? obj)		  ($*list-hash obj)]
-	  [(pair? obj)		  ($*pair-hash obj)]
-	  [($port? obj)		  ($*port-hash obj)]
-	  [($special? obj)	  ($*special-vector-hash obj)]
-	  [else			  ($*regular-vector-hash obj)] ) )
+	  [(keyword? obj)	  (%keyword-hash obj)]
+	  [(number? obj)	  (%non-fixnum-number-hash obj)]
+	  [(%immediate? obj)	  unknown-immediate-hash-value]
+	  [(%byte-block? obj)	  (%string-hash obj)]
+	  [(list? obj)		  (%%list-hash obj)]
+	  [(pair? obj)		  (%%pair-hash obj)]
+	  [(%port? obj)		  (%%port-hash obj)]
+	  [(%special? obj)	  (%%special-vector-hash obj)]
+	  [else			  (%%regular-vector-hash obj)] ) )
 
   ;
   (recursive-hash obj 0) )
 
 (define (equal?-hash obj #!optional (bound hash-default-bound))
   (##sys#check-exact bound 'hash)
-  ($hash/limit (%equal?-hash obj) bound) )
+  (%hash/limit (*equal?-hash obj) bound) )
 
 (define hash equal?-hash)
 
 ;; String Hash:
 
-(define (string-hash str #!optional (bound hash-default-bound))
+(define (string-hash str #!optional (bound hash-default-bound) . start+end)
   (##sys#check-string str 'string-hash)
   (##sys#check-exact bound 'string-hash)
-  ($hash/limit ($hash-string str) bound) )
+  (let ((str (if (pair? start+end)
+		 (let-optionals start+end ((start 0)
+					   (end (##sys#size str)))
+		   (##sys#check-range start 0 (##sys#size str) 'string-hash) 
+		   (##sys#check-range end 0 (##sys#size str) 'string-hash) 
+		   (##sys#substring str start end) )
+		 str) ) )
+    (%hash/limit (%string-hash str) bound) ) )
 
-(define (string-ci-hash str #!optional (bound hash-default-bound))
+(define (string-ci-hash str #!optional (bound hash-default-bound) . start+end)
   (##sys#check-string str 'string-ci-hash)
   (##sys#check-exact bound 'string-ci-hash)
-  ($hash/limit ($hash-string-ci str) bound) )
+  (let ((str (if (pair? start+end)
+		 (let-optionals start+end ((start 0)
+					   (end (##sys#size str)))
+		   (##sys#check-range start 0 (##sys#size str) 'string-hash-ci) 
+		   (##sys#check-range end 0 (##sys#size str) 'string-hash-ci) 
+		   (##sys#substring str start end) )
+		 str) ) )
+  (%hash/limit (%string-ci-hash str) bound) ) )
+
+(define string-hash-ci string-ci-hash)
 
 
 ;;; Hash-Tables:
@@ -429,7 +423,7 @@
 
 ;; "Raw" make-hash-table:
 
-(define %make-hash-table
+(define *make-hash-table
   (let ([make-vector make-vector])
     (lambda (test hash len min-load max-load weak-keys weak-values initial
 	     #!optional (vec (make-vector len '())))
@@ -477,7 +471,7 @@
 			[(or (eq? core-string=? test)
 			     (eq? string=? test))	  string-hash]
 			[(or (eq? core-string-ci=? test)
-			     (eq? string-ci=? test))	  string-ci-hash]
+			     (eq? string-ci=? test))	  string-hash-ci]
 			[(or (eq? core= test)
 			     (eq? = test))		  number-hash]
 			[else				  #f] ) ) ] )
@@ -560,7 +554,7 @@
 		    (warning 'make-hash-table "user test without user hash")
 		    (set! hash equal?-hash) ) ) ) )
 	  ; Done
-	  (%make-hash-table test hash size min-load max-load weak-keys weak-values initial) ) ) ) ) )
+	  (*make-hash-table test hash size min-load max-load weak-keys weak-values initial) ) ) ) ) )
 
 ;; Hash-Table Predicate:
 
@@ -607,9 +601,47 @@
   (and-let* ([thunk (##sys#slot ht 9)])
     (thunk) ) )
 
+;; hash-table-rehash!:
+
+(define (hash-table-rehash! vec1 vec2 hash)
+  (let ([len1 (##sys#size vec1)]
+	[len2 (##sys#size vec2)] )
+    (do ([i 0 (fx+ i 1)])
+	[(fx>= i len1)]
+      (let loop ([bucket (##sys#slot vec1 i)])
+	(unless (null? bucket)
+	  (let* ([pare (##sys#slot bucket 0)]
+		 [key (##sys#slot pare 0)]
+		 [hshidx (hash key len2)] )
+	    (##sys#setslot vec2 hshidx
+			   (cons (cons key (##sys#slot pare 1)) (##sys#slot vec2 hshidx)))
+	    (loop (##sys#slot bucket 1)) ) ) ) ) ) )
+
+;; hash-table-resize!:
+
+(define (hash-table-resize! ht vec len)
+  (let* ([deslen (fxmin hash-table-max-length (fx* len hash-table-new-length-factor))]
+         [newlen (hash-table-canonical-length hash-table-prime-lengths deslen)]
+         [vec2 (make-vector newlen '())] )
+    (hash-table-rehash! vec vec2 (##sys#slot ht 4))
+    (##sys#setslot ht 1 vec2) ) )
+
+;; hash-table-check-resize!:
+
+(define-inline (hash-table-check-resize! ht newsiz)
+  (let ([vec (##sys#slot ht 1)]
+        [min-load (##sys#slot ht 5)]
+        [max-load (##sys#slot ht 6)] )
+    (let ([len (##sys#size vec)] )
+      (let ([min-load-len (inexact->exact (floor (* len min-load)))]
+            [max-load-len (inexact->exact (floor (* len max-load)))] )
+        (if (and (fx< len hash-table-max-length)
+                 (fx<= min-load-len newsiz) (fx<= newsiz max-load-len))
+          (hash-table-resize! ht vec len) ) ) ) ) )
+
 ;; hash-table-copy:
 
-(define %hash-table-copy
+(define *hash-table-copy
   (let ([make-vector make-vector])
     (lambda (ht)
       (let* ([vec1 (##sys#slot ht 1)]
@@ -617,7 +649,7 @@
 	     [vec2 (make-vector len '())] )
 	(do ([i 0 (fx+ i 1)])
 	    [(fx>= i len)
-	     (%make-hash-table
+	     (*make-hash-table
 	      (##sys#slot ht 3) (##sys#slot ht 4)
 	      (##sys#slot ht 2)
 	      (##sys#slot ht 5) (##sys#slot ht 6)
@@ -634,128 +666,189 @@
 
 (define (hash-table-copy ht)
   (##sys#check-structure ht 'hash-table 'hash-table-copy)
-  (%hash-table-copy ht) )
+  (*hash-table-copy ht) )
 
 ;; hash-table-update!:
 ;;
 ;; This one was suggested by Sven Hartrumpf (and subsequently added in SRFI-69).
 ;; Modified for ht props min & max load.
 
-(define (hash-table-rehash vec1 vec2 hash)
-  (let ([len1 (##sys#size vec1)]
-	[len2 (##sys#size vec2)] )
-    (do ([i 0 (fx+ i 1)])
-	[(fx>= i len1)]
-      (let loop ([bucket (##sys#slot vec1 i)])
-	(unless (null? bucket)
-	  (let* ([pare (##sys#slot bucket 0)]
-		 [key (##sys#slot pare 0)]
-		 [hshidx (hash key len2)] )
-	    (##sys#setslot vec2 hshidx
-			   (cons (cons key (##sys#slot pare 1))
-				 (##sys#slot vec2 hshidx)))
-	    (loop (##sys#slot bucket 1)) ) ) ) ) ) )
+(define hash-table-update!
+  (let ([core-eq? eq?] )
+    (lambda (ht key
+	     #!optional (func identity)
+		        (thunk
+		         (let ([thunk (##sys#slot ht 9)])
+		           (or thunk
+			       (lambda ()
+			         (##sys#signal-hook #:access-error
+			          'hash-table-update!
+			          "hash-table does not contain key" key ht))))))
+      (##sys#check-structure ht 'hash-table 'hash-table-update!)
+      (##sys#check-closure func 'hash-table-update!)
+      (##sys#check-closure thunk 'hash-table-update!)
+      (let ([newsiz (fx+ (##sys#slot ht 2) 1)] )
+	(hash-table-check-resize! ht newsiz)
+	(let ([hash (##sys#slot ht 4)]
+	      [test (##sys#slot ht 3)]
+	      [vec (##sys#slot ht 1)] )
+	  (let* ([len (##sys#size vec)]
+	         [hshidx (hash key len)]
+	         [bucket0 (##sys#slot vec hshidx)] )
+            (if (eq? core-eq? test)
+                ; Fast path (eq? is rewritten by the compiler):
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (let ([val (func (thunk))])
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz)
+                        val )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (eq? key (##sys#slot pare 0))
+                             (let ([val (func (##sys#slot pare 1))])
+                               (##sys#setslot pare 1 val)
+                               val)
+                             (loop (##sys#slot bucket 1)) ) ) ) )
+                ; Slow path
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (let ([val (func (thunk))])
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz)
+                        val )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (test key (##sys#slot pare 0))
+                             (let ([val (func (##sys#slot pare 1))])
+                               (##sys#setslot pare 1 val)
+                               val )
+                             (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) ) ) )
 
-(define %hash-table-update!
-  (let ([core-eq? eq?]
-	[floor floor] )
-    (lambda (ht key func thunk)
-      (let ([hash (##sys#slot ht 4)]
-	    [test (##sys#slot ht 3)]
-	    [newsiz (fx+ (##sys#slot ht 2) 1)]
-	    [min-load (##sys#slot ht 5)]
-	    [max-load (##sys#slot ht 6)] )
-	(let re-enter ()
-	  (let* ([vec (##sys#slot ht 1)]
-		 [len (##sys#size vec)] )
-	    (let ([min-load-len (inexact->exact (floor (* len min-load)))]
-		  [max-load-len (inexact->exact (floor (* len max-load)))]
-		  [hshidx (hash key len)] )
-	      ; Need to resize table?
-	      (if (and (fx< len hash-table-max-length)
-		       (fx<= min-load-len newsiz) (fx<= newsiz max-load-len))
-		  ; then resize the table:
-		  (let ([vec2 (make-vector
-			       (hash-table-canonical-length
-				hash-table-prime-lengths
-				(fxmin hash-table-max-length
-				       (fx* len hash-table-new-length-factor)))
-			       '())])
-		    (hash-table-rehash vec vec2 hash)
-		    (##sys#setslot ht 1 vec2)
-		    (re-enter) )
-		  ; else update the table:
-		  (let ([bucket0 (##sys#slot vec hshidx)])
-		    (if (eq? core-eq? test)
-			; Fast path (eq? is rewritten by the compiler):
-			(let loop ([bucket bucket0])
-			  (cond [(null? bucket)
-				 (let ([val (func (thunk))])
-				   (##sys#setslot vec hshidx (cons (cons key val) bucket0))
-				   (##sys#setislot ht 2 newsiz)
-				   val) ]
-				[else
-				 (let ([pare (##sys#slot bucket 0)])
-				   (if (eq? key (##sys#slot pare 0))
-				       (let ([val (func (##sys#slot pare 1))])
-					 (##sys#setslot pare 1 val)
-					 val)
-				       (loop (##sys#slot bucket 1)) ) ) ] ) )
-			; Slow path
-			(let loop ([bucket bucket0])
-			  (cond [(null? bucket)
-				 (let ([val (func (thunk))])
-				   (##sys#setslot vec hshidx (cons (cons key val) bucket0))
-				   (##sys#setislot ht 2 newsiz)
-				   val) ]
-				[else
-				 (let ([pare (##sys#slot bucket 0)])
-				   (if (test key (##sys#slot pare 0))
-				       (let ([val (func (##sys#slot pare 1))])
-					 (##sys#setslot pare 1 val)
-					 val)
-				       (loop (##sys#slot bucket 1)) ) ) ] ) ) ) ) ) ) ) ) ) ) ) )
-
-(define (hash-table-update!
-	 ht key
-	 #!optional (func identity)
-		    (thunk
-		     (let ([thunk (##sys#slot ht 9)])
-		       (or thunk
-			   (lambda ()
-			     (##sys#signal-hook #:access-error
-			      'hash-table-update!
-			      "hash-table does not contain key" key ht))))))
-  (##sys#check-structure ht 'hash-table 'hash-table-update!)
-  (##sys#check-closure func 'hash-table-update!)
-  (##sys#check-closure thunk 'hash-table-update!)
-  (%hash-table-update! ht key func thunk) )
+(define *hash-table-update!/default
+  (let ([core-eq? eq?] )
+    (lambda (ht key func def)
+      (let ([newsiz (fx+ (##sys#slot ht 2) 1)] )
+	(hash-table-check-resize! ht newsiz)
+	(let ([hash (##sys#slot ht 4)]
+	      [test (##sys#slot ht 3)]
+	      [vec (##sys#slot ht 1)] )
+	  (let* ([len (##sys#size vec)]
+	         [hshidx (hash key len)]
+	         [bucket0 (##sys#slot vec hshidx)] )
+            (if (eq? core-eq? test)
+                ; Fast path (eq? is rewritten by the compiler):
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (let ([val (func def)])
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz)
+                        val )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (eq? key (##sys#slot pare 0))
+                             (let ([val (func (##sys#slot pare 1))])
+                               (##sys#setslot pare 1 val)
+                               val)
+                             (loop (##sys#slot bucket 1)) ) ) ) )
+                ; Slow path
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (let ([val (func def)])
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz)
+                        val )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (test key (##sys#slot pare 0))
+                             (let ([val (func (##sys#slot pare 1))])
+                               (##sys#setslot pare 1 val)
+                               val )
+                             (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) ) ) )
 
 (define (hash-table-update!/default ht key func def)
   (##sys#check-structure ht 'hash-table 'hash-table-update!/default)
   (##sys#check-closure func 'hash-table-update!/default)
-  (%hash-table-update! ht key func (lambda () def)) )
+  (*hash-table-update!/default ht key func def) )
 
-(define (hash-table-set! ht key val)
-  (##sys#check-structure ht 'hash-table 'hash-table-set!)
-  (let ([thunk (lambda _ val)])
-    (%hash-table-update! ht key thunk thunk) )
-  (void) )
+(define hash-table-set!
+  (let ([core-eq? eq?] )
+    (lambda (ht key val)
+      (##sys#check-structure ht 'hash-table 'hash-table-set!)
+      (let ([newsiz (fx+ (##sys#slot ht 2) 1)] )
+	(hash-table-check-resize! ht newsiz)
+	(let ([hash (##sys#slot ht 4)]
+	      [test (##sys#slot ht 3)]
+	      [vec (##sys#slot ht 1)] )
+	  (let* ([len (##sys#size vec)]
+	         [hshidx (hash key len)]
+	         [bucket0 (##sys#slot vec hshidx)] )
+            (if (eq? core-eq? test)
+                ; Fast path (eq? is rewritten by the compiler):
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (begin
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz) )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (eq? key (##sys#slot pare 0))
+                             (##sys#setslot pare 1 val)
+                             (loop (##sys#slot bucket 1)) ) ) ) )
+                ; Slow path
+                (let loop ([bucket bucket0])
+                  (if (null? bucket)
+                      (begin
+                        (##sys#setslot vec hshidx (cons (cons key val) bucket0))
+                        (##sys#setislot ht 2 newsiz) )
+                      (let ([pare (##sys#slot bucket 0)])
+                         (if (test key (##sys#slot pare 0))
+                             (##sys#setslot pare 1 val)
+                             (loop (##sys#slot bucket 1)) ) ) ) ) )
+            (void) ) ) ) ) ) )
 
 ;; Hash-Table Reference:
 
-(define %hash-table-ref
+(define hash-table-ref
+  (getter-with-setter
+    (let ([core-eq? eq?])
+      (lambda (ht key #!optional (def (lambda ()
+				        (##sys#signal-hook #:access-error
+				         'hash-table-ref
+				         "hash-table does not contain key" key ht))))
+        (##sys#check-structure ht 'hash-table 'hash-table-ref)
+        (##sys#check-closure def 'hash-table-ref)
+        (let  ([vec (##sys#slot ht 1)]
+	       [test (##sys#slot ht 3)] )
+          (let* ([hash (##sys#slot ht 4)]
+		 [hshidx (hash key (##sys#size vec))] )
+	    (if (eq? core-eq? test)
+	        ; Fast path (eq? is rewritten by the compiler):
+	        (let loop ([bucket (##sys#slot vec hshidx)])
+		  (if (null? bucket)
+		      (def)
+		      (let ([pare (##sys#slot bucket 0)])
+		        (if (eq? key (##sys#slot pare 0))
+			    (##sys#slot pare 1)
+			    (loop (##sys#slot bucket 1)) ) ) ) )
+	        ; Slow path
+	        (let loop ([bucket (##sys#slot vec hshidx)])
+		  (if (null? bucket)
+		      (def)
+		      (let ([pare (##sys#slot bucket 0)])
+		        (if (test key (##sys#slot pare 0))
+			    (##sys#slot pare 1)
+			    (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) )
+   hash-table-set!) )
+
+(define hash-table-ref/default
   (let ([core-eq? eq?])
     (lambda (ht key def)
-       (let  ([vec (##sys#slot ht 1)]
-	      [test (##sys#slot ht 3)] )
-	 (let* ([hash (##sys#slot ht 4)]
-		[hshidx (hash key (##sys#size vec))] )
+      (##sys#check-structure ht 'hash-table 'hash-table-ref/default)
+      (let  ([vec (##sys#slot ht 1)]
+	     [test (##sys#slot ht 3)] )
+	(let* ([hash (##sys#slot ht 4)]
+	       [hshidx (hash key (##sys#size vec))] )
 	   (if (eq? core-eq? test)
 	       ; Fast path (eq? is rewritten by the compiler):
 	       (let loop ([bucket (##sys#slot vec hshidx)])
 		 (if (null? bucket)
-		     (def)
+		     def
 		     (let ([pare (##sys#slot bucket 0)])
 		       (if (eq? key (##sys#slot pare 0))
 			   (##sys#slot pare 1)
@@ -763,30 +856,33 @@
 	       ; Slow path
 	       (let loop ([bucket (##sys#slot vec hshidx)])
 		 (if (null? bucket)
-		     (def)
+		     def
 		     (let ([pare (##sys#slot bucket 0)])
 		       (if (test key (##sys#slot pare 0))
 			   (##sys#slot pare 1)
 			   (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) ) )
 
-(define hash-table-ref
-  (getter-with-setter
-   (lambda (ht key #!optional (def (lambda ()
-				     (##sys#signal-hook #:access-error
-				      'hash-table-ref
-				      "hash-table does not contain key" key ht))))
-     (##sys#check-structure ht 'hash-table 'hash-table-ref)
-     (##sys#check-closure def 'hash-table-ref)
-     (%hash-table-ref ht key def) )
-   hash-table-set!))
-
-(define (hash-table-ref/default ht key default)
-  (##sys#check-structure ht 'hash-table 'hash-table-ref/default)
-  (%hash-table-ref ht key (lambda () default)) )
-
-(define (hash-table-exists? ht key)
-  (##sys#check-structure ht 'hash-table 'hash-table-exists?)
-  (not ($unbound? (%hash-table-ref ht key unbound-value-thunk))) )
+(define hash-table-exists?
+  (let ([core-eq? eq?])
+    (lambda (ht key)
+      (##sys#check-structure ht 'hash-table 'hash-table-exists?)
+      (let  ([vec (##sys#slot ht 1)]
+	     [test (##sys#slot ht 3)] )
+	(let* ([hash (##sys#slot ht 4)]
+	       [hshidx (hash key (##sys#size vec))] )
+	  (if (eq? core-eq? test)
+	       ; Fast path (eq? is rewritten by the compiler):
+	       (let loop ([bucket (##sys#slot vec hshidx)])
+		 (and (not (null? bucket))
+		      (let ([pare (##sys#slot bucket 0)])
+		        (or (eq? key (##sys#slot pare 0))
+			    (loop (##sys#slot bucket 1)) ) ) ) )
+	       ; Slow path
+	       (let loop ([bucket (##sys#slot vec hshidx)])
+		 (and (not (null? bucket))
+		      (let ([pare (##sys#slot bucket 0)])
+		        (or (test key (##sys#slot pare 0))
+			    (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) ) )
 
 ;; hash-table-delete!:
 
@@ -861,7 +957,7 @@
 
 ;; Hash Table Merge:
 
-(define (%hash-table-merge! ht1 ht2)
+(define (*hash-table-merge! ht1 ht2)
   (let* ([vec (##sys#slot ht2 1)]
 	 [len (##sys#size vec)] )
     (do ([i 0 (fx+ i 1)])
@@ -869,18 +965,17 @@
       (do ([lst (##sys#slot vec i) (##sys#slot lst 1)])
 	  [(null? lst)]
 	(let ([b (##sys#slot lst 0)])
-	  (%hash-table-update! ht1 (##sys#slot b 0)
-				   identity (lambda () (##sys#slot b 1))) ) ) ) ) )
+	  (*hash-table-update!/default ht1 (##sys#slot b 0) identity (##sys#slot b 1)) ) ) ) ) )
 
 (define (hash-table-merge! ht1 ht2)
   (##sys#check-structure ht1 'hash-table 'hash-table-merge!)
   (##sys#check-structure ht2 'hash-table 'hash-table-merge!)
-  (%hash-table-merge! ht1 ht2) )
+  (*hash-table-merge! ht1 ht2) )
 
 (define (hash-table-merge ht1 ht2)
   (##sys#check-structure ht1 'hash-table 'hash-table-merge)
   (##sys#check-structure ht2 'hash-table 'hash-table-merge)
-  (%hash-table-merge! (%hash-table-copy ht1) ht2) )
+  (*hash-table-merge! (*hash-table-copy ht1) ht2) )
 
 ;; Hash-Table <-> Association-List:
 
@@ -904,10 +999,11 @@
     (lambda (alist . rest)
       (##sys#check-list alist 'alist->hash-table)
       (let ([ht (apply make-hash-table rest)])
-	(for-each (lambda (x)
-		    (%hash-table-update! ht (##sys#slot x 0)
-					    identity (lambda () (##sys#slot x 1))) )
-		  alist)
+	(for-each
+	 (lambda (x)
+	   (##sys#check-pair x 'alist->hash-table)
+	   (*hash-table-update!/default  ht (##sys#slot x 0) identity (##sys#slot x 1)) )
+	 alist)
 	ht ) ) ) )
 
 ;; Hash-Table Keys & Values:
@@ -949,7 +1045,7 @@
 ;; hash-table-fold:
 ;; hash-table-map:
 
-(define (%hash-table-for-each ht proc)
+(define (*hash-table-for-each ht proc)
   (let* ([vec (##sys#slot ht 1)]
 	 [len (##sys#size vec)] )
     (do ([i 0 (fx+ i 1)] )
@@ -958,7 +1054,7 @@
 			(proc (##sys#slot bucket 0) (##sys#slot bucket 1)) )
 		      (##sys#slot vec i)) ) ) )
 
-(define (%hash-table-fold ht func init)
+(define (*hash-table-fold ht func init)
   (let* ([vec (##sys#slot ht 1)]
 	 [len (##sys#size vec)] )
     (let loop ([i 0] [acc init])
@@ -975,19 +1071,19 @@
 (define (hash-table-fold ht func init)
   (##sys#check-structure ht 'hash-table 'hash-table-fold)
   (##sys#check-closure func 'hash-table-fold)
-  (%hash-table-fold ht func init) )
+  (*hash-table-fold ht func init) )
 
 (define (hash-table-for-each ht proc)
   (##sys#check-structure ht 'hash-table 'hash-table-for-each)
   (##sys#check-closure proc 'hash-table-for-each)
-  (%hash-table-for-each ht proc) )
+  (*hash-table-for-each ht proc) )
 
 (define (hash-table-walk ht proc)
   (##sys#check-structure ht 'hash-table 'hash-table-walk)
   (##sys#check-closure proc 'hash-table-walk)
-  (%hash-table-for-each ht proc) )
+  (*hash-table-for-each ht proc) )
 
 (define (hash-table-map ht func)
   (##sys#check-structure ht 'hash-table 'hash-table-map)
   (##sys#check-closure func 'hash-table-map)
-  (%hash-table-fold ht (lambda (k v a) (cons (func k v) a)) '()) )
+  (*hash-table-fold ht (lambda (k v a) (cons (func k v) a)) '()) )
