@@ -1,7 +1,7 @@
 ;;;; tcp.scm - Networking stuff
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
-; Copyright (c) 2008, The Chicken Team
+; Copyright (c) 2008-2009, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -27,7 +27,7 @@
 
 (declare
   (unit tcp)
-  (uses data-structures ports extras scheduler)
+  (uses extras scheduler)
   (usual-integrations)
   (fixnum-arithmetic)
   (no-bound-checks)
@@ -86,23 +86,9 @@ static char addr_buffer[ 20 ];
 EOF
 ) )
 
-(register-feature! 'tcp)
+(include "unsafe-declarations.scm")
 
-(cond-expand
- (unsafe
-  (eval-when (compile)
-    (define-macro (##sys#check-structure x y . _) '(##core#undefined))
-    (define-macro (##sys#check-range x y z) '(##core#undefined))
-    (define-macro (##sys#check-pair x) '(##core#undefined))
-    (define-macro (##sys#check-list x) '(##core#undefined))
-    (define-macro (##sys#check-symbol x) '(##core#undefined))
-    (define-macro (##sys#check-string x) '(##core#undefined))
-    (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x . _) '(##core#undefined))
-    (define-macro (##sys#check-port x . _) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))))
- (else
-  (declare (emit-exports "tcp.exports"))) )
+(register-feature! 'tcp)
 
 (define-foreign-variable errno int "errno")
 (define-foreign-variable strerror c-string "strerror(errno)")
@@ -187,7 +173,7 @@ EOF
 ) )
 
 (unless (##net#startup)
-  (##sys#signal-hook #:network-error "can not initialize Winsock") )
+  (##sys#signal-hook #:network-error "cannot initialize Winsock") )
 
 (define ##net#getservbyname 
   (foreign-lambda* int ((c-string serv) (c-string proto))
@@ -253,7 +239,7 @@ EOF
 		       (when (eq? 0 p)
 			 (##sys#update-errno)
 			 (##sys#signal-hook
-			  #:network-error 'tcp-connect (##sys#string-append "can not compute port from service - " strerror)
+			  #:network-error 'tcp-connect (##sys#string-append "cannot compute port from service - " strerror)
 			  s) )
 		       p) )
 		    (loop (fx+ i 1)) ) ) ) ) ) ) ) )
@@ -276,7 +262,7 @@ EOF
   (let ((s (##net#socket _af_inet style 0)))
     (when (eq? _invalid_socket s)
       (##sys#update-errno)
-      (##sys#error "can not create socket") )
+      (##sys#error "cannot create socket") )
     ;; PLT makes this an optional arg to tcp-listen. Should we as well?
     (when (eq? -1 ((foreign-lambda* int ((int socket)) 
 		     "int yes = 1; 
@@ -292,7 +278,7 @@ EOF
       (let ((b (##net#bind s addr _sockaddr_in_size)))
 	(when (eq? -1 b)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "can not bind to socket - " strerror) s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "cannot bind to socket - " strerror) s port) )
 	(values s addr) ) ) ) )
 
 (define-constant default-backlog 10)
@@ -304,7 +290,7 @@ EOF
       (let ((l (##net#listen s w)))
 	(when (eq? -1 l)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "can not listen on socket - " strerror) s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "cannot listen on socket - " strerror) s port) )
 	(##sys#make-structure 'tcp-listener s) ) ) ) )
 
 (define (tcp-listener? x) 
@@ -316,7 +302,7 @@ EOF
   (let ((s (##sys#slot tcpl 1)))
     (when (fx= -1 (##net#close s))
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-close (##sys#string-append "can not close TCP socket - " strerror) tcpl) ) ) )
+      (##sys#signal-hook #:network-error 'tcp-close (##sys#string-append "cannot close TCP socket - " strerror) tcpl) ) ) )
 
 (define-constant +input-buffer-size+ 1024)
 (define-constant +output-chunk-size+ 8192)
@@ -345,7 +331,7 @@ EOF
     (lambda (fd)
       (unless (##net#make-nonblocking fd)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error (##sys#string-append "can not create TCP ports - " strerror)) )
+	(##sys#signal-hook #:network-error (##sys#string-append "cannot create TCP ports - " strerror)) )
       (let* ((buf (make-string +input-buffer-size+))
 	     (data (vector fd #f #f))
 	     (buflen 0)
@@ -377,7 +363,7 @@ EOF
 				  (##sys#update-errno)
 				  (##sys#signal-hook 
 				   #:network-error
-				   (##sys#string-append "can not read from socket - " strerror) 
+				   (##sys#string-append "cannot read from socket - " strerror) 
 				   fd) ) ) )
 			  (else
 			   (set! buflen n)
@@ -399,7 +385,7 @@ EOF
 			 (##sys#update-errno)
 			 (##sys#signal-hook
 			  #:network-error
-			  (##sys#string-append "can not check socket for input - " strerror) 
+			  (##sys#string-append "cannot check socket for input - " strerror) 
 			  fd) )
 		       (eq? f 1) ) ) )
 	       (lambda ()
@@ -410,7 +396,7 @@ EOF
 		     (##sys#update-errno)
 		     (##sys#signal-hook
 		      #:network-error
-		      (##sys#string-append "can not close socket input port - " strerror)
+		      (##sys#string-append "cannot close socket input port - " strerror)
 		      fd) ) ) )
 	       #f
 	       (lambda (p n dest start)	; read-string!
@@ -428,28 +414,33 @@ EOF
 			      m
 			      (loop n m start) ) ) ) ) )
 	       (lambda (p limit)	; read-line
-		 (let loop ((str #f))
+		 (let loop ((str #f)
+			    (limit (or limit (##sys#fudge 21))))
 		   (cond ((fx< bufindex buflen)
 			  (##sys#scan-buffer-line
 			   buf 
-			   buflen
+			   (fxmin buflen limit)
 			   bufindex
 			   (lambda (pos2 next)
-			     (let ((dest (##sys#make-string (fx- pos2 bufindex))))
+			     (let* ((len (fx- pos2 bufindex))
+				    (dest (##sys#make-string len)))
 			       (##core#inline "C_substring_copy" buf dest bufindex pos2 0)
 			       (set! bufindex next)
-			       (cond ((eq? pos2 next) ; no line-terminator encountered
+			       (cond ((eq? pos2 limit) ; no line-terminator, hit limit
+				      (if str (##sys#string-append str dest) dest))
+				     ((eq? pos2 next)  ; no line-terminator, hit buflen
 				      (read-input)
 				      (if (fx>= bufindex buflen)
 					  (or str "")
-					  (loop (if str (##sys#string-append str dest) dest)) ) )
+					  (loop (if str (##sys#string-append str dest) dest)
+						(fx- limit len)) ) )
 				     (else 
 				      (##sys#setislot p 4 (fx+ (##sys#slot p 4) 1))
 				      (if str (##sys#string-append str dest) dest)) ) ) ) ) )
 			 (else
 			  (read-input)
 			  (if (fx< bufindex buflen)
-			      (loop str)
+			      (loop str limit)
 			      #!eof) ) ) ) ) ) )
 	     (output
 	      (lambda (s)
@@ -474,7 +465,7 @@ EOF
 				  (##sys#update-errno)
 				  (##sys#signal-hook 
 				   #:network-error
-				   (##sys#string-append "can not write to socket - " strerror) 
+				   (##sys#string-append "cannot write to socket - " strerror) 
 				   fd) ) ) )
 			  ((fx< n len)
 			   (loop (fx- len n) (fx+ offset n)) ) ) ) ) ) )
@@ -499,7 +490,7 @@ EOF
 		   (when (and iclosed (eq? -1 (##net#close fd)))
 		     (##sys#update-errno)
 		     (##sys#signal-hook
-		      #:network-error (##sys#string-append "can not close socket output port - " strerror) fd) ) ) )
+		      #:network-error (##sys#string-append "cannot close socket output port - " strerror) fd) ) ) )
 	       (and outbuf
 		    (lambda ()
 		      (when (fx> (##sys#size outbuf) 0)
@@ -546,7 +537,7 @@ EOF
     (when (eq? -1 f)
       (##sys#update-errno)
       (##sys#signal-hook 
-       #:network-error 'tcp-accept-ready? (##sys#string-append "can not check socket for input - " strerror) 
+       #:network-error 'tcp-accept-ready? (##sys#string-append "cannot check socket for input - " strerror) 
        tcpl) )
     (eq? 1 f) ) )
 
@@ -561,7 +552,7 @@ EOF
 (define general-strerror (foreign-lambda c-string "strerror" int))
 
 (define (tcp-connect host . more)
-  (let ((port (:optional more #f))
+  (let ((port (optional more #f))
 	(tmc (tcp-connect-timeout)))
     (##sys#check-string host)
     (unless port
@@ -574,13 +565,13 @@ EOF
 	(##net#close s)
 	(##sys#update-errno)
 	(##sys#signal-hook 
-	 #:network-error 'tcp-connect (##sys#string-append "can not connect to socket - " strerror) 
+	 #:network-error 'tcp-connect (##sys#string-append "cannot connect to socket - " strerror) 
 	 host port) )
       (when (eq? -1 s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "can not create socket - " strerror) host port) )
+	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "cannot create socket - " strerror) host port) )
       (unless (##net#gethostaddr addr host port)
-	(##sys#signal-hook #:network-error 'tcp-connect "can not find host address" host) )
+	(##sys#signal-hook #:network-error 'tcp-connect "cannot find host address" host) )
       (unless (##net#make-nonblocking s)
 	(##sys#update-errno)
 	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "fcntl() failed - " strerror)) )
@@ -604,30 +595,37 @@ EOF
 		  (loop) ) ) )
 	    (fail) ) )
       (let ((err (get-socket-error s)))
-	(cond ((= err -1) 
+	(cond ((= err -1)
+	       (##net#close s)
 	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "getsockopt() failed - " strerror)))
-	      ((> err 0) 
-	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "can not create socket - " (general-strerror err))))))
+	      ((> err 0)
+	       (##net#close s)
+	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "cannot create socket - " (general-strerror err))))))
       (##net#io-ports s) ) ) )
 
 (define (##sys#tcp-port->fileno p)
-  (##sys#slot (##sys#port-data p) 0) )
+  (let ((data (##sys#port-data p)))
+    (if (vector? data)			; a meagre test, but better than nothing
+	(##sys#slot data 0)
+	(error '##sys#tcp-port->fileno "argument does not appear to be a TCP port" p))))
 
 (define (tcp-addresses p)
+  (##sys#check-port p 'tcp-addresses)
   (let ((fd (##sys#tcp-port->fileno p)))
     (values 
      (or (##net#getsockname fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "can not compute local address - " strerror) p) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "cannot compute local address - " strerror) p) )
      (or (##net#getpeername fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "can not compute remote address - " strerror) p) ) ) ) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "cannot compute remote address - " strerror) p) ) ) ) )
 
 (define (tcp-port-numbers p)
- (let ((fd (##sys#tcp-port->fileno p)))
-   (values
-    (or (##net#getsockport fd)
-        (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "can not compute local port - " strerror) p) )
-    (or (##net#getpeerport fd)
-        (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "can not compute remote port - " strerror) p) ) ) ) )
+  (##sys#check-port p 'tcp-port-numbers)
+  (let ((fd (##sys#tcp-port->fileno p)))
+    (values
+     (or (##net#getsockport fd)
+	 (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "cannot compute local port - " strerror) p) )
+     (or (##net#getpeerport fd)
+	 (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "cannot compute remote port - " strerror) p) ) ) ) )
 
 (define (tcp-listener-port tcpl)
   (##sys#check-structure tcpl 'tcp-listener 'tcp-listener-port)
@@ -635,7 +633,7 @@ EOF
 	 (port (##net#getsockport fd)) )
     (when (eq? -1 port)
       (##sys#signal-hook
-       #:network-error 'tcp-listener-port (##sys#string-append "can not obtain listener port - " strerror) 
+       #:network-error 'tcp-listener-port (##sys#string-append "cannot obtain listener port - " strerror) 
        tcpl fd) )
     port) )
 
